From Coq Require Import Qcanon.
From iris.proofmode Require Import tactics.
From iris.bi Require Import fractional.
From iris.algebra Require Import frac.
From lrust.lifetime Require Export lifetime.
Set Default Proof Using "Type".

Class frac_borG Σ := frac_borG_inG :> inG Σ fracR.

Definition frac_bor `{invG Σ, lftG Lat E0 Σ, frac_borG Σ} κ
        (φ : Qp → monPred _ _) : monPred _ _ :=
  (* First, we close up to a bunch of things: iff, lifetime inclusion and view
     inclusion. We also existentially bind γ, the ghost name of receipts. *)
  (∃ γ κ' V0 (φ' : Qp → monPred _ _), κ ⊑ κ' ∗ monPred_in V0 ∗
    ▷ □ (∀ q, φ' q ↔ φ q) ∗
    ▷ □ ⎡(∀ q1 q2, φ' (q1+q2)%Qp ↔ φ' q1 ∗ φ' q2) V0⎤  ∗

    (* Then we use an (internal) atomic persistent borrow. *)
    &in_at{κ'} (∃ (qfresh : Qp) (qdep qused : Qc),
         (* We divide the 1 fraction into 3 parts... *)
         ⌜0 ≤ qdep⌝%Qc ∗ ⌜0 ≤ qused⌝%Qc ∗ ⌜(qfresh + qdep + qused = 1)%Qc⌝ ∗
         (* First part: the piece of φ' which is still fresh. *)
         ⎡φ' qfresh V0⎤ ∗
         (* Second part: some tokens have been left in deposit while the
            fractured borrow is accessed. *)
         (if decide (0 < qdep)%Qc is left H return _
          then ⎡(mk_Qp qdep H).[κ'] V0⎤ else True) ∗
         (* Third part: the piece of φ' that is "already used". It can be given
            back when the lifetime ends, but cannot be used when accessing the
            fractured borrow.  *)
         (if decide (0 < qused)%Qc is left H return _
          then φ' (mk_Qp qused H) else True) ∗
         (* We keep a set of "receipts", to prove that we will always have
            enough fractions of the lifetime in deposit. *)
         (if decide (0 < qfresh+qused)%Qc is left H return _
          then ⎡own γ (mk_Qp (qfresh+qused) H)⎤
          else True (* Dummy, never happens *))))%I.
Notation "&frac{ κ }" := (frac_bor κ) (format "&frac{ κ }") : bi_scope.

Section frac_bor.
  Context `{invG Σ, lftG Lat E0 Σ, frac_borG Σ}.
  Implicit Types E : coPset.

  Global Instance frac_bor_contractive κ n :
    Proper (pointwise_relation _ (dist_later n) ==> dist n) (frac_bor κ).
  Proof. solve_contractive. Qed.
  Global Instance frac_bor_ne κ n :
    Proper (pointwise_relation _ (dist n) ==> dist n) (frac_bor κ).
  Proof. solve_proper. Qed.
  Global Instance frac_bor_proper κ :
    Proper (pointwise_relation _ (≡) ==> (≡)) (frac_bor κ).
  Proof. solve_proper. Qed.

  Lemma frac_bor_iff κ φ φ' :
    ▷ □ (∀ q, φ q ↔ φ' q) -∗ &frac{κ}φ -∗ &frac{κ}φ'.
  Proof.
    iIntros "#Hφφ' H". iDestruct "H" as (γ κ' V0 φ0) "(? & ? & #Hφ0φ & ?)".
    iExists γ, κ', V0, φ0. iFrame. iNext. iAlways. iIntros (q).
    iSplit; iIntros "H".
    - iApply "Hφφ'". by iApply "Hφ0φ".
    - iApply "Hφ0φ". by iApply "Hφφ'".
  Qed.

  Global Instance frac_bor_persistent φ κ : Persistent (&frac{κ}φ) := _.

  Lemma bor_fracture' `{LatBottom Lat} φ E κ :
    ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗
      ▷ □ (∀ q1 q2, φ (q1+q2)%Qp ↔ φ q1 ∗ φ q2) -∗ &{κ}(φ 1%Qp) ={E}=∗ &frac{κ}φ.
  Proof.
    iIntros (?) "#LFT #Hfrac Hφ". iMod (own_alloc 1%Qp) as (γ) "Hown". done.
    iMod (bor_acc_atomic_strong with "LFT Hφ") as "[H|H]". done.
    - iDestruct "H" as (P' κ') "(#Hκκ' & Hφ & Hclose)". iCombine "Hfrac Hφ" as "HH".
      iDestruct (monPred_in_intro with "HH") as (V) "(#HV & #Hfrac' & Hφ)".
      iMod ("Hclose" with "[] [-]") as "Hφ"; last first.
      { iExists γ, κ', V, φ. iFrame "#". iSplitR; [|iSplitR].
        - iIntros "!>!>!#*". by iApply (bi.iff_refl True%I).
        - iModIntro. iNext. iAlways. iIntros (??).
          rewrite !bi.intuitionistically_elim. iApply ("Hfrac'" $! _ _).
        - by iApply (in_at_bor_share with "[Hφ]"). }
      { iNext. iSplit. iDestruct "Hφ" as "[$ _]". iExists 1%Qp, 0, 0.
        iDestruct "Hφ" as "[_ Hφ]". repeat (iSplit; [done|]).
        rewrite /= -(proj2 (Qp_eq 1 (mk_Qp (1+0) _))) //. iFrame. }
      iIntros "!> Hφ H† !>". iNext.
      iDestruct "Hφ" as (q q' q'' ?? Hsum) "(Hφ1 & Htok & Hφ2 & _)".
      destruct (decide (0 < q')%Qc).
      { iExFalso. rewrite -(embed_pure False). iModIntro ⎡_⎤%I.
        iApply (lft_tok_dead_subj with "Htok [H†]"). iApply "H†". }
      destruct (Qcle_lt_or_eq 0 q') as [|<-]; [done|done|].
      rewrite Qcplus_0_r in Hsum. destruct (decide (0 < q'')%Qc).
      { iDestruct ("Hfrac" with "[Hφ1 Hφ2]") as "H".
        - iSplitL "Hφ1"; [|done]. by iApply monPred_in_elim.
        - by rewrite (proj2 (Qp_eq (q + mk_Qp q'' _) 1) Hsum). }
      destruct (Qcle_lt_or_eq 0 q'') as [|<-]; [done|done|]. rewrite Qcplus_0_r in Hsum.
      rewrite (proj2 (Qp_eq q 1) Hsum). by iApply monPred_in_elim.
    - iDestruct (monPred_in_intro with "Hfrac") as (V) "[HV Hfrac']".
      iDestruct "H" as (κ') "(#? & #? & >_)". iExists γ, κ', V, φ. iFrame "#".
      iSplitR; [|iSplitR].
      + iIntros "!> !> !# *". by iApply (bi.iff_refl True%I).
      + iModIntro. iNext. iAlways. iIntros (??).
        rewrite !bi.intuitionistically_elim. iApply ("Hfrac'" $! _ _).
      + by iApply in_at_bor_fake.
  Qed.
  Lemma bor_fracture `{LatBottom Lat} P φ `{!AsFractional P φ 1} E κ :
    ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ &{κ}P ={E}=∗ &frac{κ}φ.
  Proof.
    iIntros (?) "#? ?". rewrite [P]as_fractional. iApply bor_fracture'=>//.
    iIntros "!> !# %%". by rewrite fractional -(bi.iff_refl True%I).
  Qed.

  Lemma frac_bor_acc `{LatBottom Lat} φ E q κ :
    ↑lftN ⊆ E →
    ⎡lft_ctx⎤ -∗ &frac{κ}φ -∗ q.[κ] ={E}=∗ ∃ q', ▷ φ q' ∗ (▷ φ q' ={E}=∗ q.[κ]).
  Proof.
    assert (Inhabited Qc) by exact (populate 0%Qc).
    iIntros (?) "#LFT Hfrac Hκ".
    iDestruct "Hfrac" as (γ κ' V0 φ') "#(Hκκ' & HV0 & Hφ' & Hfrac & Hshr)".
    iMod (lft_incl_acc with "Hκκ' Hκ") as (qκ') "[Hκ Hclose]". done.
    rewrite -{1}(Qp_div_2 qκ'). rewrite lft_tok_split_obj.
    iDestruct "Hκ" as "[Hκ1 Hκ2]".
    iMod (in_at_bor_acc with "LFT Hshr Hκ1") as (Vb) "[H Hclose']"; try done.
    iCombine "H HV0" as "HH".
    iDestruct (monPred_in_intro with "HH") as (V) "(#HV & H & %)".
    iAssert ⎡monPred_in Vb (Vb ⊔ V)⎤%I as "HVVb"; [solve_lat|].
    iDestruct ("H" with "HVVb")
      as (qφ0 qtok qφ1) "(>% & >% & >% & Hφ0 & Htok' & Hφ1 & Hown)".
    destruct (Qp_lower_bound (qκ'/2) (qφ0/2)) as (qq & qκ'0 & qφ0' & Hqκ' & Hqφ).
    iExists qq.
    iAssert (▷ φ qq ∗ ⎡▷ φ' (qφ0' + qφ0 / 2) V0⎤)%Qp%I with "[Hφ0]" as "[$ Hqφ]".
    { rewrite -{1}(Qp_div_2 qφ0) {1}Hqφ -assoc_L embed_later -bi.later_sep. iNext.
      iDestruct ("Hfrac" with "Hφ0") as "[Hφ0 $]". iApply "Hφ'".
      by iApply (monPred_in_elim with "HV0"). }
    iAssert (▷ ⎡own γ qq⎤ ∗
             ▷ if decide (0 < qφ0' + qφ0 / 2 + qφ1)%Qc is left H
             then ⎡own γ (mk_Qp _ H)⎤ else True)%I with "[Hown]" as "[>Hown1 Hown2]".
    { destruct (decide (0 < qφ0 + qφ1)%Qc) as [|[]]; last by apply Qcplus_pos_nonneg.
      destruct decide as [|[]]; first last.
      { assert (0 < / 2)%Qc by reflexivity.
        auto using Qcplus_pos_nonneg, Qcplus_pos_pos, Qcmult_pos_pos. }
      rewrite -bi.later_sep -embed_sep -own_op frac_op' monPred_at_embed. iNext.
      rewrite [(_ + _)%Qp](_ : _ = {| Qp_car := qφ0 + qφ1 |}) //.
      apply Qp_eq. apply Qp_eq in Hqφ. simpl in Hqφ.
      rewrite /= !Qcplus_assoc -Hqφ (Qcplus_diag (qφ0/2)) Qcmult_div_r //. }
    rewrite monPred_objectively_unfold Hqκ'. iDestruct ("Hκ2" $! V0) as "[Hκ2 Hκ2']".
    iMod ("Hclose'" with "[-Hown1 Hclose Hκ2']") as "Hκ1".
    { iIntros "#HVb". iExists (qφ0' + qφ0 / 2)%Qp, (qtok + qq)%Qc, qφ1. iNext. iFrame.
      iSplit; [|iSplit; [done|iSplit; [|iSplitR "Hφ1"]]].
      - iPureIntro. apply Qcplus_nonneg_nonneg, Qclt_le_weak=>//.
      - iPureIntro. simpl.
        replace (qφ0' + qφ0 / 2%Z + (qtok + qq))%Qc
          with (qq + qφ0' + qφ0 / 2%Z + qtok)%Qc by ring.
        apply Qp_eq in Hqφ. simpl in Hqφ. rewrite -Hqφ Qcplus_diag Qcmult_div_r //.
      - destruct (decide (0 < qtok)%Qc), (decide (0 < qtok + qq)%Qc)=>//.
        + erewrite (proj2 (Qp_eq (mk_Qp (_ + _) _) (_ + qq))).
          rewrite lft_tok_fractional monPred_at_embed. by iFrame "Htok'". done.
        + destruct (Qcle_lt_or_eq 0 qtok) as [|<-]=>//.
          erewrite (proj2 (Qp_eq (mk_Qp (_ + _) _) qq)). iFrame.
          simpl. apply Qcplus_0_l.
      - destruct (decide (0 < qφ1)%Qc); [|done]. iCombine "HV HVb" as "HH".
        iDestruct (monPred_in_intro with "HH") as (VV) "(HVV & % & %)".
        iApply (monPred_in_elim with "HVV"). iFrame. }
    rewrite -Hqκ'. iClear "Hκκ' HV". clear dependent qφ0' qtok qφ1 Vb qφ0 V.
    iIntros "!> Hφ". iApply ("Hclose" with "[>-]"). clear κ.
    iMod (in_at_bor_acc with "LFT Hshr Hκ1") as (Vb) "[H Hclose']"; [done|].
    iCombine "H HV0" as "HH".
    iDestruct (monPred_in_intro with "HH") as (V) "(#HV & H & %)".
    iAssert ⎡monPred_in Vb (Vb ⊔ V)⎤%I as "HVVb"; [solve_lat|].
    iDestruct ("H" with "HVVb")
      as (qφ0 qtok qφ1) "(>% & >% & >% & Hφ0 & Htok' & Hφ1 & Hown)".
    destruct (decide (0 < qφ0 + qφ1)%Qc) as [|[]]; last by auto using Qcplus_pos_nonneg.
    iClear "Hshr". rewrite monPred_at_embed. iDestruct "Hown" as ">Hown".
    iCombine "Hown1" "Hown" as "Hown". iDestruct (own_valid with "Hown") as %Hval.
    assert (qq + qφ0 + qφ1 ≤ 1)%Qc. {
      destruct (decide (0 < qφ0 + qφ1)%Qc) as [|[]]; [|by apply Qcplus_pos_nonneg].
      by rewrite -Qcplus_assoc. }
    assert (0 ≤ qtok - qq)%Qc. {
      replace (qtok - qq)%Qc with ((qφ0 + qtok + qφ1) - (qq + qφ0 + qφ1))%Qc by ring.
      change 0%Qc with (1 - 1)%Qc. apply Qcplus_le_compat.
        by apply reflexive_eq. by apply Qcopp_le_compat. }
    replace qtok with (qtok - qq + qq)%Qc by ring.
    destruct (decide (0 < qtok - qq + qq)%Qc) as [|[]]; last by apply Qcplus_nonneg_pos.
    assert (∀ H,
      (mk_Qp (qtok - qq + qq) H).[κ'] V0 ⊣⊢
       (if decide (0 < qtok - qq)%Qc is left H return _ then (mk_Qp _ H).[κ'] V0 else True)
       ∗ qq.[κ'] V0) as ->.
    { intros ?. destruct (decide (0 < qtok - qq)%Qc) as [|EQ%Qcnot_lt_le%Qcle_antisym]=>//.
      - rewrite -monPred_at_sep -lft_tok_fractional. by erewrite (proj2 (Qp_eq _ _)).
      - erewrite (proj2 (Qp_eq _ _)), (left_id _ _)=>//=. by rewrite -EQ Qcplus_0_l. }
    iDestruct "Htok'" as "[Htok1 >Htok2]". iCombine "Htok2" "Hκ2'" as "Htok2".
    rewrite -monPred_at_sep -lft_tok_fractional -Hqκ'.
    iDestruct (monPred_in_elim with "HV0 Htok2") as "$".
    iApply "Hclose'". iIntros "#HVb". iExists qφ0, (qtok - qq)%Qc, (qφ1 + qq)%Qc.
    replace (qφ0 + (qtok - qq) + (qφ1 + qq))%Qc with (qφ0 + qtok + qφ1)%Qc by ring.
    iNext. iFrame "Hφ0 %". iSplit; [|iSplitL "Htok1"; [|iSplitR "Hown"]].
    - iPureIntro. apply Qcplus_nonneg_nonneg, Qclt_le_weak=>//.
    - by destruct (decide (0 < qtok - qq))%Qc.
    - destruct (decide (0 < qφ1 + qq))%Qc as [|[]]; last by apply Qcplus_nonneg_pos.
      destruct (decide (0 < qφ1)%Qc) as [|<-%Qcnot_lt_le%Qcle_antisym]=>//.
      + erewrite (proj2 (Qp_eq (mk_Qp (_ + _) _) _)).
        * rewrite !bi.intuitionistically_elim.
          iDestruct (monPred_in_elim with "HV0 Hfrac") as "Hfrac'".
          iApply "Hfrac'". iSplitL "Hφ"; [by iApply "Hφ'"|]. iCombine "HV HVb" as "HH".
          iDestruct (monPred_in_intro with "HH") as (VV) "(HVV & % & %)".
          iApply (monPred_in_elim with "HVV"). iFrame.
        * simpl. ring.
      + erewrite (proj2 (Qp_eq (mk_Qp _ _) qq)). by iApply "Hφ'". simpl; ring.
    - destruct decide as [|[]].
      + erewrite ->(proj2 (Qp_eq _ _)). iFrame. simpl. ring.
      + auto using Qcplus_pos_pos, Qcplus_nonneg_pos.
  Qed.

  Lemma frac_bor_shorten φ κ κ' : κ ⊑ κ' -∗ &frac{κ'}φ -∗ &frac{κ}φ.
  Proof.
    iIntros "Hκκ' H". iDestruct "H" as (γ κ0 V0 φ') "[H ?]".
    iExists γ, κ0, V0, φ'. iFrame. by iApply (lft_incl_trans with "Hκκ'").
  Qed.

  Lemma frac_bor_lft_incl `{LatBottom Lat} κ κ' q:
    ⎡lft_ctx⎤ -∗ &frac{κ}(λ q', (q * q').[κ']) -∗ κ ⊑ κ'.
  Proof.
    iIntros "#LFT#Hbor". iApply lft_incl_intro. iAlways. iIntros (q') "Hκ'".
    iMod (frac_bor_acc with "LFT Hbor Hκ'") as (q'') "[>Htok Hclose]". done.
    iExists _. iIntros "{$Htok} !> Hκ'". by iApply "Hclose".
  Qed.
End frac_bor.

Typeclasses Opaque frac_bor.
