From iris.proofmode Require Import tactics.
From iris.algebra Require Import auth csum frac agree.
From iris.bi Require Import fractional.
From lrust.lang Require Import memcpy.
From lrust.lifetime Require Import at_borrow.
From lrust.typing Require Import typing option.
From lrust.typing.lib.rwlock Require Import rwlock rwlockreadguard rwlockwriteguard.
From gpfsl.gps Require Import middleware protocols.
From lrust.logic Require Import gps.
Set Default Proof Using "Type".

(** SYNCHRONIZATION CONDITIONS of rwlock *)
(*  ---> stands for Release-Acquire synchronization.
  write_unlock  --->  read_lock
  read_unlock   --->  read_lock
  read_unlock   --->  write_lock

  This leads to the following C11 implementation:
  - write_unlock uses a release write
  - read_lock uses an acquire CAS
  - write_lock uses an acquire CAS
  - read_unlock uses a release CAS *)

Section rwlock_functions.
  Context `{typeG Σ, rwlockG Σ}.

  (* Constructing a rwlock. *)
  Definition rwlock_new ty : val :=
    funrec: <> ["x"] :=
      let: "r" := new [ #(S ty.(ty_size))] in
      "r" +ₗ #0 <- #0;;
      "r" +ₗ #1 <-{ty.(ty_size)} !"x";;
       delete [ #ty.(ty_size) ; "x"];; return: ["r"].

  Lemma rwlock_new_type ty `{!TyWf ty} :
    typed_val (rwlock_new ty) (fn(∅; ty) → rwlock ty).
  Proof.
    intros E L. iApply type_fn; [solve_typing..|]. iIntros "/= !#".
      iIntros (_ ret ϝ arg). inv_vec arg=>x. simpl_subst.
    iApply (type_new (S ty.(ty_size))); [solve_typing..|].
    iIntros (r tid) "#LFT HE Hna HL Hk HT". simpl_subst.
    rewrite tctx_interp_cons tctx_interp_singleton !tctx_hasty_val.
    iDestruct "HT" as "[Hr Hx]". destruct x as [[|lx|]|]; try done.
    iDestruct "Hx" as "[Hx Hx†]". iDestruct "Hx" as (vl) "[Hx↦ Hx]".
    destruct r as [[|lr|]|]; try done. iDestruct "Hr" as "[Hr Hr†]".
    iDestruct "Hr" as (vl') "Hr". rewrite uninit_own.
    iDestruct "Hr" as "[Hr↦ >SZ]". destruct vl' as [|]; iDestruct "SZ" as %[=].
    rewrite heap_mapsto_vec_cons. iDestruct "Hr↦" as "[Hr↦0 Hr↦1]". wp_op.
    rewrite shift_loc_0. wp_write. wp_op. iDestruct (ty.(ty_size_eq) with "Hx") as %Hsz.
    wp_apply (wp_memcpy with "[$Hr↦1 $Hx↦]"); [done||lia..|].
    iIntros "[Hr↦1 Hx↦]". wp_seq.
    iApply wp_hist_inv; [done|iIntros "HI"].
    iApply (type_type _ _ _ [ #lx ◁ box (uninit (ty_size ty)); #lr ◁ box (rwlock ty)]
        with "[] LFT HE Hna HL Hk [-]"); last first.
    { rewrite tctx_interp_cons tctx_interp_singleton !tctx_hasty_val' //=. iFrame.
      iSplitL "Hx↦".
      - iExists _. rewrite uninit_own. auto.
      - iExists (_::_). rewrite heap_mapsto_vec_cons. iFrame. simpl. iFrame. auto. }
    iApply type_delete; [solve_typing..|].
    iApply type_jump; solve_typing.
  Qed.

  (* The other direction: getting ownership out of a rwlock.
     Note: as we ignore poisonning, this cannot fail. *)
  Definition rwlock_into_inner ty : val :=
    funrec: <> ["x"] :=
      let: "r" := new [ #ty.(ty_size)] in
      "r" <-{ty.(ty_size)} !("x" +ₗ #1);;
       delete [ #(S ty.(ty_size)) ; "x"];; return: ["r"].

  Lemma rwlock_into_inner_type ty `{!TyWf ty} :
    typed_val (rwlock_into_inner ty) (fn(∅; rwlock ty) → ty).
  Proof.
    intros E L. iApply type_fn; [solve_typing..|]. iIntros "/= !#".
      iIntros (_ ret ϝ arg). inv_vec arg=>x. simpl_subst.
    iApply (type_new ty.(ty_size)); [solve_typing..|].
    iIntros (r tid) "#LFT HE Hna HL Hk HT". simpl_subst.
    rewrite tctx_interp_cons tctx_interp_singleton !tctx_hasty_val.
    iDestruct "HT" as "[Hr Hx]". destruct x as [[|lx|]|]; try done.
    iDestruct "Hx" as "[Hx Hx†]".
    iDestruct "Hx" as ([|[[]|]vl]) "[Hx↦ [HI Hx]]"; try iDestruct "Hx" as ">[]".
    destruct r as [[|lr|]|]; try done. iDestruct "Hr" as "[Hr Hr†]".
    iDestruct "Hr" as (vl') "Hr". rewrite uninit_own heap_mapsto_vec_cons.
    iDestruct "Hr" as "[Hr↦ >%]". iDestruct "Hx↦" as "[Hx↦0 Hx↦1]". wp_op.
    iDestruct "Hx" as "[% Hx]". iDestruct (ty.(ty_size_eq) with "Hx") as %Hsz.
    wp_apply (wp_memcpy with "[$Hr↦ $Hx↦1]"); [done||lia..|].
    iIntros "[Hr↦ Hx↦1]". wp_seq.
    iApply (type_type _ _ _ [ #lx ◁ box (uninit (S (ty_size ty))); #lr ◁ box ty]
        with "[] LFT HE Hna HL Hk [-]"); last first.
    { rewrite tctx_interp_cons tctx_interp_singleton !tctx_hasty_val' //. iFrame.
      iSplitR "Hr↦ Hx".
      - iExists (_::_). rewrite heap_mapsto_vec_cons uninit_own -Hsz. iFrame. auto.
      - iExists vl. iFrame. }
    iApply type_delete; [solve_typing..|].
    iApply type_jump; solve_typing.
  Qed.

  Definition rwlock_get_mut : val :=
    funrec: <> ["x"] :=
      let: "x'" := !"x" in
      "x" <- "x'" +ₗ #1;;
      return: ["x"].

  Lemma rwlock_get_mut_type ty `{!TyWf ty} :
    typed_val rwlock_get_mut (fn(∀ α, ∅; &uniq{α} (rwlock ty)) → &uniq{α} ty).
  Proof.
    intros E L. iApply type_fn; [solve_typing..|]. iIntros "/= !#".
      iIntros (α ϝ ret arg). inv_vec arg=>x. simpl_subst.
    iApply type_deref; [solve_typing..|]. iIntros (x'). simpl_subst.
    iIntros (tid) "#LFT HE Hna HL HC HT".
    rewrite tctx_interp_cons tctx_interp_singleton !tctx_hasty_val.
    iDestruct "HT" as "[Hx Hx']". destruct x' as [[|lx'|]|];  try iDestruct "Hx'" as "[]".
    iAssert (&{α} (⎡ hist_inv ⎤ ∗ ∃ (z : Z), lx' ↦ #z ∗ ⌜-1 ≤ z⌝) ∗
        (&uniq{α} ty).(ty_own) tid [ #(lx' +ₗ 1)])%I with "[> Hx']" as "[_ Hx']".
    { iApply bor_sep; [done..|]. iApply (bor_proper with "Hx'"). iSplit.
      - iIntros "[H1 H2]". iDestruct "H1" as "[? H1]".
        iDestruct "H1" as (z) "[??]". iDestruct "H2" as (vl) "[??]".
        iExists (_::_). rewrite heap_mapsto_vec_cons. iFrame. iFrame.
      - iIntros "H". iDestruct "H" as ([|[[| |z]|]vl]) "[H↦ [HI H]]"; try done.
        rewrite heap_mapsto_vec_cons. iFrame "HI".
        iDestruct "H" as "[H1 H2]". iDestruct "H↦" as "[H↦1 H↦2]".
        iSplitL "H1 H↦1"; eauto. iExists _. iFrame. }
    destruct x as [[|lx|]|]; try done. iDestruct "Hx" as "[Hx Hx†]".
    iDestruct "Hx" as (vl) "[Hx↦ Hx]". rewrite uninit_own. wp_op.
    iApply (type_type _ _ _
            [ #lx ◁ box (uninit 1); #(lx' +ₗ 1) ◁ &uniq{α}ty]
            with "[] LFT HE Hna HL HC [-]"); last first.
    { rewrite tctx_interp_cons tctx_interp_singleton !tctx_hasty_val' //. iFrame.
      iNext. iExists _. rewrite uninit_own. iFrame. }
    iApply type_assign; [solve_typing..|].
    iApply type_jump; solve_typing.
  Qed.

  (* Acquiring a read lock. *)
  Definition rwlock_try_read : val :=
    funrec: <> ["x"] :=
      let: "r" := new [ #2 ] in
      let: "x'" := !"x" in
    withcont: "k":
    withcont: "loop":
      "loop" []
    cont: "loop" [] :=
      let: "n" := !ʳˡˣ"x'" in
        if: "n" ≤ #-1 then
          "r" <-{Σ none} ();;
          "k" []
        else
          if: CAS "x'" "n" ("n" + #1) Relaxed AcqRel Relaxed then
            "r" <-{Σ some} "x'";;
            "k" []
          else "loop" []
    cont: "k" [] :=
      delete [ #1; "x"];; return: ["r"].

  Lemma rwlock_try_read_type ty `{!TyWf ty} :
    typed_val rwlock_try_read
        (fn(∀ α, ∅; &shr{α}(rwlock ty)) → option (rwlockreadguard α ty)).
  Proof.
    intros E L. iApply type_fn; [solve_typing..|]. iIntros "/= !#".
      iIntros (α ϝ ret arg). inv_vec arg=>x. simpl_subst.
    iApply type_new; [solve_typing..|]. iIntros (r). simpl_subst.
    iApply type_deref; [solve_typing..|]. iIntros (x'). simpl_subst.
    iApply (type_cont [] [ϝ ⊑ₗ []] (λ _, [x ◁ box (&shr{α}(rwlock ty));
                                            r ◁ box (option (rwlockreadguard α ty))]));
      [iIntros (k)|iIntros "/= !#"; iIntros (k arg); inv_vec arg];
      simpl_subst; last first.
    { iApply type_delete; [solve_typing..|].
      iApply type_jump; solve_typing. }
    iApply (type_cont [] [ϝ ⊑ₗ []] (λ _, [x ◁ _; x' ◁ _; r ◁ _]));
      [iIntros (loop)|iIntros "/= !#"; iIntros (loop arg); inv_vec arg];
      simpl_subst.
    { iApply type_jump; solve_typing. }
    iIntros (tid) "#LFT #HE Hna HL Hk HT".
    rewrite 2!tctx_interp_cons tctx_interp_singleton !tctx_hasty_val.
    iDestruct "HT" as "(Hx & Hx' & Hr)". destruct x' as [[|lx|]|]; try done.
    iDestruct "Hx'" as (β) "#[Hαβ Hinv]".
    iMod (lctx_lft_alive_tok α with "HE HL") as (qα) "(Hα & HL & Hclose)";
      [solve_typing..|].
    iMod (lft_incl_acc with "Hαβ Hα") as (qβ) "[Hβtok Hclose']". done.
    iDestruct (lft_tok_split_obj (qβ/2) (qβ/2) β with "[Hβtok]")
      as "[Hβtok1 Hβtok2]". { by rewrite Qp_div_2. }
    iDestruct "Hinv" as (γ γs tyO tyS) "(EqO & EqS & R)".
    iDestruct "R" as (tR vR) "R".
    wp_bind (!ʳˡˣ(LitV lx))%E.
    iApply (GPS_aSP_Read rwlockN _ _ _ β _
              (λ _ _ v, ∃ st, ⌜v = #(Z_of_rwlock_st st)⌝)%I
              with "[$LFT $Hβtok1 $R]"); [solve_ndisj|done..|by left| |].
    { iIntros "!>" (t' [] v [_ Ext']) "!>". iSplit; last iSplit.
      - iDestruct 1 as (st Eqvs) "F".
        iIntros "!>". iSplitR ""; iExists st; [by iFrame "F"|done].
      - iDestruct 1 as (st Eqvs) "F".
        iIntros "!>". iSplitR ""; iExists st; [by iFrame "F"|done].
      - iIntros (Eqvs) "!>". iSplit; [done|]. by iExists (Some (inl ())). }
    iIntros "!>" (t' [] v') "(_ & #R' & Hβtok1 & Int)".
    iDestruct (acq_exist with "Int") as (st') "Int".
    iDestruct (acq_pure_elim with "Int") as %?. subst v'.
    wp_let. wp_op; case_bool_decide as Hm1; wp_if.
    - iDestruct (monPred_objectively_elim with "Hβtok2") as "Hβtok2".
      iMod ("Hclose'" with "[$]") as "Hα". iMod ("Hclose" with "Hα HL") as "HL".
      iApply (type_type _ _ _
              [ x ◁ box (&shr{α}(rwlock ty)); r ◁ box (uninit 2) ]
              with "[] LFT HE Hna HL Hk"); first last.
      { rewrite tctx_interp_cons tctx_interp_singleton !tctx_hasty_val. iFrame. }
      iApply (type_sum_unit (option $ rwlockreadguard α ty));
        [solve_typing..|]; first last.
      simpl. iApply type_jump; solve_typing.
    - wp_op. wp_bind (CAS _ _ _ _ _ _).
      set Q : time → () → vProp Σ :=
        (λ t _, (qβ / 2).[β] ∗
          (∃ qν ν, (qν).[ν] ∗
            (□ tyS (β ⊓ ν)) ∗
            rwown γs (reading_st qν ν) ∗
            (□ ((1).[ν] ={↑lftN ∪ ↑histN,↑histN}▷=∗ [†ν])) ∗
            GPS_SWSharedReader lx t (() : unitProtocol) #(Z_of_rwlock_st st' + 1) qν γ))%I.
      set Q1: time → () → vProp Σ :=
        (λ t _, rwlock_interp γs β tyO tyS false lx γ t () #(Z_of_rwlock_st st'))%I.
      set Q2: time → () → vProp Σ :=
        (λ t _, rwlock_interp γs β tyO tyS true lx γ t () #(Z_of_rwlock_st st'))%I.
      set R : time → () → lit → vProp Σ := (λ _ _ _, True)%I.
      set P : vProp Σ := (<obj> (qβ / 2).[β])%I.
      iMod (rel_True_intro True%I tid with "[//]") as "#rTrue".

      iApply (GPS_aSP_SharedWriter_CAS_int_weak rwlockN (rwlock_interp γs β tyO tyS)
                rwlock_noCAS _ β _ _ _ _ (Z_of_rwlock_st st') #(Z_of_rwlock_st st' + 1)
                t' () _ _ P Q Q1 Q2 R _ (⊤ ∖ ↑rwlockN)
                with "[$LFT $Hβtok1 $R' Hβtok2 rTrue]");
        [solve_ndisj|done..|by left|by right|by left|iSplitL ""|].
      { iIntros "!> !>" (?? v1 ?) "[F|[F|F]]".
        - iApply rwlock_interp_comparable. by iLeft.
        - iApply rwlock_interp_comparable. by iRight.
        - iDestruct "F" as %?. subst v1. iExists (-1). iSplit; [done|].
          iPureIntro. by constructor. }
      { simpl. iSplitL ""; last first.
        { rewrite -{2}(bi.True_sep' P).
          iApply (rel_sep_objectively with "[$rTrue Hβtok2]"). by iModIntro. }
        rewrite -(bi.True_sep' (∀ _ _, _)%I).
        iApply (rel_sep_objectively with "[$rTrue]").
        iIntros "!>" (t1 [] [_ Ext1]). iSplit; last first; last iSplitL ""; last iSplitL "".
        { iIntros "!>" (v ?) "!>". iSplit; last iSplit; by iIntros "$". }
        { rewrite /rwlock_noCAS. iIntros "!> _" (Eq).
          inversion Eq as [Eq1]. rewrite Eq1 in Hm1. exfalso. by apply Hm1. }
        { iIntros "!>". iDestruct 1 as (st2) "(>#Eq & #Share & F)".
          iIntros "!>". iSplitL ""; iNext; iExists st2; by iFrame "Eq Share". }
        iIntros "Hβtok2". iDestruct 1 as (st2) "(>Eqst & #Share & g & INT)".
        iDestruct "Eqst" as %Eqst. inversion Eqst as [Eqst2]. clear Eqst.
        iDestruct (monPred_objectively_elim with "Hβtok2") as "Hβtok2".
        iModIntro.
        destruct st2 as [[[]|[[ν q] n]]|]; simpl in Eqst2.
        - exfalso. clear -Eqst2 Hm1. rewrite Eqst2 in Hm1. omega.
        - rewrite Eqst2. iDestruct "INT" as "[$ INT]".
          iExists (). iSplitL ""; [done|]. iIntros (t2 Ext2) "W' !>". iSplitL "".
          { rewrite /Q1 Eqst2. by iIntros "!> $". } iIntros "!> !>".
          iDestruct "INT" as (q') "(#H† & Hh & #Hshr & Hqq' & [Htok1 Htok2] & R)".
          iDestruct "Hqq'" as %Hqq'.
          iDestruct (GPS_SWSharedWriter_Reader_update with "W' R") as "[W' [R1 R2]]".
          iMod (rwown_update_read ν n q q' γs Hqq' with "g") as "[g rst]".
          iModIntro. iSplitL "Htok2 rst R1 Hβtok2 H†".
          + iFrame "Hβtok2". iExists _,_. rewrite Eqst2.
            by iFrame "Htok2 Hshr rst H† R1".
          + iExists (Some (inr (ν, (q + q' / 2)%Qp, (n + 1)%positive))).
            iSplitL ""; [done|]. iFrame "Share g W'". iExists _.
            iFrame "H† Hh Hshr Htok1 R2". by rewrite -Qp_plus_assoc Qp_div_2.
        - rewrite Eqst2. iDestruct "INT" as "(Hown & $ & R)".
          iExists (). iSplitL ""; [done|]. iIntros (t2 Ext2) "W' !>". iSplitL "".
          { rewrite /Q1 Eqst2. by iIntros "!> $". } iIntros "!> !>".
          iDestruct (GPS_SWSharedWriter_Reader_update with "W' R") as "[W' [R1 R2]]".
          iMod (lft_create with "LFT") as (ν) "[[Htok1 Htok2] #H†]". solve_ndisj.
          iMod (rwown_update_write_read ν γs with "g") as "[g rst]".
          iDestruct (lft_intersect_acc with "Hβtok2 Htok2") as (q) "[Htok Hclose]".
          iApply (fupd_mask_mono (↑lftN ∪ ↑histN)).
          { apply union_subseteq; split; solve_ndisj. }
          iMod (rebor _ _ (β ⊓ ν) with "LFT [] Hown") as "[Hst Hh]";
            [apply union_subseteq_l|iApply lft_intersect_incl_l|].
          iMod ("Share" with "[%] Hst Htok") as "[#Hshr Htok]". apply union_subseteq_l.
          iDestruct ("Hclose" with "Htok") as "[Hβtok2 Htok2]".
          iModIntro. iSplitL "Htok2 rst R1 Hβtok2".
          + iFrame "Hβtok2". iExists _,_. rewrite Eqst2.
            by iFrame "Htok2 Hshr rst H† R1".
          + iExists (Some (inr (ν, (1 / 2)%Qp, 1%positive))). iSplitL ""; [done|].
            iFrame "Share g W'". iExists _. iFrame "H† Hshr Htok1 R2".
            rewrite Qp_div_2. iSplitR ""; [|done].
            iIntros "Hν". iApply "Hh". rewrite -lft_dead_or. auto. }

      iNext. simpl.
      iIntros (b t2 [] v2) "(Hβtok1&_& [([%[%%]]&_& Hβtok2 & Hβ) | ([% _] &_&_& P)])".
      + subst b v2. iDestruct "Hβ" as (q' ν) "(Hν & Hshr & Hreading & #H† & R3)".
        wp_case.
        iMod ("Hclose'" with "[$]") as "Hα". iMod ("Hclose" with "Hα HL") as "HL".
        iApply (type_type  _ _ _
                   [ x ◁ box (&shr{α}(rwlock ty)); r ◁ box (uninit 2);
                     #lx ◁ rwlockreadguard α ty]
              with "[] LFT HE Hna HL Hk"); first last.
        { rewrite 2!tctx_interp_cons tctx_interp_singleton !tctx_hasty_val
                  tctx_hasty_val' //. iFrame "Hx Hr".
          iExists _, _, _, _. iFrame "Hν Hreading H† Hαβ".
          iDestruct ("EqS" with "Hshr") as "Hshr".
          iSplitL "Hshr".
          - iApply ty_shr_mono; last done.
            iApply lft_intersect_mono; first done. iApply lft_incl_refl.
          - iExists _,_,_. iSplitL "R3"; first by iExists _,_.
            iFrame "EqO EqS". by iExists _, _. }
        iApply (type_sum_assign (option $ rwlockreadguard α ty)); [solve_typing..|].
        simpl. iApply type_jump; solve_typing.
      + subst b.
        iDestruct (rel_objectively_elim with "P") as "Hβtok2".
        iDestruct (monPred_objectively_elim with "Hβtok2") as "Hβtok2".
        wp_case. iMod ("Hclose'" with "[$]") as "Hα".
        iMod ("Hclose" with "Hα HL") as "HL".
        iSpecialize ("Hk" with "[]"); first solve_typing.
        iApply ("Hk" $! [#] with "Hna HL").
        rewrite 2!tctx_interp_cons tctx_interp_singleton !tctx_hasty_val. iFrame.
        iExists _. iSplit. done. iExists _. iFrame "Hαβ".
        iExists _,_,_,_. iFrame "EqO EqS". by iExists _,_.
  Qed.

  (* Acquiring a write lock. *)
  Definition rwlock_try_write : val :=
    funrec: <> ["x"] :=
    withcont: "k":
      let: "r" := new [ #2 ] in
      let: "x'" := !"x" in
      if: CAS "x'" #0 #(-1) Relaxed AcqRel Relaxed then
        "r" <-{Σ some} "x'";;
        "k" ["r"]
      else
        "r" <-{Σ none} ();;
        "k" ["r"]
    cont: "k" ["r"] :=
      delete [ #1; "x"];; return: ["r"].

  Lemma rwlock_try_write_type ty `{!TyWf ty} :
    typed_val rwlock_try_write
        (fn(∀ α, ∅; &shr{α}(rwlock ty)) → option (rwlockwriteguard α ty)).
  Proof.
    intros E L. iApply type_fn; [solve_typing..|]. iIntros "/= !#".
      iIntros (α ϝ ret arg). inv_vec arg=>x. simpl_subst.
    iApply (type_cont [_] [ϝ ⊑ₗ []] (λ r, [x ◁ box (&shr{α}(rwlock ty));
                                        r!!!0 ◁ box (option (rwlockwriteguard α ty))]));
      [iIntros (k)|iIntros "/= !#"; iIntros (k arg); inv_vec arg=>r];
      simpl_subst; last first.
    { iApply type_delete; [solve_typing..|]. iApply type_jump; solve_typing. }
    iApply type_new; [solve_typing..|]. iIntros (r). simpl_subst.
    iApply type_deref; [solve_typing..|].
    iIntros (x' tid) "#LFT #HE Hna HL Hk HT". simpl_subst.
    rewrite 2!tctx_interp_cons tctx_interp_singleton !tctx_hasty_val.
    iDestruct "HT" as "(Hx & Hx' & Hr)". destruct x' as [[|lx|]|]; try done.
    iDestruct "Hx'" as (β) "#[Hαβ Hinv]".
    iDestruct "Hinv" as (γ γs tyO tyS) "(EqO & EqS & R)".
    iDestruct "R" as (tP vP) "R".
    iMod (lctx_lft_alive_tok α with "HE HL") as (qα) "(Hα & HL & Hclose)"; [solve_typing..|].
    iMod (lft_incl_acc with "Hαβ Hα") as (qβ) "[Hβtok Hclose']". done.
    wp_bind (CAS _ _ _ _ _ _).

    set Q : time → () → vProp Σ :=
      (λ t _, &{β}tyO ∗ rwown γs writing_st ∗
         GPS_SWSharedWriter lx t (() : unitProtocol) #(-1) γ ∗
         GPS_SWSharedReader lx t (() : unitProtocol) #(-1) 1 γ)%I.
    set Q1: time → () → vProp Σ :=
        (λ t _, rwlock_interp γs β tyO tyS false lx γ t () #0)%I.
    set Q2: time → () → vProp Σ :=
        (λ t _, rwlock_interp γs β tyO tyS true lx γ t () #0)%I.
    iMod (rel_True_intro True%I tid with "[//]") as "#rTrue".
    iApply (GPS_aSP_SharedWriter_CAS_int_weak rwlockN
              (rwlock_interp γs β tyO tyS) rwlock_noCAS _ β qβ _ _ _
              0 #(-1) tP () _ _ True%I Q Q1 Q2 (λ _ _ _, True)%I _ (⊤ ∖ ↑rwlockN)
              with "[$LFT $Hβtok $R rTrue]");
      [solve_ndisj|done..|by left|by right|by left|iSplitL ""|].
    { iIntros "!> !>" (?? v1 ?) "[F|[F|F]]".
        - iApply rwlock_interp_comparable. by iLeft.
        - iApply rwlock_interp_comparable. by iRight.
        - iDestruct "F" as %?. subst v1. iExists (-1). iSplit; [done|].
          iPureIntro. by constructor. }
    { simpl. iSplitL ""; last done.
      rewrite -(bi.True_sep' (∀ _ _, _)%I).
      iApply (rel_sep_objectively with "[$rTrue]"). 
      iIntros "!>" (t1 [] [_ Ext1]). iSplit; last first; last iSplitL ""; last iSplitL "".
      { iIntros "!>" (v ?) "!>". iSplit; last iSplit; by iIntros "$". }
      { rewrite /rwlock_noCAS. iIntros "!> _" (Eq). by inversion Eq. }
      { iIntros "!>". iDestruct 1 as (st2) "(>#Eq & #Share & F)".
        iIntros "!>". iSplitL ""; iNext; iExists st2; by iFrame "Eq Share". }
      iIntros "_". iDestruct 1 as (st2) "(>Eqst & #Share & g & INT)".
      iDestruct "Eqst" as %Eqst. inversion Eqst as [Eqst2]. clear Eqst.
      destruct st2 as [[[]|[[]]]|]; simpl in Eqst2; [done|done|..].
      iDestruct "INT" as "(Hown & $ & R)". iExists (). iSplitL ""; [done|].
      iIntros "!>" (t3 Ext3) "W' !>". iSplitL "".
      { rewrite /Q1. by iIntros "!> $". } iIntros "!> !>".
      iMod (rwown_update_write with "g") as "[g wst]". iModIntro.
      iFrame "Hown wst".
      iDestruct (GPS_SWSharedWriter_Reader_update with "W' R") as "[$ $]".
      iExists (Some (inl ())). iSplitL ""; [done|]. by iFrame "Share g". }

    iNext. simpl.
    iIntros (b t' [] v')
            "(Hβtok & _ &[([%[%%]] &_& Hβ & wst & W & R') | ([% _] & _)])".
    - subst b v'. iDestruct "R" as "[R AT]".
      iMod (GPS_aSP_SharedWriter_revert with "LFT Hβtok [$W $AT] R'") as "[Hβtok W]";
        [done..|].
      iMod ("Hclose'" with "Hβtok") as "Hα". iMod ("Hclose" with "Hα HL") as "HL".
      wp_case.
      iApply (type_type  _ _ _
                   [ x ◁ box (&shr{α}(rwlock ty)); r ◁ box (uninit 2);
                     #lx ◁ rwlockwriteguard α ty]
              with "[] LFT HE Hna HL Hk"); first last.
      { rewrite 2!tctx_interp_cons tctx_interp_singleton !tctx_hasty_val
                tctx_hasty_val' //. iFrame "Hx Hr".
        iExists γs, β. iSplitL "Hβ"; [by iApply (bor_iff with "[$EqO] Hβ")|].
        iFrame "Hαβ wst". iExists _,_,_. iSplitL "W".
        - iExists _. iDestruct "W" as "[$ _]".
        - iFrame "EqO EqS". iExists _,_. iFrame "R AT". }
      iApply (type_sum_assign (option $ rwlockwriteguard α ty)); [solve_typing..|].
      simpl. iApply type_jump; solve_typing.
    - subst b.
      iMod ("Hclose'" with "Hβtok") as "Hα". iMod ("Hclose" with "Hα HL") as "HL".
      wp_case.
      iApply (type_type _ _ _
              [ x ◁ box (&shr{α}(rwlock ty)); r ◁ box (uninit 2) ]
              with "[] LFT HE Hna HL Hk"); first last.
      { rewrite tctx_interp_cons tctx_interp_singleton !tctx_hasty_val. iFrame. }
      iApply (type_sum_unit (option $ rwlockwriteguard α ty));
        [solve_typing..|]; first last.
      rewrite /option /=. iApply type_jump; solve_typing.
  Qed.
End rwlock_functions.
