From iris.base_logic.lib Require Export iprop own.
From iris.algebra Require Export auth gmap agree frac.

From gpfsl.lang Require Export lang.
From gpfsl Require Import vprop.

Record protocolT : Type := {
  pr_stateT     :> Type;
  pr_inhab      : Inhabited pr_stateT;
  pr_eqdec      : EqDecision pr_stateT;
  pr_count      : Countable pr_stateT;
  pr_steps      : SqSubsetEq pr_stateT;
  pr_steps_pre  : PreOrder (⊑);
}.

Existing Instances pr_inhab pr_eqdec pr_count pr_steps pr_steps_pre.

Notation eventT Prtcl := (pr_stateT Prtcl * (val * view))%type.
Notation stateT Prtcl := (gmap time (eventT Prtcl))%type.
Notation interpC Σ Prtcl
  := (bool -c> loc -c> gname -c> time -c> pr_stateT Prtcl -c> val -c> vPropC Σ).
Notation interpCasC Σ Prtcl
  := (loc -c> gname -c> time -c> pr_stateT Prtcl -c> val -c> vPropC Σ).

Section CMRAs.
  Context {Prtcl : protocolT}.

  Definition stateMapUR : ucmraT := gmapUR time (agreeR $ leibnizC (eventT Prtcl)).
  Definition toStateMapUR : (stateT Prtcl) → stateMapUR :=
    fmap (@to_agree (leibnizC _)).

  (* cmra for protocol states *)
  Definition stateR : cmraT := authR $ authUR stateMapUR.
  (* cmra for the exclusive write *)
  Definition exWriteR : cmraT := authR $ optionUR (prodR fracR (agreeR $ leibnizC time)).
  (* cmra for the non-atomic write *)
  Definition naWriteR : cmraT := optionUR $ agreeR (leibnizC (view * time_ids)).
  (* the real one *)
  Definition gpsR : cmraT := prodR stateR (prodR exWriteR naWriteR).

  Lemma stateMap_absorb (ζ ζ': stateMapUR):
    ζ ⋅ ζ' ≡ ζ ↔ ζ' ≼ ζ.
  Proof.
    split; first by (move => <-; apply cmra_included_r).
    move => /lookup_included INCL i.
    rewrite lookup_op. move : (INCL i).
    destruct (ζ !! i) eqn:Hi; rewrite Hi; simpl;
       destruct (ζ' !! i) eqn:Hi'; rewrite Hi'; simpl;
       [|done|by rewrite option_included; naive_solver|done].
    rewrite Some_included_total -Some_op.
    move => /agree_included. rewrite cmra_comm. by move => <-.
  Qed.

  Lemma stateMap_idemp (ζ: stateMapUR):
    ζ ⋅ ζ ≡ ζ.
  Proof. by apply stateMap_absorb. Qed.

  Lemma stateMap_valid ζ: ✓ toStateMapUR ζ.
  Proof.
    intros i. rewrite /toStateMapUR lookup_fmap. by case (ζ !! i).
  Qed.

  Lemma stateMap_included ζ1 ζ2:
    ζ1 ⊆ ζ2 ↔ toStateMapUR ζ1 ≼ toStateMapUR ζ2.
  Proof.
    rewrite lookup_included /toStateMapUR.
    setoid_rewrite lookup_fmap.
    split; intros HI i; specialize (HI i).
    - destruct (ζ2 !! i) eqn:HE2; rewrite HE2; simpl;
        destruct (ζ1 !! i) eqn:HE1; rewrite HE1; simpl;
        [cbn in HI; by subst|apply option_included; by left|by inversion HI|done].
    - destruct (ζ2 !! i) eqn:HE2; simpl;
        destruct (ζ1 !! i) eqn:HE1; simpl; [|done| |done].
      + rewrite HE1 HE2 /= in HI.
        apply (Some_included_total (to_agree _)), to_agree_included in HI.
        by inversion HI.
      + rewrite HE1 HE2 /= in HI. apply option_included in HI. naive_solver.
  Qed.

  Lemma stateMap_eq ζ1 ζ2:
    toStateMapUR ζ1 ≡ toStateMapUR ζ2 → ζ1 = ζ2.
  Proof.
    move => Eq. apply map_eq => i. rewrite /toStateMapUR in Eq.
    move : Eq => /(_ i). rewrite 2!lookup_fmap.
    destruct (ζ1 !! i)  eqn:HE; rewrite HE /=;
      destruct (ζ2 !! i)  eqn:HE'; rewrite HE' /=;
      [|inversion 1|inversion 1|done].
    move => Eq. apply Some_equiv_inj, to_agree_inj in Eq. by inversion Eq.
  Qed.

  Lemma stateMap_local_update (ζ ζ' ζ0: stateMapUR)
    (INCL: ζ ≼ ζ') (VALID: ✓ ζ'):
    (ζ, ζ0) ~l~> (ζ', ζ').
  Proof.
    apply local_update_unital_discrete => z Val Eq. split; [done|].
    symmetry. apply stateMap_absorb. etrans; last exact INCL.
    rewrite Eq. apply cmra_included_r.
  Qed.

End CMRAs.

Arguments gpsR _ : clear implicits.

Class gpsG Σ Prtcl := { gps_inG :> inG Σ (gpsR Prtcl); }.

Definition gpsΣ Prtcl : gFunctors := #[ GFunctor (constRF (gpsR Prtcl)) ].
