From iris.algebra Require Import gmap gset excl auth frac agree.
From iris.base_logic.lib Require Import own.
From gpfsl.lang Require Import lang lattice_cmra.

Set Default Proof Using "Type".

Canonical Structure cellC := (leibnizC cell).

(* Cmra for histories *)
(* optionC is used to denote deallocation:
   if the location is deallocated, then it is None, otherwise it is (Some C) *)
Definition histUR : ucmraT
  := gmapUR loc (prodR (agreeR (optionC cellC)) fracR).

(* Cmra to manage block deallocation *)
Definition hist_freeableUR : ucmraT :=
  gmapUR block (prodR fracR (gmapR Z (exclR unitC))).

(* Cmra to manage racing reads *)
Definition hist_readUR : ucmraT := gmapUR loc (fracLatR (gset_Lat time_id)).
Definition hist_writesUR : ucmraT := gmapUR loc (prodR fracR (agreeR (gsetC time_id))).

(* Cmra to manage views *)
Definition viewR := authR (latUR view_Lat).

(* All those Cmra structures we need *)
Class histG Σ := HistG {
  hist_inG :> inG Σ (authR histUR);
  hist_freeable_inG :> inG Σ (authR hist_freeableUR);
  hist_readG :> inG Σ (authR hist_readUR);
  hist_na_view_inG :> inG Σ viewR;
  hist_at_write_inG :> inG Σ (authR hist_writesUR);
  hist_tview_inG :> inG Σ (authR (latUR tview_Lat));
  hist_physG :> inG Σ (authR (optionUR (exclR stateC)));

  hist_name : gname;
  hist_freeable_name : gname;
  hist_atwrite_name : gname;
  hist_atread_name : gname;
  hist_naread_name : gname;
  hist_gtime_name : gname;
  hist_phys_name : gname;
}.

(** From base structure to CMRA *)
Definition to_hist (M: memory) : histUR :=
  fmap (λ C : cell,
        (to_agree (if (decide (cell_deallocated C)) then None else Some C), 1%Qp))
       (gmap_uncurry M).

Definition to_atw (𝓝: view) : hist_writesUR :=
  fmap (λ ts, (1%Qp, to_agree ts.(tawrite))) 𝓝.

Definition to_nar (𝓝: view) : hist_readUR :=
  fmap (λ ts, (1%Qp, to_latT ts.(tnread))) 𝓝.

Definition to_atr (𝓝: view) : hist_readUR :=
  fmap (λ ts, (1%Qp, to_latT ts.(taread))) 𝓝.

(* to_hist cut *)
Lemma memory_uncurry_insert_nonempty (M: memory) (l: loc) (C : cell) (NE: C ≠ ∅) :
  gmap_uncurry (<[l:=C]> M) = <[l := C]> (gmap_uncurry M).
Proof.
  rewrite {1}/insert /memory_cell_insert. apply map_eq=>l'.
  generalize (lookup_gmap_uncurry M l'),
     (lookup_gmap_uncurry (gmap_curry (<[l:=C]> (gmap_uncurry M))) l').
  setoid_rewrite lookup_gmap_curry. destruct (decide (l = l')) as [<-|].
  - rewrite !lookup_insert=>_ /=. destruct lookup=>/=.
    + move=>?. f_equal. by apply map_eq.
    + apply map_choose in NE as (t & ? & EQ)=>/(_ t). by rewrite EQ.
  - generalize (lookup_gmap_uncurry_None M l'),
       (lookup_gmap_uncurry_None (gmap_curry (<[l:=C]> (gmap_uncurry M))) l').
    setoid_rewrite lookup_gmap_curry. rewrite lookup_insert_ne //.
    (do 2 case:(_ !! l'))=>[??|?|?|] //= HN1 HN2 HS1 HS2.
    + f_equal. apply map_eq, HS2.
    + setoid_rewrite <-HS1 in HN1. naive_solver.
    + naive_solver.
Qed.

Lemma to_hist_insert_alloc (M: memory) (l: loc) (C : cell)
  (ALLOC: ¬ cell_deallocated C) (NE: C ≠ ∅):
  to_hist (<[l:=C]> M) = <[l:= (to_agree (Some C), 1%Qp)]> (to_hist M).
Proof.
  rewrite /to_hist {1}/insert /memory_cell_insert.
  rewrite (_: Some C = if (decide (cell_deallocated C)) then None else Some C);
    last by rewrite decide_False.
  rewrite -fmap_insert. f_equal. by apply memory_uncurry_insert_nonempty.
Qed.

Lemma to_hist_insert_dealloc (M: memory) (l: loc) (C : cell)
  (DEALLOC: cell_deallocated C) :
  to_hist (<[l:=C]> M) = <[l:= (to_agree None, 1%Qp)]> (to_hist M).
Proof.
  rewrite /to_hist {1}/insert /memory_cell_insert.
  rewrite {2}(_: None = if (decide (cell_deallocated C)) then None else Some C);
    last by rewrite decide_True.
  rewrite -fmap_insert. f_equal. apply memory_uncurry_insert_nonempty.
  move => EM. rewrite EM in DEALLOC. apply DEALLOC.
Qed.

Lemma to_hist_lookup_None M l:
  M !!c l = ∅ → to_hist M !! l = None.
Proof.
  rewrite lookup_fmap /memory_cell_lookup.
  destruct (gmap_uncurry M !! l) as [?|] eqn:Eq; [|done].
  move => ?. exfalso. by apply (gmap_uncurry_non_empty _ _ _ Eq).
Qed.

Lemma to_hist_lookup_Some M C l q:
  (to_hist M) !! l ≡ Some (A:=ofe_car _) (to_agree (Some C), q) →
  M !!c l = C ∧ ¬ cell_deallocated C ∧ C ≠ ∅.
Proof.
  rewrite lookup_fmap /=.
  destruct (gmap_uncurry M !! l) eqn:Eq; last by inversion 1.
  intros [HC%(inj to_agree) _]%(inj Some)%(inj2 pair).
  case_match; inversion HC; setoid_subst.
  rewrite /memory_cell_lookup Eq. repeat split; [done|].
  by apply (gmap_uncurry_non_empty _ _ _ Eq).
Qed.

(* to_atw *)
Lemma to_atw_insert 𝓝 l t ws rsa rsn :
  to_atw (<[l:=[{ t,ws,rsn,rsa }]]> 𝓝) = <[l:= (1%Qp, to_agree ws)]> (to_atw 𝓝).
Proof. by rewrite /to_atw fmap_insert. Qed.

Lemma to_atw_set_write_time 𝓝 l t :
  to_atw (set_write_time 𝓝 l t) ≡ to_atw 𝓝.
Proof.
  move => l'. rewrite 2!lookup_fmap.
  case (decide (l' = l)) => [->|?].
  - rewrite lookup_partial_alter /=.
    by rewrite -2!lookup_fmap -map_fmap_compose lookup_fmap.
  - by rewrite lookup_partial_alter_ne.
Qed.

Lemma to_atw_add_aread_id 𝓝 l t :
  to_atw (add_aread_id 𝓝 l t) ≡ (to_atw 𝓝).
Proof.
  move => l'. rewrite 2!lookup_fmap.
  case (decide (l' = l)) => [->|?].
  - rewrite lookup_partial_alter /=.
    by rewrite -2!lookup_fmap -map_fmap_compose lookup_fmap.
  - by rewrite lookup_partial_alter_ne.
Qed.

Lemma to_atw_add_nread_id 𝓝 l t :
  to_atw (add_nread_id 𝓝 l t) ≡ (to_atw 𝓝).
Proof.
  move => l'. rewrite 2!lookup_fmap.
  case (decide (l' = l)) => [->|?].
  - rewrite lookup_partial_alter /=.
    by rewrite -2!lookup_fmap -map_fmap_compose lookup_fmap.
  - by rewrite lookup_partial_alter_ne.
Qed.

Lemma to_atw_add_awrite_id 𝓝 l t ws :
  𝓝 !!aw l = Some ws →
  to_atw (add_awrite_id 𝓝 l t) ≡ <[l := (1%Qp, to_agree (ws ∪ {[t]}))]> (to_atw 𝓝).
Proof.
  move => Eql l'. rewrite lookup_fmap.
  case (decide (l' = l)) => [->|?].
  - rewrite lookup_insert lookup_partial_alter /=.
    rewrite -2!lookup_fmap -map_fmap_compose lookup_fmap.
    destruct (view_lookup_of_awp _ _ _ Eql) as [p [Eq1 Eq2]].
    by rewrite Eq2 /= Eq1 union_comm_L.
  - rewrite lookup_insert_ne; [|done].
    rewrite lookup_partial_alter_ne; [|done]. by rewrite lookup_fmap.
Qed.

Lemma to_atw_lookup_None 𝓝 l:
  𝓝 !! l = None ↔ to_atw 𝓝 !! l = None.
Proof.
  rewrite lookup_fmap. split. move => -> //.
  destruct (𝓝 !! l) as [[]|] eqn:Eql; [|done]. by rewrite Eql.
Qed.

Lemma to_atw_lookup_Some 𝓝 l p:
  𝓝 !! l = Some p → to_atw 𝓝 !! l = Some (1%Qp, to_agree p.(tawrite)).
Proof. rewrite lookup_fmap => -> //. Qed.

Lemma to_atw_lookup_r_Some 𝓝 l rs:
  𝓝 !!aw l = Some rs → to_atw 𝓝 !! l = Some (1%Qp, to_agree rs).
Proof.
  move => Eql. destruct (view_lookup_of_awp _ _ _ Eql) as [p [Eq1 Eq2]].
  rewrite -Eq1. by apply to_atw_lookup_Some.
Qed.

Lemma to_atw_valid 𝓝: ✓ to_atw 𝓝.
Proof. intros l. rewrite lookup_fmap. case (_ !! l) => //. Qed.

(* to_nar *)
Lemma to_nar_insert 𝓝 l t ws rsa rsn :
  to_nar (<[l:=[{ t,ws,rsn,rsa }]]> 𝓝) = <[l:= (1%Qp, to_latT rsn)]> (to_nar 𝓝).
Proof.
  apply leibniz_equiv => l'. rewrite lookup_fmap.
  case (decide (l' = l)) => [->|?].
  - by rewrite 2!lookup_insert /=.
  - do 2 (rewrite lookup_insert_ne; [|done]). by rewrite lookup_fmap.
Qed.

Lemma to_nar_set_write_time 𝓝 l t :
  to_nar (set_write_time 𝓝 l t) = to_nar 𝓝.
Proof.
  apply leibniz_equiv => l'. rewrite 2!lookup_fmap.
  case (decide (l' = l)) => [->|?].
  - rewrite lookup_partial_alter /=.
    by rewrite -2!lookup_fmap -map_fmap_compose lookup_fmap.
  - by rewrite lookup_partial_alter_ne.
Qed.

Lemma to_nar_add_awrite_id 𝓝 l t :
  to_nar (add_awrite_id 𝓝 l t) = to_nar 𝓝.
Proof.
  apply leibniz_equiv => l'. rewrite 2!lookup_fmap.
  case (decide (l' = l)) => [->|?].
  - rewrite lookup_partial_alter /=.
    by rewrite -2!lookup_fmap -map_fmap_compose lookup_fmap.
  - by rewrite lookup_partial_alter_ne.
Qed.

Lemma to_nar_add_aread_id 𝓝 l t :
  to_nar (add_aread_id 𝓝 l t) = to_nar 𝓝.
Proof.
  apply leibniz_equiv => l'.
  rewrite 2!lookup_fmap.
  case (decide (l' = l)) => [->|?].
  - rewrite lookup_partial_alter /=.
    by rewrite -2!lookup_fmap -map_fmap_compose lookup_fmap.
  - by rewrite lookup_partial_alter_ne.
Qed.

Lemma to_nar_add_nread_id 𝓝 l t rs :
  𝓝 !!nr l = Some rs →
  to_nar (add_nread_id 𝓝 l t) = <[l := (1%Qp, to_latT (rs ∪ {[t]}))]> (to_nar 𝓝).
Proof.
  move => Eql. apply leibniz_equiv => l'.
  rewrite lookup_fmap.
  case (decide (l' = l)) => [->|?].
  - rewrite lookup_insert lookup_partial_alter /=.
    rewrite -2!lookup_fmap -map_fmap_compose lookup_fmap.
    destruct (view_lookup_of_nrp _ _ _ Eql) as [p [Eq1 Eq2]].
    by rewrite Eq2 /= Eq1 union_comm_L.
  - rewrite lookup_insert_ne; [|done].
    rewrite lookup_partial_alter_ne; [|done]. by rewrite lookup_fmap.
Qed.

Lemma to_nar_lookup_None 𝓝 l:
  𝓝 !! l = None ↔ to_nar 𝓝 !! l = None.
Proof.
  rewrite lookup_fmap. split. move => -> //.
  destruct (𝓝 !! l) as [[]|] eqn:Eql; [|done]. by rewrite Eql.
Qed.

Lemma to_nar_lookup_Some 𝓝 l p:
  𝓝 !! l = Some p → to_nar 𝓝 !! l = Some (1%Qp, to_latT p.(tnread)).
Proof. rewrite lookup_fmap => -> //. Qed.

Lemma to_nar_lookup_r_Some 𝓝 l rs:
  𝓝 !!nr l = Some rs → to_nar 𝓝 !! l = Some (1%Qp, to_latT rs).
Proof.
  move => Eql. destruct (view_lookup_of_nrp _ _ _ Eql) as [p [Eq1 Eq2]].
  rewrite -Eq1. by apply to_nar_lookup_Some.
Qed.

Lemma to_nar_valid 𝓝: ✓ to_nar 𝓝.
Proof. intros l. rewrite lookup_fmap. case (_ !! l) => //. Qed.

(* to_atr *)
Lemma to_atr_insert 𝓝 l t ws rsa rsn :
  to_atr (<[l:=[{ t,ws,rsn,rsa }]]> 𝓝) = <[l:= (1%Qp, to_latT rsa)]> (to_atr 𝓝).
Proof.
  apply leibniz_equiv => l'. rewrite lookup_fmap.
  case (decide (l' = l)) => [->|?].
  - by rewrite 2!lookup_insert /=.
  - do 2 (rewrite lookup_insert_ne; [|done]). by rewrite lookup_fmap.
Qed.

Lemma to_atr_set_write_time 𝓝 l t :
  to_atr (set_write_time 𝓝 l t) = to_atr 𝓝.
Proof.
  apply leibniz_equiv => l'. rewrite 2!lookup_fmap.
  case (decide (l' = l)) => [->|?].
  - rewrite lookup_partial_alter /=.
    by rewrite -2!lookup_fmap -map_fmap_compose lookup_fmap.
  - by rewrite lookup_partial_alter_ne.
Qed.

Lemma to_atr_add_nread_id 𝓝 l t :
  to_atr (add_nread_id 𝓝 l t) = to_atr 𝓝.
Proof.
  apply leibniz_equiv => l'. rewrite 2!lookup_fmap.
  case (decide (l' = l)) => [->|?].
  - rewrite lookup_partial_alter /=.
    by rewrite -2!lookup_fmap -map_fmap_compose lookup_fmap.
  - by rewrite lookup_partial_alter_ne.
Qed.

Lemma to_atr_add_awrite_id 𝓝 l t :
  to_atr (add_awrite_id 𝓝 l t) = to_atr 𝓝.
Proof.
  apply leibniz_equiv => l'. rewrite 2!lookup_fmap.
  case (decide (l' = l)) => [->|?].
  - rewrite lookup_partial_alter /=.
    by rewrite -2!lookup_fmap -map_fmap_compose lookup_fmap.
  - by rewrite lookup_partial_alter_ne.
Qed.

Lemma to_atr_add_aread_id 𝓝 l t rs :
  𝓝 !!ar l = Some rs →
  to_atr (add_aread_id 𝓝 l t) = <[l := (1%Qp, to_latT (rs ∪ {[t]}))]> (to_atr 𝓝).
Proof.
  move => Eql. apply leibniz_equiv => l'. rewrite lookup_fmap.
  case (decide (l' = l)) => [->|?].
  - rewrite lookup_insert lookup_partial_alter /=.
    rewrite -2!lookup_fmap -map_fmap_compose lookup_fmap.
    destruct (view_lookup_of_arp _ _ _ Eql) as [p [Eq1 Eq2]].
    by rewrite Eq2 /= Eq1 union_comm_L.
  - rewrite lookup_insert_ne; [|done].
    rewrite lookup_partial_alter_ne; [|done]. by rewrite lookup_fmap.
Qed.

Lemma to_atr_lookup_None 𝓝 l:
  𝓝 !! l = None ↔ to_atr 𝓝 !! l = None.
Proof.
  rewrite lookup_fmap. split. move => -> //.
  destruct (𝓝 !! l) as [[]|] eqn:Eql; [|done]. by rewrite Eql.
Qed.

Lemma to_atr_lookup_Some 𝓝 l p:
  𝓝 !! l = Some p → to_atr 𝓝 !! l = Some (1%Qp, to_latT p.(taread)).
Proof. rewrite lookup_fmap => -> //. Qed.

Lemma to_atr_lookup_r_Some 𝓝 l rs:
  𝓝 !!ar l = Some rs → to_atr 𝓝 !! l = Some (1%Qp, to_latT rs).
Proof.
  move => Eql. destruct (view_lookup_of_arp _ _ _ Eql) as [p [Eq1 Eq2]].
  rewrite -Eq1. by apply to_atr_lookup_Some.
Qed.

Lemma to_atr_valid 𝓝: ✓ to_atr 𝓝.
Proof. intros l. rewrite lookup_fmap. case (_ !! l) => //. Qed.

Lemma add_aread_awrite_comm 𝓝 l tr tw:
  add_awrite_id (add_aread_id 𝓝 l tr) l tw =
  add_aread_id (add_awrite_id 𝓝 l tw) l tr.
Proof.
  apply leibniz_equiv => l'.
  case (decide (l' = l)) => [->|?].
  - rewrite 4!lookup_partial_alter /=.
    by rewrite -4!lookup_fmap -2!map_fmap_compose.
  - by do 4 (rewrite lookup_partial_alter_ne; [|done]).
Qed.
