From iris.algebra Require Import gmap coPset auth functions.
From iris.base_logic.lib Require Import invariants.
From iris.proofmode Require Import tactics.
From gpfsl Require Export vprop.
From gpfsl.lang Require Import lattice_cmra.
Set Default Proof Using "Type".

(* Non-atomic ("thread-local") invariants. *)

Definition na_inv_pool_name := gname.

Class na_invG View `{!LatBottom bot} Σ :=
  { na_inv_enable_inG :> inG Σ (@ofe_funR positive (λ _, authUR (latUR View)));
    na_inv_disable_inG :> inG Σ (authR (optionUR (exclR (latC View)))) }.
Definition na_invΣ View `{!LatBottom bot} : gFunctors :=
  #[ GFunctor (constRF (@ofe_funR positive (λ _, authUR (latUR View))));
     GFunctor (constRF (authR (optionUR (exclR (latC View))))) ].
Instance subG_na_invG View `{!LatBottom bot} Σ :
  subG (na_invΣ View) Σ → na_invG View Σ.
Proof. solve_inG. Qed.

Section defs.
  Context `{invG Σ, !@na_invG View bot BOT Σ}.
  Notation vProp := (monPred (lat_bi_index View) (uPredI (iResUR Σ))).

  Definition to_ofe_funR (E : coPset) (Vs : positive → View) : ofe_funR _ :=
    λ i, if decide (i ∈ E) then ● to_latT (Vs i) else ε.

  Definition na_own (p : na_inv_pool_name) (E : coPset) : vProp :=
    (∃ Vs, ⎡own p (to_ofe_funR E Vs)⎤ ∗ ∀ i, monPred_in (Vs i))%I.

  Definition na_inv (p : na_inv_pool_name) (N : namespace) (P : vProp) : vProp :=
    (∃ i V γ, ⌜i ∈ (↑N:coPset)⌝ ∗ monPred_in V ∗
            ⎡inv N ((∃ V', P (V ⊔ V') ∗
                           own p (ofe_fun_singleton i (◯ to_latT V')) ∗
                           own γ (● None)) ∨
                    (∃ V', own p (ofe_fun_singleton i (● to_latT V')) ∗
                           own γ (● Excl' (to_latT V'))))⎤)%I.
End defs.

Instance: Params (@na_inv) 8.
Instance: Params (@na_own) 6.
Typeclasses Opaque na_own na_inv.

Section proofs.
  Context `{invG Σ, !@na_invG View bot BOT Σ}.
  Notation vProp := (monPred (lat_bi_index View) (uPredI (iResUR Σ))).

  Global Instance na_own_timeless p E : Timeless (na_own p E).
  Proof. rewrite /na_own. apply _. Qed.

  Global Instance na_inv_ne p N : NonExpansive (na_inv p N).
  Proof. rewrite /na_inv. solve_proper. Qed.
  Global Instance na_inv_proper p N : Proper ((≡) ==> (≡)) (na_inv p N).
  Proof. apply (ne_proper _). Qed.

  Global Instance na_inv_persistent p N (P : vProp) : Persistent (na_inv p N P).
  Proof. rewrite /na_inv; apply _. Qed.

  Lemma na_alloc : (|==> ∃ p, na_own p ⊤)%I.
  Proof.
    rewrite /na_own. iDestruct (monPred_in_intro True%I with "[//]") as (V) "[HV _]".
    iMod (own_alloc (to_ofe_funR ⊤ (λ _, V))) as (p) "?".
    { intro i. split; [|done]. intros n. apply ucmra_unit_leastN. }
    iExists _, _. by iFrame.
  Qed.

  Lemma na_own_disjoint p E1 E2 : na_own p E1 -∗ na_own p E2 -∗ ⌜E1 ## E2⌝.
  Proof.
    rewrite /na_own. iIntros "H1 H2".
    iDestruct "H1" as (Vs1) "[H1 _]". iDestruct "H2" as (Vs2) "[H2 _]".
    iDestruct (own_valid_2 with "H1 H2") as %Hv. iIntros "!%" (i Hi1 Hi2).
    specialize (Hv i). by rewrite /to_ofe_funR /= ofe_fun_lookup_op !decide_True in Hv.
  Qed.

  Lemma na_own_union p E1 E2 :
    E1 ## E2 → na_own p (E1 ∪ E2) ⊣⊢ na_own p E1 ∗ na_own p E2.
  Proof.
    intros DISJ. rewrite /na_own. iSplit.
    - iIntros "H". iDestruct "H" as (Vs) "[Hown #Hin]".
      rewrite -!(bi.exist_intro Vs). iFrame "Hin".
      rewrite -embed_sep -own_op 
        (_:to_ofe_funR (E1 ∪ E2) Vs ≡ to_ofe_funR E1 Vs ⋅ to_ofe_funR E2 Vs) //.
      intros i. rewrite ofe_fun_lookup_op /to_ofe_funR. specialize (DISJ i).
      do 3 destruct decide; try set_solver; by rewrite ?right_id ?left_id.
    - iIntros "[H1 H2]".
      iDestruct "H1" as (Vs1) "[H1 #Hin1]". iDestruct "H2" as (Vs2) "[H2 #Hin2]".
      iExists (λ i, if decide (i ∈ E1) then Vs1 i else Vs2 i). iSplitL.
      + iCombine "H1 H2" as "H".
        rewrite (_:to_ofe_funR E1 Vs1 ⋅ to_ofe_funR E2 Vs2 ≡ _) //.
        intros i. rewrite ofe_fun_lookup_op /to_ofe_funR. specialize (DISJ i).
        do 3 destruct decide; try set_solver; by rewrite ?right_id ?left_id.
      + iIntros (i). destruct decide; [iApply "Hin1"|iApply "Hin2"].
  Qed.

  Lemma na_own_acc E2 E1 tid :
    E2 ⊆ E1 → na_own tid E1 -∗ na_own tid E2 ∗ (na_own tid E2 -∗ na_own tid E1).
  Proof.
    intros HF. assert (E1 = E2 ∪ (E1 ∖ E2)) as -> by exact: union_difference_L.
    rewrite na_own_union; last by set_solver+. iIntros "[$ $]". auto.
  Qed.

  Lemma na_inv_alloc p E N (P : vProp) : ▷ P ={E}=∗ na_inv p N P.
  Proof.
    iIntros "HP".
    iDestruct (monPred_in_intro with "HP") as (V) "[#HV HP]".
    iMod (own_unit (@ofe_funUR positive (λ _, authUR (latUR View))) p) as "Hempty".
    iMod (own_alloc (● None)) as (γ) "Hdis"; [done|].
    destruct (fresh_inv_name ∅ N) as (i & _ & HiN). unfold na_inv.
    iExists i, V, γ. iFrame "# %". iMod (inv_alloc N with "[-]") as "$"; [|done].
    iLeft. iExists bot. iNext. iFrame. rewrite (_:ε ≡ _) //. intros i'.
    rewrite /ofe_fun_singleton /ofe_fun_insert. destruct decide=>//. by subst i.
  Qed.

  Lemma na_inv_open p E F N (P : vProp) :
    ↑N ⊆ E → ↑N ⊆ F →
    na_inv p N P -∗ na_own p F ={E}=∗ ▷ P ∗ na_own p (F∖↑N) ∗
                       (▷ P ∗ na_own p (F∖↑N) ={E}=∗ na_own p F).
  Proof.
    rewrite /na_inv. iIntros (??) "#Hnainv Htoks".
    iDestruct "Hnainv" as (i V γ) "(% & HinV & Hinv)".
    rewrite [F as X in na_own p X](union_difference_L (↑N) F) //.
    rewrite [X in (X ∪ _)](union_difference_L {[i]} (↑N)) ?na_own_union; [|set_solver..].
    iDestruct "Htoks" as "[[Htoki $] $]". rewrite [_ p {[i]}]/na_own.
    iDestruct "Htoki" as (Vs) "[Hown #HinVs]".
    iInv N as "[HP|>Htoki]" "Hclose"; last first.
    { iDestruct "Htoki" as (V') "[Hown' _]".
      iDestruct (own_valid_2 with "Hown Hown'") as %Hv. specialize (Hv i).
      rewrite /= ofe_fun_lookup_op ofe_fun_lookup_singleton /to_ofe_funR
        decide_True // in Hv. set_solver. }
    iDestruct "HP" as (V') "(HP & >#HV' & >Hdis)".
    iDestruct (own_valid_2 with "Hown HV'") as %Hv. specialize (Hv i).
    rewrite /= ofe_fun_lookup_op ofe_fun_lookup_singleton /to_ofe_funR
      decide_True in Hv; [|set_solver].
    apply auth_valid_discrete_2, proj1, latT_included in Hv. simpl in Hv. rewrite [in P _]Hv.
    iSpecialize ("HinVs" $! i). iCombine "HinV HinVs" as "Hin".
    iDestruct (monPred_in_intro with "Hin") as (V0) "[HV0 HinV0]".
    rewrite monPred_at_sep. iDestruct "HinV0" as %[??].
    rewrite (lat_join_lub V (Vs i) V0) // -monPred_at_later.
    iDestruct (monPred_in_elim (▷ P)%I with "HV0 HP") as "$".
    iMod (own_update with "Hdis") as "[Hdis1 Hdis2]".
    { by apply auth_update_alloc, (alloc_option_local_update (Excl (to_latT (Vs i)))). }
    iMod ("Hclose" with "[Hown Hdis1]") as "_".
    { iNext. iRight. iExists (Vs i). rewrite (_:to_ofe_funR {[i]} Vs ≡ _); [by iFrame|].
      intros i'. rewrite /to_ofe_funR /ofe_fun_singleton /ofe_fun_insert.
      do 2 destruct decide; set_solver. }
    iIntros "!> [HP $]".
    iInv N as "[HP'|>Htoki]" "Hclose".
    { iDestruct "HP'" as (?) "(_ & _ & >Hdis1)".
      iDestruct (own_valid_2 with "Hdis1 Hdis2") as %[Hv' _]. exfalso.
      move:(Hv' 0%nat). rewrite /= left_id option_includedN. naive_solver. }
    iDestruct "Htoki" as (Vi) "[Hown Hdis1]".
    iDestruct (own_valid_2 with "Hdis1 Hdis2") as %[Hv' _]. move:(Hv' 0%nat).
    rewrite /= left_id=>/Excl_includedN /discrete /(inj to_latT)=>EQVi.
    (* FIXME : rewrite EQVi should work. *)
    assert (@ofe_fun_singleton _ _ (λ _, _) i (● to_latT Vi) ≡
            ofe_fun_singleton i (● to_latT (Vs i))) as -> by by f_equiv; rewrite EQVi.
    iMod (own_update_2 with "Hdis1 Hdis2") as "Hdis".
    { apply auth_update_dealloc, delete_option_local_update, _. }
    iCombine "HP Hin" as "H".
    iDestruct (monPred_in_intro with "H") as (V'') "(#HV'' & HP & % & %)".
    iMod (own_update with "Hown") as "[Hown1 Hown2]".
    { etrans.
      - by apply ofe_fun_singleton_update, auth_update_alloc,
          (op_local_update_discrete _ _ (to_latT V'')).
      - by rewrite -ofe_fun_op_singleton. }
    iMod ("Hclose" with "[-Hown1]") as "_".
    - iLeft. iExists V''. iFrame. iNext.
      (* FIXME : idem *)
      assert (@ofe_fun_singleton _ _ (λ _, _) i (◯ (to_latT V'' ⋅ ε)) ≡
              ofe_fun_singleton i (◯ (to_latT V''))) as ->; [|done].
      f_equiv. by rewrite (right_id ε).
    - iExists (λ _, V''). iSplitL; last by auto.
      assert (ofe_fun_singleton i (● (to_latT V'' ⋅ to_latT (Vs i))) ≡
              to_ofe_funR {[i]} (λ _ : positive, V'')) as ->; [|done].
      intros i'. unfold ofe_fun_singleton, ofe_fun_insert, to_ofe_funR.
      destruct (decide (i = i')) as [<-|]=>/=; last first.
      { rewrite decide_False //. set_solver. }
      rewrite decide_True; [|set_solver]. rewrite lat_op_join'.
      do 4 f_equiv. solve_lat.
  Qed.
End proofs.
