From Coq.QArith Require Import Qcanon.

From iris.base_logic.lib Require Export iprop.
From iris.algebra Require Export auth gmap agree frac.
From iris.bi Require Import fractional.
From iris.proofmode Require Import tactics.

From gpfsl.lang Require Import history base_lifting cbends.
From gpfsl.gps Require Import gps_cmra.
From gpfsl Require Import vprop.

Set Default Proof Using "Type".

Definition isInvView l (rs ws : time_ids) (C: cell) Vo :=
  atr_local l rs Vo ∧ atw_local l ws Vo ∧
  ∀ t m, C !! t = Some m → seen_local l t Vo ∧ m.(mrel) ⊑ Some Vo.

Lemma inv_view_insert_hist l ws rs C Vo t' m'
  (FRESH: C !! t' = None) (JOIN: isInvView l rs ws C Vo) :
  isInvView l rs (ws ∪ {[t']}) (<[t':=m']> C)
            (Vo ⊔ {[l := [{ t',{[t']},∅,∅ }] ]} ⊔ default ∅ m'.(mrel)).
Proof.
  destruct JOIN as [AT [AW JOIN]]. split; last split.
  { eapply atr_local_mono; [done|..|done]. solve_lat. }
  { eapply atw_local_join.
    - eapply atw_local_mono; [done|..|done]. solve_lat.
    - eapply atw_local_mono; [done|..]; last first.
      + rewrite /atw_local. instantiate (1:={[l := [{ t',{[t']},∅,∅ }]]}).
        by rewrite view_lookup_aw_insert /=.
      + solve_lat. }
  move => t m.
  case (decide (t = t')) => [->|?];
    [rewrite lookup_insert => [[->]] /=
    |rewrite lookup_insert_ne => Eqt; last done]; split.
  - rewrite /seen_local 2!view_lookup_w_join /= view_lookup_w_insert /=.
    solve_lat.
  - destruct m.(mrel); solve_lat.
  - rewrite /seen_local. etrans; first apply (JOIN _ _ Eqt).
    rewrite 2!view_lookup_w_join. solve_lat.
  - etrans; first apply (JOIN _ _ Eqt). solve_lat.
Qed.

Instance inv_view_mono l rs ws C: Proper ((⊑) ==> (impl)) (isInvView l rs ws C).
Proof.
  move => ??? [? [? JV]]. split; last split.
  by eapply atr_local_mono. by eapply atw_local_mono.
  move => t m /JV [? ?]. split; [by eapply seen_local_mono|solve_lat].
Qed.

Lemma inv_view_singleton_hist l rs ws t m V:
  atr_local l rs V → atw_local l ws V →
  isInvView l rs ws {[t := m]}
            ((default ∅ m.(mrel) ⊔ {[l := [{t,∅,∅,∅}]]}) ⊔ V).
Proof.
  move => AT AW. split; last split.
  { eapply atr_local_mono; [done|..|done]. solve_lat. }
  { eapply atw_local_mono; [done|..|done]. solve_lat. }
  move => t' m' /lookup_singleton_Some [<- <-]. split.
  - rewrite /seen_local 2!view_lookup_w_join /= view_lookup_w_insert.
    solve_lat.
  - destruct m.(mrel); solve_lat.
Qed.

Section Definitions.
  Context `{!noprolG Σ, !gpsG Σ Prtcl}
          (IP : interpC Σ Prtcl) (IW: interpCasC Σ Prtcl)
          (l : loc).
  Local Notation iProp := (iProp Σ).
  Local Notation vProp := (vProp Σ).
  Implicit Types
           (s : pr_stateT Prtcl) (ζ : stateT Prtcl) (e : eventT Prtcl)
           (t : time) (v : val) (V : view) (C : cell)
           (μ : time → pr_stateT Prtcl).

  Notation F := (IP true l).        (* full interp *)
  Notation F_read := (IP false l).  (* read interp *)

  (* μ assigns write events (identified by timestamps) to protocol states *)
  (* μ must be monotone for write events in C.
     The write events need to be restricted further to true values only. *)
  Definition isState C t : Prop := ∃ m, C !! t = Some m ∧ m.(mval) ≠ DVal.
  Definition fsorted μ C : Prop :=
    ∀ (t1 t2 : time),
      isState C t1 → isState C t2 → (t1 ≤ t2)%positive → μ t1 ⊑ μ t2.

  Definition stsorted (ζ: stateT Prtcl) :=
    ∀ (t1 t2 : time) e1 e2,
      ζ !! t1 = Some e1 → ζ !! t2 = Some e2 → (t1 ≤ t2)%positive → e1.1 ⊑ e2.1.

  (* full resources are available at BEs (block ends)
    that are after the exwr (exclusive write) identified by tx,
    and that are of true values (VVal v) *)
  Definition gps_res (π: time → val → view → iProp) C Va : iProp :=
    ([∗ map] t ↦ m ∈ C,
      match m.(mval) with
      | VVal v => π t v (default ∅ m.(mrel) ⊔ Va)
      | AVal => π t #☠ (default ∅ m.(mrel) ⊔ Va)
      | _ => True
      end)%I.

  (* Block-ends have F or IW *)
  Definition resBE γF μ C tx :=
    gps_res (λ t v, if (decide (tx ≤ t)%positive)
                    then F γF t (μ t) v
                    else IW l γF t(μ t) v)%I (cbends C).

  (* The rest has something *)
  Definition resD γF μ C tx :=
    gps_res (λ t v, if (decide (tx ≤ t)%positive)
                    then F γF t (μ t) v ∨ F_read γF t (μ t) v
                    else F γF t (μ t) v ∨ F_read γF t (μ t) v ∨ IW l γF t (μ t) v)%I
      (C ∖ (cbends C)).

  (* To get the ghost state, we attach the written value and view
      (of events in the history C) with the associated protocol state (from μ) *)
  Definition toState μ C : stateT Prtcl :=
    map_imap (λ t m,
                match m.(mval) with
                | VVal v => Some (μ t, (v, default ∅ m.(mrel)))
                | AVal => Some (μ t, (#☠, default ∅ m.(mrel)))
                | _ => None
                end) C.

  (* States disconnected from exwr must be final, where disconnected
      means that there is an unused timestamp td in between. *)
  Definition disconnected_from (tx t: time) C :=
    ∃ (td: time), (tx ≤ td)%positive ∧ (td ≤ t)%positive ∧ C !! td = None.
  (* TODO: finality is too strong, can we weaken it? *)
  Definition final_st s := ∀ s', s' ⊑ s.
  (* Final say that anything disconnected from the exclusive write must be final.
    We might need to restrict `disconnected` to writes with state only *)
  Definition FinalInv μ C tx :=
    ∀ t, isState C t → disconnected_from tx t C → final_st (μ t).

  Definition SWConnect C (tx: time) : Prop :=
    ∀ t, is_Some (C !! t) → (tx ≤ t)%positive → ¬ disconnected_from tx t C.

  Definition gps_ownA γ ζ (tx: time) (Va: view) rs : iProp :=
    own γ (● (● toStateMapUR ζ ⋅ ◯ toStateMapUR ζ),
              (● Some (1%Qp, to_agree tx), Some $ to_agree (Va, rs))).

  Definition gps_ownA_M γ ζ: iProp :=
    own γ (● (● toStateMapUR ζ ⋅ ◯ toStateMapUR ζ), (ε,ε)).

  Definition gps_ownA_E γ (tx: time): iProp :=
    own γ (ε, (● Some (1%Qp, to_agree tx), ε)).

  Definition naWrite γ (Va: view) rs : iProp :=
    own γ (ε: stateR, (ε : exWriteR, Some $ to_agree (Va, rs))).

  Definition gps_inv γ γF (bsw: bool) Vo : iProp
    := (∃ μ C (tx : time) (Va: view) rsa rsn ws,
        hist l 1 C ∗ atread l 1 rsa ∗ atwrite l 1 ws ∗ naread l 1 rsn
        ∗ gps_ownA γ (toState μ C) tx Va rsn
        ∗ resBE γF μ C tx Va ∗ resD γF μ C tx Va
        ∗ ⌜FinalInv μ C tx ∧ fsorted μ C
          ∧ good_hist C ∧ isInvView l rsa ws C Vo ∧ Va ⊑ Vo ∧ na_local l rsn Va
          ∧ is_Some (toState μ C !! tx)
          ∧ (∀ t m, C !! t = Some m → m.(mval) ≠ DVal)
          ∧ if bsw then SWConnect C tx else True⌝)%I.

  Notation gps_own γ ζ := (@own _ _ (gps_inG) γ (◯ ζ, (ε,ε))).

  Definition Writer γ ζ := (gps_own γ (● toStateMapUR ζ ⋅ ◯ toStateMapUR ζ)).

  Definition Reader γ ζ := (gps_own γ (◯ toStateMapUR ζ)).

  Definition exwr γ (tx: time) q := own γ ((ε, (◯ Some (q, to_agree tx), ε))).

  Program Definition naViewSeen_def γ Va rs : vProp :=
    MonPred (λ V0, naWrite γ Va rs ∗ ⌜Va ⊑ V0 ∧ na_local l rs Va⌝)%I _.
  Next Obligation. solve_proper. Qed.
  Definition naViewSeen_aux : seal (@naViewSeen_def). by eexists. Qed.
  Definition naViewSeen := unseal (@naViewSeen_aux).
  Definition naViewSeen_eq : @naViewSeen = _ := seal_eq _.

  Program Definition PrtSeen (t: time) s v γ : vProp :=
    MonPred (λ V0, ∃ V Va rs, naViewSeen γ Va rs V0
                          ∗ Reader γ {[t := (s,(v,V))]}
                          ∗ ⌜seen_local l t V0⌝)%I _.
  Next Obligation. solve_proper. Qed.

  Program Definition PrtSeenView (t: time) s v γ : vProp :=
    MonPred (λ V0, ∃ V Va rs, naViewSeen γ Va rs V0
                          ∗ Reader γ {[t := (s,(v,V))]}
                          ∗ ⌜seen_local l t V0 ∧ V ⊑ V0⌝)%I _.
  Next Obligation. solve_proper. Qed.

  Lemma PrtSeenView_PrtSeen t s v γ:
    PrtSeenView t s v γ -∗ PrtSeen t s v γ.
  Proof.
    constructor. iIntros (?). iDestruct 1 as (???) "(na & R & (% & %))".
    iExists _,_,_. by iFrame.
  Qed.

End Definitions.

Section Properties.
  Context `{!noprolG Σ, !gpsG Σ Prtcl}
          (IP : interpC Σ Prtcl)
          (l : loc).
  Local Notation iProp := (iProp Σ).
  Local Notation vProp := (vProp Σ).
  Implicit Types
           (s : pr_stateT Prtcl) (ζ : stateT Prtcl) (e : eventT Prtcl)
           (t : time) (v : val) (V : view) (C : cell)
           (μ : time → pr_stateT Prtcl).

  Notation F := (IP true l).        (* full interp *)
  Notation F_read := (IP false l).  (* read interp *)

  Global Instance naViewSeen_persistent : Persistent (naViewSeen l γ Va rs).
  Proof.
    rewrite naViewSeen_eq /Persistent.
    iStartProof iProp. by iIntros (????) "#?".
  Qed.

  Global Instance naViewSeen_timeless : Timeless (naViewSeen l γ Va rs).
  Proof.
    intros. rewrite /Timeless. rewrite naViewSeen_eq.
    iStartProof iProp. by iIntros (?) ">$".
  Qed.

  Global Instance PrtSeen_persistent: Persistent (PrtSeen l t s v γ).
  Proof. intros. rewrite /Persistent. iStartProof iProp. by iIntros (?) "#?". Qed.

  Global Instance PrtSeen_timeless : Timeless (PrtSeen l t s v γ).
  Proof. intros. rewrite /Timeless. iStartProof iProp. by iIntros (?) ">$". Qed.

  Global Instance PrtSeenView_persistent: Persistent (PrtSeenView l t s v γ).
  Proof. intros. rewrite /Persistent. iStartProof iProp. by iIntros (?) "#?". Qed.

  Global Instance PrtSeenView_timeless : Timeless (PrtSeenView l t s v γ).
  Proof. intros. rewrite /Timeless. iStartProof iProp. by iIntros (?) ">$". Qed.

  Global Instance Reader_persistent γ ζ: Persistent (Reader γ ζ) := _.

  Lemma exwr_update γ t t':
    gps_ownA_E γ t -∗ exwr γ t 1%Qp ==∗ gps_ownA_E γ t' ∗ exwr γ t' 1%Qp.
  Proof.
    iIntros "oE ox". iCombine "oE ox" as "oEx".
    iMod (own_update _ _
              (ε, (● Some (1%Qp, to_agree t') ⋅ ◯ Some (1%Qp, to_agree t'),ε))
              with "oEx") as "[$ $]"; [|done].
    apply prod_update; [done|]. apply prod_update; [|done].
    by apply auth_update, option_local_update, exclusive_local_update.
  Qed.

  Global Instance exwr_fractional γ t: Fractional (exwr γ t)%I.
  Proof.
    intros ??. by rewrite -own_op 2!pair_op -(auth_frag_op (Some _))
                  -Some_op pair_op agree_idemp frac_op' 2!left_id.
  Qed.

  (* Writers *)
  Lemma Writer_latest γ ζ1 ζ2 :
    Writer γ ζ2 ∗ Reader γ ζ1 -∗ ⌜ζ1 ⊆ ζ2⌝.
  Proof.
    rewrite -own_op pair_op -auth_frag_op
            -(cmra_assoc (● toStateMapUR ζ2)) -auth_frag_op own_valid.
    iDestruct 1 as %[[VALID _]%auth_valid_discrete_2 _]. iPureIntro.
    apply stateMap_included. by etrans; [apply cmra_included_r|apply VALID].
  Qed.

  Lemma Writer_Exclusive γ ζ1 ζ2 :
    Writer γ ζ1 -∗ Writer γ ζ2 -∗ False.
  Proof.
    iIntros "Writer1 Writer2". iCombine "Writer1" "Writer2" as "oW".
    by iDestruct (own_valid with "oW") as %[].
  Qed.

  Lemma Writer_exact γ ζ ζ': gps_ownA_M γ ζ -∗ Writer γ ζ' -∗ ⌜ζ = ζ'⌝.
  Proof.
    iIntros "oA oW". iCombine "oA" "oW" as "oS".
    iDestruct (own_valid with "oS")
      as %[[[VALID%Excl_included _]%auth_included _]%auth_valid_discrete_2 _].
    iPureIntro. by apply stateMap_eq.
  Qed.

  Lemma gps_ownA_split γ ζ tx Va rs:
    gps_ownA γ ζ tx Va rs ≡ (gps_ownA_M γ ζ ∗ gps_ownA_E γ tx ∗ naWrite γ Va rs)%I.
  Proof. by rewrite -2!own_op 2!pair_op /= 3!right_id 3!left_id. Qed.

  Lemma Writer_exact_2 γ ζ ζ' tx Va rs:
    gps_ownA γ ζ tx Va rs -∗ Writer γ ζ' -∗ ⌜ζ = ζ'⌝.
  Proof.
    rewrite gps_ownA_split. iIntros "[oA _]". by iApply (Writer_exact with "oA").
  Qed.

  Lemma Writer_update γ ζ ζ' (Sub: ζ ⊆ ζ'):
    gps_ownA_M γ ζ ∗ Writer γ ζ ==∗ gps_ownA_M γ ζ' ∗ Writer γ ζ'.
  Proof.
    rewrite -2!own_op 2!pair_op.
    apply own_update, prod_update; simpl; [|done].
    have ? := stateMap_valid ζ'.
    apply auth_update, auth_local_update; [|done|done].
    apply stateMap_local_update; [by apply stateMap_included|done].
  Qed.

  Lemma Writer_update_2 γ ζ0 ζ ζ' (Sub: ζ ⊆ ζ'):
    gps_ownA_M γ ζ0 -∗ Writer γ ζ ==∗ gps_ownA_M γ ζ' ∗ Writer γ ζ'.
  Proof.
    iIntros "oA Writer". iDestruct (Writer_exact with "oA Writer") as %Eq.
    rewrite Eq. by iApply (Writer_update with "[$oA $Writer]").
  Qed.

  Lemma ExWrite_agree γ (t1 t2: time) q :
    gps_ownA_E γ t1 -∗ exwr γ t2 q -∗ ⌜t1 = t2⌝.
  Proof.
    iIntros "oA oS". iCombine "oA" "oS" as "oA".
    by iDestruct (own_valid with "oA")
      as %[_ [[[_ INCL%to_agree_included]%Some_pair_included_total_2 _]%auth_valid_discrete_2 _]].
  Qed.

  Lemma ExWrite_frac_agree γ (t1 t2: time) q1 q2 :
    exwr γ t1 q1 -∗ exwr γ t2 q2 -∗ ⌜t1 = t2⌝.
  Proof.
    iIntros "o1 o2". iCombine "o1 o2" as "o".
    by iDestruct (own_valid with "o") as %[_ [[_ ?%agree_op_invL'] _]].
  Qed.

  Lemma naWrite_agree γ (V1 V2: view) rs1 rs2 :
    naWrite γ V1 rs1 -∗ naWrite γ V2 rs2 -∗ ⌜V1 = V2 ∧ rs1 = rs2⌝.
  Proof.
    iIntros "V1 V2". iCombine "V1" "V2" as "V".
    iDestruct (own_valid with "V") as %[_ [_ VAL]].
    iPureIntro. move : VAL => /=. rewrite Some_valid.
    move => /agree_op_invL' [//].
  Qed.

  Lemma naWrite_agree_2 γ ζ tx Va rs Va' rs':
    gps_ownA γ ζ tx Va rs -∗ naWrite γ Va' rs' -∗ ⌜Va = Va' ∧ rs = rs'⌝.
  Proof.
    rewrite gps_ownA_split. iIntros "(_&_&na1) na2".
    by iApply (naWrite_agree with "na1 na2").
  Qed.

  (* Writers and Readers *)
  Lemma Writer_fork_Reader γ ζ : Writer γ ζ -∗ Reader γ ζ.
  Proof. iIntros "[_ $]". Qed.

  Lemma Reader_extract γ ζ1 ζ2 (Sub: ζ2 ⊆ ζ1) :
    Reader γ ζ1 -∗ Reader γ ζ2.
  Proof.
    apply own_mono, prod_included. split; [|done]. repeat apply auth_frag_mono.
    by apply stateMap_included.
  Qed.

  Lemma Writer_fork_sub_Reader γ ζ ζ' (Sub: ζ' ⊆ ζ):
    Writer γ ζ -∗ Reader γ ζ'.
  Proof.
    iIntros "Writer". iDestruct (Writer_fork_Reader with "Writer") as "?".
    by iApply (Reader_extract _ ζ).
  Qed.

  Lemma Writer_update_insert_Reader γ ζ0 ζ t svV (FRESH: ζ !! t = None) :
    gps_ownA_M γ ζ0 -∗ Writer γ ζ
    ==∗ gps_ownA_M γ (<[t := svV]> ζ)
      ∗ Writer γ (<[t := svV]> ζ) ∗ Reader γ ({[ t := svV ]}).
  Proof.
    iIntros "own W".
    iMod (Writer_update_2 with "own W") as "[$ W]"; [by apply insert_subseteq|].
    iDestruct (Writer_fork_sub_Reader with "W") as "#$"; [|by iFrame "W"].
    by apply insert_mono, gmap_subseteq_empty.
  Qed.

  Lemma gps_ownA_M_fork_Reader γ ζ:
    gps_ownA_M γ ζ ==∗ gps_ownA_M γ ζ ∗ Reader γ ζ.
  Proof.
    rewrite -own_op pair_op /gps_ownA_M.
    apply own_update, prod_update; [|done].
    apply auth_update_alloc, local_update_unital_discrete => z Val.
    rewrite -> (left_id _ _) => <-. split; [done|].
    by rewrite (cmra_comm (◯ _)) -cmra_assoc -auth_frag_op stateMap_idemp.
  Qed.

  Lemma gps_ownA_M_fork_Reader_2 γ ζ ζ' (Sub: ζ' ⊆ ζ):
    gps_ownA_M γ ζ ==∗ gps_ownA_M γ ζ ∗ Reader γ ζ'.
  Proof.
    iIntros "oA". iMod (gps_ownA_M_fork_Reader with "oA") as "[$ oR]".
    iModIntro. by iApply (Reader_extract _ ζ).
  Qed.

  Lemma gps_ownA_M_Reader_sub γ ζ ζ':
    gps_ownA_M γ ζ' -∗ Reader γ ζ -∗ ⌜ζ ⊆ ζ'⌝.
  Proof.
    iIntros "oA R". iCombine "oA" "R" as "oAR".
    rewrite own_valid.
    iDestruct "oAR" as %[[[_ Valid]%auth_included _]%auth_valid_discrete_2 _].
    iPureIntro. move : Valid => /=. rewrite -> (left_id _ _).
    by rewrite stateMap_included.
  Qed.

  Lemma gps_ownA_Reader_sub γ tx Va rs ζ ζ':
    gps_ownA γ ζ' tx Va rs -∗ Reader γ ζ -∗ ⌜ζ ⊆ ζ'⌝.
  Proof.
    rewrite gps_ownA_split. iIntros "[oA _]".
    iApply (gps_ownA_M_Reader_sub with "oA").
  Qed.

  Lemma gps_ownA_M_fork_Reader_singleton γ ζ t svV (HL: ζ !! t = Some svV):
    gps_ownA_M γ ζ ==∗ gps_ownA_M γ ζ ∗ Reader γ {[t:= svV]}.
  Proof. by apply gps_ownA_M_fork_Reader_2, map_singleton_lookup_subseteq. Qed.

  Lemma gps_ownA_Reader_singleton_sub γ tx Va rs t svV ζ:
    gps_ownA γ ζ tx Va rs  -∗ Reader γ {[t := svV]} -∗ ⌜ζ !! t = Some svV⌝.
  Proof.
    iIntros "oA R". iDestruct (gps_ownA_Reader_sub with "oA R") as %SUB.
    iPureIntro. specialize (SUB t). rewrite lookup_singleton in SUB.
    destruct (ζ !! t) as [?|]; by inversion SUB.
  Qed.

  Lemma FinalInv_mono μ C t1 t2 (LE: (t1 ≤ t2)%positive):
    FinalInv μ C t1 → FinalInv μ C t2.
  Proof.
    move => FIN t IS [td [Le Le']]. apply (FIN _ IS).
    exists td. split; [|done]. by etrans.
  Qed.

  (* toState *)
  Lemma toState_singleton_lookup μ C t svV
    (SUB: {[t := svV]} ⊆ toState μ C) :
    toState μ C !! t = Some svV.
  Proof. eapply lookup_weaken; [|exact SUB]. by rewrite lookup_insert. Qed.

  Lemma toState_lookup_state μ C t s vV :
    toState μ C !! t = Some (s, vV) → μ t = s.
  Proof.
    rewrite lookup_imap bind_Some => [[? [_ Hv]]].
    case_match; [congruence|done|by inversion Hv].
  Qed.

  Lemma toState_lookup_message μ C t s vV :
    toState μ C !! t = Some (s, vV)
    → ∃ m, C !! t = Some m ∧ memval_val_rel m.(mval) vV.1 ∧
           vV.2 = default ∅ m.(mrel).
  Proof.
    rewrite lookup_imap bind_Some.
    move => [m [? Hv]]. exists m. split; [done|].
    case_match; inversion Hv; subst; clear Hv; simpl; split=>//; constructor.
  Qed.

  Lemma toState_lookup_state_inv μ C t m v
    (ISV: memval_val_rel m.(mval) v) (HL: C !! t = Some m) :
    toState μ C !! t = Some (μ t, (v, default ∅ m.(mrel))).
  Proof. rewrite lookup_imap HL /=. by inversion ISV. Qed.

  Lemma toState_insert μ C t m v s
    (ISV: memval_val_rel m.(mval) v) :
    toState (<[t:= s]> μ) (<[t:=m]> C)
      = <[t:= (s,(v,default ∅ m.(mrel)))]> (toState μ C).
  Proof.
    apply map_eq => t'. rewrite /toState lookup_imap.
    case (decide (t' = t)) => [->|?].
    - rewrite 2!lookup_insert fn_lookup_insert /=. by inversion ISV.
    - rewrite !lookup_insert_ne // lookup_imap. destruct (C !! t') as [m'|]; last done.
      rewrite /= fn_lookup_insert_ne //.
  Qed.

  Global Instance fsorted_mono μ : Proper ((⊆) ==> (flip impl)) (fsorted μ).
  Proof.
    move => ??? FS ?? [?[??]] [?[??]].
    apply FS; eexists; (split; [by eapply lookup_weaken|done]).
  Qed.

  Global Instance stsorted_mono : Proper ((⊆) ==> (flip impl)) (@stsorted Prtcl).
  Proof. move => ??? FS ??????. apply FS; by eapply lookup_weaken. Qed.

  Lemma toState_stsorted μ C (FS: fsorted μ C): stsorted (toState μ C).
  Proof.
    move => ?? [s1 [v1 V1]] [s2 [v2 V2]] HL1 HL2 /=.
    have ? := toState_lookup_state _ _ _ _ _ HL1.
    have ? := toState_lookup_state _ _ _ _ _ HL2.
    subst s1 s2.
    apply toState_lookup_message in HL1 as [? [? [Eqv1 ?]]].
    apply toState_lookup_message in HL2 as [? [? [Eqv2 ?]]].
    apply FS; eexists; (split; first done); by [inversion Eqv1|inversion Eqv2].
  Qed.

  Lemma Writer_stsorted IW γ γF bs Vo ζ E:
    ▷ gps_inv IP IW l γ γF bs Vo -∗ Writer γ ζ
    ={E}=∗ ▷ gps_inv IP IW l γ γF bs Vo ∗ Writer γ ζ ∗ ⌜stsorted ζ⌝.
  Proof.
    iDestruct 1 as (μ C tx Va rsa rsn ws) "(? & ? & ? & ? & >oA & ? & rD & >H)".
    iDestruct "H" as %(?&? & SORTED & ?).
    iIntros "W". iDestruct (Writer_exact_2 with "oA W") as %?. subst ζ.
    iModIntro. iFrame "W". iSplit.
    - iExists μ, C, tx, Va, rsa, rsn, ws. iNext. by iFrame.
    - iPureIntro. by apply toState_stsorted.
  Qed.

  Lemma Reader_stsorted IW γ γF bs Vo ζ E:
    ▷ gps_inv IP IW l γ γF bs Vo -∗ Reader γ ζ
    ={E}=∗ ▷ gps_inv IP IW l γ γF bs Vo ∗ ⌜stsorted ζ⌝.
  Proof.
    iDestruct 1 as (μ C tx Va rsa rsn ws ) "(? & ? & ? & ? & >oA & ? & rD & >H)".
    iDestruct "H" as %(?&? & SORTED & ?).
    iIntros "R". iModIntro. iSplit.
    - iExists μ, C, tx , Va, rsa, rsn, ws. iNext. by iFrame.
    - iDestruct (gps_ownA_Reader_sub with "oA R") as %SUB.
      iPureIntro. rewrite -> SUB. by apply toState_stsorted.
  Qed.

  Lemma Reader_singleton_agree IW γ γF t1 s1 vV1 t2 s2 vV2 bs Vo E:
    ▷ gps_inv IP IW l γ γF bs Vo -∗
    Reader γ {[t1 := (s1, vV1)]} -∗
    Reader γ {[t2 := (s2, vV2)]} ={E}=∗
    ▷ gps_inv IP IW l γ γF bs Vo ∗ ⌜(s1 ⊑ s2 ∧ t1 ⊑ t2) ∨ (s2 ⊑ s1 ∧ t2 ⊑ t1)⌝.
  Proof.
    iDestruct 1 as (μ C tx Va rsa rsn ws ) "(? & ? & ? & ? & >oA & ? & rD & >H)".
    iDestruct "H" as %(? & SORTED & ?). iIntros "R1 R2".
    iDestruct (gps_ownA_Reader_sub with "oA R1") as %Sub1%toState_singleton_lookup.
    iDestruct (gps_ownA_Reader_sub with "oA R2") as %Sub2%toState_singleton_lookup.
    iModIntro. iSplit.
    - iExists μ, C, tx, Va, rsa, rsn, ws. iNext. by iFrame.
    - iPureIntro. apply toState_stsorted in SORTED.
      case (decide (t1 ≤ t2)%positive) => [?|/Pos.lt_nle /Pos.lt_le_incl NE];
        [left|right].
      + split; [by apply (SORTED _ _ _ _ Sub1 Sub2)|done].
      + split; [by apply (SORTED _ _ _ _ Sub2 Sub1)|done].
  Qed.

  Lemma naViewSeen_local_seen γ ζ tx Va Va' rs rs' V:
    gps_ownA γ ζ tx Va rs -∗ naViewSeen l γ Va' rs' V -∗
      ⌜Va = Va' ∧ rs = rs' ∧ Va ⊑ V ∧ na_local l rs Va ∧ na_local l rs V⌝.
  Proof.
    iIntros "oA". rewrite naViewSeen_eq. iDestruct (1) as "(naW & % & %)".
    iDestruct (naWrite_agree_2 with "oA naW") as %[]. subst Va rs.
    iPureIntro. do 4 (split; [done|]). eapply na_local_mono; [done| |eauto]. done.
  Qed.

  Lemma gps_ownA_alloc ζ t V rs:
    (|==> ∃ γ, gps_ownA γ ζ t V rs ∗ Writer γ ζ ∗ exwr γ t 1%Qp ∗ naWrite γ V rs)%I.
  Proof.
    rewrite /gps_ownA /Writer /naWrite /exwr.
    do 3 setoid_rewrite <- own_op. rewrite 4!pair_op /= 2!left_id_L -Some_op.
    setoid_rewrite agree_idemp.
    iApply own_alloc. split; last split; simpl; [..|done].
    - rewrite 2!right_id. apply auth_valid_discrete_2. split; [done|].
      apply auth_valid_discrete_2; split; [done|]. apply stateMap_valid.
    - by apply auth_valid_discrete_2.
  Qed.

  Lemma SWConnect_singleton t m: SWConnect {[t := m]} t.
  Proof.
    move => t' [m' /lookup_singleton_Some [? ?]] Le [td [? [? NIN]]].
    subst t' m'. have ?: td = t by apply : anti_symm. subst td.
    by rewrite lookup_insert in NIN.
  Qed.

End Properties.

Lemma seen_local_view_sqsubseteq l t V (SL: seen_local l t V):
  {[l := [{t,∅,∅,∅}]]} ⊑ V.
Proof.
  move => l'. case (decide (l' = l)) => [->|?]; [|by rewrite lookup_insert_ne].
  apply view_sqsubseteq.
  rewrite view_lookup_w_insert view_lookup_ar_insert
          view_lookup_aw_insert view_lookup_nr_insert /=.
  split; [done|].
  rewrite /view_lookup_awrite /view_lookup_aread /view_lookup_nread.
  destruct (V !! l) as [[]|] eqn:Eql; simpl; [done|].
  move : SL. by rewrite /seen_local (view_lookup_w' _ _ _ Eql).
Qed.

Section GName.
  Context `{!noprolG Σ, !gpsG Σ Prtcl} (l : loc)
          (IP : gname → interpC Σ Prtcl).
  Local Notation iProp := (iProp Σ).
  Local Notation vProp := (vProp Σ).

  Lemma GPSRaw_Init_strong_vs_gname b t v s m rsa rsn ws bs V E
    (fγ : gname → gname) (Q: gname → vProp)
    (ISV: memval_val_rel m.(mval) v) (SL: seen_local l t V)
    (AT: atr_local l rsa V) (AW: atw_local l ws V) (NA: na_local l rsn V)
    (LEV: m.(mrel) ⊑ Some V):
    let C : cell := {[ t := m ]} in good_hist C →
    let V': view := default ∅ m.(mrel) in
    hist l 1 C -∗ atread l 1 rsa -∗ atwrite l 1 ws -∗ naread l 1 rsn -∗
    (∀ γ, exwr γ t 1%Qp
        -∗ Writer γ {[t := (s, (v,V'))]}
        -∗ Reader γ {[t := (s, (v,V'))]}
        -∗ naViewSeen l γ V rsn V
        ={E}=∗ ▷?b (IP γ) true l (fγ γ) t s v V ∗ Q γ V)
    ={E}=∗ ∃ γ, ∀ IW, ▷?b (gps_inv (IP γ) IW l γ (fγ γ) bs V) ∗ Q γ V.
  Proof.
    iIntros (C GOOD V') "hist at aw nar F".
    set μ : time → Prtcl := λ _, s.
    set ζ := toState μ C.

    have Eqζ: ζ = {[t := (s, (v, V'))]}.
    { rewrite /ζ /C /=. apply map_eq => t'.
      case (decide (t' = t)) => [->|?].
      - rewrite lookup_insert.
        apply toState_lookup_state_inv; [done|by rewrite lookup_insert].
      - rewrite /toState lookup_imap.
        by do 2 (rewrite lookup_insert_ne; [|done]). }

    iMod (gps_ownA_alloc ζ t V rsn) as (γ) "(oA & W & ox & #na)".
    iExists γ.
    iDestruct (Writer_fork_sub_Reader _ ζ ζ with "W") as "#R"; [done|].
    iMod ("F" $! γ with "ox [W] [R] []") as "(F&$)";
      [by rewrite Eqζ|by rewrite Eqζ|rewrite naViewSeen_eq; by iFrame "na"|..].

    subst ζ. iIntros "!>" (IW). iNext.
    iExists μ, C, t, V, rsa, rsn, ws. iFrame "hist at aw nar oA".
    iSplitL "F"; last iSplit.
    + rewrite /resBE /gps_res /cbends mblock_ends_singleton.
      iApply big_sepM_singleton.
      inversion ISV; (rewrite decide_True; [iFrame "F"|done]).
    + by rewrite /resD /gps_res /cbends mblock_ends_singleton
                  map_difference_diag big_sepM_empty.
    + iPureIntro. split; last split; last split; last split; last split;
        last split; last split; last split; [..|destruct bs; [|done]];
          [| |done| |done|done|..|by apply SWConnect_singleton].
      * rewrite /C. move => ?[?[/lookup_singleton_Some[<- <-]]] _ [td[?[? DIS]]].
        have ?: td = t by apply : anti_symm. subst td.
        by rewrite lookup_insert in DIS.
      * by move => ?? [?[/lookup_singleton_Some[<- _]]] _
                   [?[/lookup_singleton_Some[<- _]]].
      * do 2 (split; [done|]). by move => ?? /lookup_singleton_Some [<- <-].
      * rewrite Eqζ lookup_insert. by eexists.
      * move => ?? /lookup_singleton_Some [? <-]. by inversion ISV.
  Qed.
End GName.

Section TwoLoc.
  Context `{!noprolG Σ, !gpsG Σ Prtcl1, !gpsG Σ Prtcl2} (l1 l2 : loc)
          (IP1 : gname → interpC Σ Prtcl1)
          (IP2 : gname → interpC Σ Prtcl2).
  Local Notation iProp := (iProp Σ).
  Local Notation vProp := (vProp Σ).

  Lemma GPSRaw_Init_strong_vs_2 b (t1 t2: time) v1 v2
    (s1: Prtcl1) (s2: Prtcl2) m1 m2 rsa1 rsn1 rsa2 rsn2 ws1 ws2 bs1 bs2 V E
    (fγ1 fγ2 : gname → gname) (Q: gname → gname → vProp)
    (ISV1: memval_val_rel m1.(mval) v1)
    (ISV2: memval_val_rel m2.(mval) v2)
    (SL1: seen_local l1 t1 V) (SL2: seen_local l2 t2 V)
    (AT1: atr_local l1 rsa1 V) (AT2: atr_local l2 rsa2 V)
    (AW1: atw_local l1 ws1 V) (AW2: atw_local l2 ws2 V)
    (NA1: na_local l1 rsn1 V) (NA2: na_local l2 rsn2 V)
    (LEV1: m1.(mrel) ⊑ Some V) (LEV2: m2.(mrel) ⊑ Some V):
    let C1 : cell := {[ t1 := m1 ]} in good_hist C1 →
    let C2 : cell := {[ t2 := m2 ]} in good_hist C2 →
    let V1: view := default ∅ m1.(mrel) in
    let V2: view := default ∅ m2.(mrel) in
    hist l1 1 C1 -∗ hist l2 1 C2 -∗
    atread l1 1 rsa1 -∗ atread l2 1 rsa2 -∗
    atwrite l1 1 ws1 -∗ atwrite l2 1 ws2 -∗
    naread l1 1 rsn1 -∗ naread l2 1 rsn2 -∗
    (∀ γ1 γ2, exwr (Prtcl:=Prtcl1) γ1 t1 1%Qp
        -∗ Writer γ1 {[t1 := (s1, (v1,V1))]}
        -∗ Reader γ1 {[t1 := (s1, (v1,V1))]}
        -∗ naViewSeen (Prtcl:=Prtcl1) l1 γ1 V rsn1 V
        -∗ exwr (Prtcl:=Prtcl2) γ2 t2 1%Qp
        -∗ Writer γ2 {[t2 := (s2, (v2,V2))]}
        -∗ Reader γ2 {[t2 := (s2, (v2,V2))]}
        -∗ naViewSeen (Prtcl:=Prtcl2) l2 γ2 V rsn2 V
        ={E}=∗ ▷?b (IP1 γ2) true l1 (fγ1 γ1) t1 s1 v1 V ∗
            ▷?b (IP2 γ1) true l2 (fγ2 γ2) t2 s2 v2 V ∗
            Q γ1 γ2 V)
    ={E}=∗ ∃ γ1 γ2, ∀ IW1 IW2,
          ▷?b (gps_inv (IP1 γ2) IW1 l1 γ1 (fγ1 γ1) bs1 V) ∗
          ▷?b (gps_inv (IP2 γ1) IW2 l2 γ2 (fγ2 γ2) bs2 V) ∗
          Q γ1 γ2 V.
  Proof.
    iIntros (C1 GOOD1 C2 GOOD2 V1 V2) "hist1 hist2 at1 at2 aw1 aw2 nar1 nar2 F".
    set μ1 : time → Prtcl1 := λ _, s1.
    set μ2 : time → Prtcl2 := λ _, s2.
    set ζ1 := toState μ1 C1.
    set ζ2 := toState μ2 C2.

    have Eqζ1: ζ1 = {[t1 := (s1, (v1, V1))]}.
    { rewrite /ζ1 /C1 /=. apply map_eq => t'.
      case (decide (t' = t1)) => [->|?].
      - rewrite lookup_insert.
        apply toState_lookup_state_inv; [done|by rewrite lookup_insert].
      - rewrite /toState lookup_imap.
        by do 2 (rewrite lookup_insert_ne; [|done]). }
    have Eqζ2: ζ2 = {[t2 := (s2, (v2, V2))]}.
    { rewrite /ζ2 /C2 /=. apply map_eq => t'.
      case (decide (t' = t2)) => [->|?].
      - rewrite lookup_insert.
        apply toState_lookup_state_inv; [done|by rewrite lookup_insert].
      - rewrite /toState lookup_imap.
        by do 2 (rewrite lookup_insert_ne; [|done]). }

    iMod (gps_ownA_alloc (Prtcl:=Prtcl1) ζ1 t1 V rsn1)
          as (γ1) "(oA1 & W1 & ox1 & #na1)".
    iMod (gps_ownA_alloc (Prtcl:=Prtcl2) ζ2 t2 V rsn2)
          as (γ2) "(oA2 & W2 & ox2 & #na2)".
    iExists γ1, γ2.
    iDestruct (Writer_fork_sub_Reader _ ζ1 ζ1 with "W1") as "#R1"; [done|].
    iDestruct (Writer_fork_sub_Reader _ ζ2 ζ2 with "W2") as "#R2"; [done|].
    iMod ("F" $! γ1 γ2 with "ox1 [W1] [R1] [] ox2 [W2] [R2] []")
      as "(F1 & F2 &$)";
      [by rewrite Eqζ1|by rewrite Eqζ1|rewrite naViewSeen_eq; by iFrame "na1"|
       by rewrite Eqζ2|by rewrite Eqζ2|rewrite naViewSeen_eq; by iFrame "na2"|..].

    subst ζ1 ζ2. iIntros "!>" (IW1 IW2).
    iSplitL "hist1 at1 aw1 nar1 oA1 F1"; iNext.
    - iExists μ1, C1, t1, V, rsa1, rsn1, ws1. iFrame "hist1 at1 aw1 nar1 oA1".
      iSplitL "F1"; last iSplit.
      + rewrite /resBE /gps_res /cbends mblock_ends_singleton.
        iApply big_sepM_singleton.
        inversion ISV1; (rewrite decide_True; [iFrame "F1"|done]).
      + by rewrite /resD /gps_res /cbends mblock_ends_singleton
                    map_difference_diag big_sepM_empty.
      + iPureIntro. split; last split; last split; last split; last split;
          last split; last split; last split; [..|destruct bs1; [|done]];
          [| |done| |done|done|..|by apply SWConnect_singleton].
        * rewrite /C1. move => ?[?[/lookup_singleton_Some[<- <-]]] _ [td[?[? DIS]]].
          have ?: td = t1 by apply : anti_symm. subst td.
          by rewrite lookup_insert in DIS.
        * by move => ?? [?[/lookup_singleton_Some[<- _]]] _
                     [?[/lookup_singleton_Some[<- _]]].
        * do 2 (split; [done|]). by move => ?? /lookup_singleton_Some [<- <-].
        * rewrite Eqζ1 lookup_insert. by eexists.
        * move => ?? /lookup_singleton_Some [? <-]. by inversion ISV1.
    - iExists μ2, C2, t2, V, rsa2, rsn2, ws2. iFrame "hist2 at2 aw2 nar2 oA2".
      iSplitL "F2"; last iSplit.
      + rewrite /resBE /gps_res /cbends mblock_ends_singleton.
        iApply big_sepM_singleton.
        inversion ISV2; (rewrite decide_True; [iFrame "F2"|done]).
      + by rewrite /resD /gps_res /cbends mblock_ends_singleton
                    map_difference_diag big_sepM_empty.
      + iPureIntro. split; last split; last split; last split; last split;
          last split; last split; last split; [..|destruct bs2; [|done]];
          [| |done| |done|done|..|by apply SWConnect_singleton].
        * rewrite /C2. move => ?[?[/lookup_singleton_Some[<- <-]]] _ [td[?[? DIS]]].
          have ?: td = t2 by apply : anti_symm. subst td.
          by rewrite lookup_insert in DIS.
        * by move => ?? [?[/lookup_singleton_Some[<- _]]] _
                     [?[/lookup_singleton_Some[<- _]]].
        * do 2 (split; [done|]). by move => ?? /lookup_singleton_Some [<- <-].
        * rewrite Eqζ2 lookup_insert. by eexists.
        * move => ?? /lookup_singleton_Some [? <-]. by inversion ISV2.
  Qed.
End TwoLoc.

Section RawRules.
  Context `{!noprolG Σ, !gpsG Σ Prtcl} (IP : interpC Σ Prtcl) (l : loc).
  Local Notation iProp := (iProp Σ).
  Local Notation vProp := (vProp Σ).
  Implicit Types
           (s : pr_stateT Prtcl) (ζ : stateT Prtcl) (e : eventT Prtcl)
           (t : time) (v : val) (V : view) (C : cell)
           (μ : time → pr_stateT Prtcl).

  Notation F := (IP true l).
  Notation F_read := (IP false l).

  (** DEALLOC *)
  Lemma GPSRaw_dealloc IW b γ γF bs Vo E (SUB: ↑histN ⊆ E) :
     hist_inv -∗ ▷?b gps_inv IP IW l γ γF bs Vo
     ={E}=∗ ∃ t s v, (l ↦ v) Vo ∗ ▷?b F γF t s v Vo.
  Proof.
    iIntros "#Inv".
    iDestruct 1 as (μ C tx Va rsa rsn ws)
                    "(>hist & >at & >aw & >nar & own & rB & _ & >FACT)".
    iDestruct "FACT" as %(?&?&GH&[AT [AW JV]]&?&NAL&IS&ND&_).
    (* FIXME: duplicating Read *)
    (* get the top write *)
    destruct (gmap_top_nonempty (flip (⊑)) (toState μ C)) as [t [[s [v V]] Eqt]].
    { move => Eq. rewrite Eq in IS. by destruct IS. }
    have EQm: toState μ C !! t = Some (s, (v, V))
        by apply (gmap_top_lookup _ _ _ _ Eqt).
    have TOP: ∀ k', k' ∈ dom (gset _) (toState μ C) → k' ⊑ t
        by apply (gmap_top_top _ _ _ _ Eqt).
    destruct (toState_lookup_message _ _ _ _ _ EQm) as [m [EqCm [Eqvm EqVm]]].
    have Eqsm := toState_lookup_state _ _ _ _ _ EQm.

    (* the top write is certainly a block-end *)
    have Letx: (tx ≤ t)%positive by apply TOP, elem_of_dom.
    have HLa: (cbends C) !! t = Some m.
    { apply map_filter_lookup_Some. split; [done|]. rewrite /cell_adj.
      move => t0 m0 Eq0 /= NEq0 ?. subst t0.
      apply (irreflexive_eq (R:=(⊏)) t t); [done|].
      eapply (strict_transitive_l _ (t+1)%positive).
      - split. apply Pos.lt_le_incl. lia. apply Pos.lt_nle. lia.
      - apply TOP, elem_of_dom. specialize (ND _ _ Eq0).
        destruct (m0.(mval)) as [| |v0] eqn:Eqv0; [|done|];
          eexists; (apply toState_lookup_state_inv; [|done]);
          rewrite Eqv0; by constructor. }

    rewrite -{1}(insert_id _ _ _ EqCm).

    iInv histN as (σ) ">[Hσ ctx]" "HClose".
    iMod (hist_ctx_hist_drop_singleton with "ctx hist") as "[ctx hist]";
      [|by inversion Eqvm|].
    { move => t' [m' Eqt']. apply TOP, elem_of_dom.
      specialize (ND _ _ Eqt').
      destruct m'.(mval) eqn:Eqv'; [|done|]; eexists;
        (apply toState_lookup_state_inv; [|done]); rewrite Eqv'; by constructor. }

    iDestruct (hist_ctx_hist_good with "ctx hist") as %GH'.
    iMod ("HClose" with "[ctx Hσ]") as "_". { iExists _. by iFrame. }
    iExists t, s.
    rewrite /resBE /gps_res (big_sepM_lookup _ _ _ _ HLa) Eqsm.
    destruct (JV _ _ EqCm) as [SL LE].
    have ?: default ∅ (mrel m) ⊔ Va ⊑ Vo
      by apply lat_join_lub; [rewrite LE|].
    have ?: alloc_local l {[t := m]} Vo.
    { exists t, m. rewrite lookup_insert.
      split; [done|split;by [inversion Eqvm|]]. }
    have ?: na_local l rsn Vo by eapply na_local_mono; [..|apply NAL]; done.
    iModIntro. inversion Eqvm as [|EqA].
    - iExists _. rewrite (decide_True _ _ Letx). iFrame "rB".
      rewrite own_loc_na_eq. iExists _, _. iSplit; [|done]. iExists _,_,_.
      by iFrame "hist at aw nar".
    - iExists _. rewrite (decide_True _ _ Letx). iFrame "rB".
      rewrite own_loc_na_eq. iExists _, _. iSplit.
      + iExists _,_,_. by iFrame "hist at aw nar".
      + iPureIntro. rewrite -EqA. split; [by constructor|done].
  Qed.

  Lemma GPSRaw_SW_dealloc IW b ζ t s v V γ γF bs V0 Va rsn E
    (SUB: ↑histN ⊆ E)
    (TOP: ∀ t', is_Some (ζ !! t') → t' ⊑ t)
    (LE0: V0 ⊑ V) (SL0: seen_local l t V) :
     hist_inv -∗ ▷?b gps_inv IP IW l γ γF bs V
     -∗ Writer γ ζ -∗ Reader γ {[t := (s, (v, V0))]}
     -∗ naViewSeen l γ Va rsn V
     ={E}=∗ (l ↦ v) V ∗ ▷?b F γF t s v V.
  Proof.
    iIntros "#Inv".
    iDestruct 1 as (μ C tx Va' rsa rsn' ws)
                    "(>hist & >at & >aw & >nar & >own & rB & _ & >FACT)".
    iDestruct "FACT" as %(?&?&GH&[AT [AW JV]]&?&NAL&IS&ND&_).
    iIntros "W R na".
    iDestruct (gps_ownA_Reader_singleton_sub with "own R") as %EQL.
    iDestruct (Writer_exact_2 with "own W") as %?. subst ζ.
    iDestruct (naViewSeen_local_seen with "own na") as %(?&?&LEa&LNA&LNA2). subst Va' rsn'.
    (* FIXME: duplicating Read *)
    (* get the top write *)
    destruct (gmap_top_nonempty (flip (⊑)) (toState μ C)) as [tm [sm Eqt]].
    { move => Eq. rewrite Eq in IS. by destruct IS. }
    have ?: tm = t.
    { apply : anti_symm.
      - apply TOP. exists sm. by apply (gmap_top_lookup _ _ _ _ Eqt).
      - apply (gmap_top_top _ _ _ _ Eqt), elem_of_dom. by eexists. }
    subst tm.
    destruct (toState_lookup_message _ _ _ _ _ EQL) as [m [EqCm [Eqvm EqVm]]].
    have Eqsm := toState_lookup_state _ _ _ _ _ EQL.

    have Letx: (tx ≤ t)%positive by apply TOP.
    (* the top write is certainly a block-end *)
    have HLa: (cbends C) !! t = Some m.
    { apply map_filter_lookup_Some. split; [done|]. rewrite /cell_adj.
      move => t0 m0 Eq0 /= NEq0 ?. subst t0.
      apply (irreflexive_eq (R:=(⊏)) t t); [done|].
      eapply (strict_transitive_l _ (t+1)%positive).
      - split. apply Pos.lt_le_incl. lia. apply Pos.lt_nle. lia.
      - apply TOP. specialize (ND _ _ Eq0).
        destruct (m0.(mval)) as [| |v0] eqn:Eqv0; [|done|];
          eexists; (apply toState_lookup_state_inv; [|done]);
          rewrite Eqv0; by constructor. }

    rewrite -{1}(insert_id _ _ _ EqCm).
    iInv histN as (σ) ">[Hσ ctx]" "HClose".
    iMod (hist_ctx_hist_drop_singleton with "ctx hist") as "[ctx hist]";
      [|by inversion Eqvm|].
    { move => t' [m' Eqt']. apply TOP.
      specialize (ND _ _ Eqt').
      destruct m'.(mval) eqn:Eqv'; [|done|]; eexists;
        (apply toState_lookup_state_inv; [|done]); rewrite Eqv'; by constructor. }

    iDestruct (hist_ctx_hist_good with "ctx hist") as %GH'.
    iMod ("HClose" with "[ctx Hσ]") as "_". { iExists _. by iFrame. }
    rewrite /resBE /gps_res (big_sepM_lookup _ _ _ _ HLa) Eqsm.
    have Lem: m.(mrel) ⊑ Some V.
    { destruct m.(mrel); [|done]. simpl in EqVm. by rewrite -EqVm. }
    have ?: from_option id ∅ (mrel m) ⊔ Va ⊑ V.
    { apply lat_join_lub; [by rewrite Lem|done]. }
    have ?: alloc_local l {[t := m]} V.
    { exists t, m. rewrite lookup_insert. split; [done|split;by [inversion Eqvm|]]. }
    iModIntro. inversion Eqvm as [|EqA].
    - rewrite (decide_True _ _ Letx). iFrame "rB".
      rewrite own_loc_na_eq. iExists _, _. iSplit; [|done].
      iExists _,_,_. by iFrame "hist at aw nar".
    - simpl in *. subst v. rewrite (decide_True _ _ Letx). iFrame "rB".
      rewrite own_loc_na_eq. iExists _, _. iSplit; [|done].
      iExists _,_,_. by iFrame "hist at aw nar".
  Qed.

  (** INIT *)
  Lemma GPSRaw_Init_vs IW b (t: time) v s m bs V rsa rsn ws (fγ : gname → gname)
    (ISV: memval_val_rel m.(mval) v)
    (AT: atr_local l rsa V) (ATW: atw_local l ws V)
    (NA: na_local l rsn V):
    let C : cell := {[ t := m ]} in good_hist C →
    let V': view := default ∅ m.(mrel) in
    hist l 1 C -∗ atread l 1 rsa -∗ atwrite l 1 ws -∗ naread l 1 rsn -∗
    (∀ γF, ▷?b F γF t s v V) ==∗
      ∃ γ,
       ▷?b (gps_inv IP IW l γ (fγ γ) bs (V' ⊔ {[l := [{t,∅,∅,∅}]]} ⊔ V))
        ∗ exwr γ t 1%Qp
        ∗ Writer γ {[t := (s, (v,V'))]}
        ∗ Reader γ {[t := (s, (v,V'))]}
        ∗ naViewSeen l γ V rsn V.
  Proof.
    iIntros (C GOOD V') "hist at aw nar F".
    set μ : time → Prtcl := λ _, s.
    set ζ := toState μ C.

    have Eqζ: ζ = {[t := (s, (v, V'))]}.
    { rewrite /ζ /C /=. apply map_eq => t'.
      case (decide (t' = t)) => [->|?].
      - rewrite lookup_insert.
        apply toState_lookup_state_inv; [done|by rewrite lookup_insert].
      - rewrite /toState lookup_imap.
        by do 2 (rewrite lookup_insert_ne; [|done]). }

    iMod (gps_ownA_alloc ζ t V rsn) as (γ) "(oA & W & ox & #na)".
    iModIntro. iExists γ.
    iDestruct (Writer_fork_sub_Reader _ ζ ζ with "W") as "#R"; [done|].
    iFrame "ox". subst ζ. rewrite -Eqζ. iFrame "W R".
    iSplit; last first. { rewrite naViewSeen_eq. by iFrame "na". }
    iNext. iExists μ, C, t, V, rsa, rsn, ws. iFrame "hist at aw nar oA".
    iDestruct ("F" $! (fγ γ)) as "F". iSplitL "F"; last iSplit.
    - rewrite /resBE /gps_res /cbends mblock_ends_singleton.
      iApply big_sepM_singleton.
      inversion ISV; (rewrite decide_True; [iFrame "F"|done]).
    - by rewrite /resD /gps_res /cbends mblock_ends_singleton
                  map_difference_diag big_sepM_empty.
    - iPureIntro. split; last split; last split; last split; last split;
        last split; last split; last split; [..|destruct bs; [|done]];
        [| |done| |solve_lat|done|..|by apply SWConnect_singleton].
      + rewrite /C. move => ?[?[/lookup_singleton_Some[<- <-]]] _ [td[?[? DIS]]].
        have ?: td = t by apply : anti_symm. subst td.
        by rewrite lookup_insert in DIS.
      + by move => ?? [?[/lookup_singleton_Some[<- _]]] _
                   [?[/lookup_singleton_Some[<- _]]].
      + by eapply inv_view_singleton_hist.
      + rewrite Eqζ lookup_insert. by eexists.
      + move => ?? /lookup_singleton_Some [? <-]. by inversion ISV.
  Qed.

  Lemma GPSRaw_Init_strong_vs IW b (t: time) v s m bs V rsa rsn ws
    (fγ : gname → gname) (Q: gname → vProp) E
    (ISV: memval_val_rel m.(mval) v)
    (SL: seen_local l t V) (LEV: m.(mrel) ⊑ Some V)
    (AT: atr_local l rsa V) (ATW: atw_local l ws V) (NA: na_local l rsn V) :
    let C : cell := {[ t := m ]} in good_hist C →
    let V': view := default ∅ m.(mrel) in
    hist l 1 C -∗ atread l 1 rsa -∗ atwrite l 1 ws -∗ naread l 1 rsn -∗
    (∀ γ, exwr γ t 1%Qp
        -∗ Writer γ {[t := (s, (v,V'))]}
        -∗ Reader γ {[t := (s, (v,V'))]}
        -∗ naViewSeen l γ V rsn V
        ={E}=∗ ▷?b F (fγ γ) t s v V ∗ Q γ V)
    ={E}=∗ ∃ γ, ▷?b (gps_inv IP IW l γ (fγ γ) bs V) ∗ Q γ V.
  Proof.
    iIntros (C GOOD V') "hist at aw nar F".
    set μ : time → Prtcl := λ _, s.
    set ζ := toState μ C.

    have Eqζ: ζ = {[t := (s, (v, V'))]}.
    { rewrite /ζ /C /=. apply map_eq => t'.
      case (decide (t' = t)) => [->|?].
      - rewrite lookup_insert.
        apply toState_lookup_state_inv; [done|by rewrite lookup_insert].
      - rewrite /toState lookup_imap.
        by do 2 (rewrite lookup_insert_ne; [|done]). }

    iMod (gps_ownA_alloc ζ t V rsn) as (γ) "(oA & W & ox & #na)".
    iExists γ.
    iDestruct (Writer_fork_sub_Reader _ ζ ζ with "W") as "#R"; [done|].
    iMod ("F" $! γ with "ox [W] [R] [na]") as "(F & $)";
      [by rewrite Eqζ..|rewrite naViewSeen_eq; by iFrame "na"|].
    subst ζ. iIntros "!> !>". iExists μ, C, t, V, rsa, rsn, ws.
    iFrame "hist at aw nar oA". iSplitL "F"; last iSplit.
    - rewrite /resBE /gps_res /cbends mblock_ends_singleton.
      iApply big_sepM_singleton.
      inversion ISV; (rewrite decide_True; [iFrame "F"|done]).
    - by rewrite /resD /gps_res /cbends mblock_ends_singleton
                  map_difference_diag big_sepM_empty.
    - iPureIntro. split; last split; last split; last split; last split;
        last split; last split; last split; [..|destruct bs; [|done]];
        [| |done| |done|done|..|by apply SWConnect_singleton].
      + rewrite /C. move => ?[?[/lookup_singleton_Some[<- <-]]] _ [td[?[? DIS]]].
        have ?: td = t by apply : anti_symm. subst td.
        by rewrite lookup_insert in DIS.
      + by move => ?? [?[/lookup_singleton_Some[<- _]]] _
                   [?[/lookup_singleton_Some[<- _]]].
      + do 2 (split; [done|]). move => ?? /lookup_singleton_Some [<- <-] //.
      + rewrite Eqζ lookup_insert. by eexists.
      + move => ?? /lookup_singleton_Some [? <-]. by inversion ISV.
  Qed.

  (** READ *)
  Lemma GPSRaw_read_resources
    IW (isEx: bool) R μ C q (tx tx' tr t': time) m' v γ γF Va E
    (Eqt': C !! t' = Some m') (Eqv': memval_val_rel (mval m') v) (Letr: tr ⊑ t'):
    ((F_read γF t' (μ t') v ={E}=∗ F_read γF t' (μ t') v ∗ R t' (μ t') v) ∧
      (F γF t' (μ t') v ={E}=∗ F γF t' (μ t') v ∗ R t' (μ t') v) ∧
      (if isEx then True
       else IW l γF t' (μ t') v ={E}=∗ IW l γF t' (μ t') v ∗ R t' (μ t') v)) ∅ -∗
    (if isEx then exwr γ tx q ∗ ⌜(tx ≤ tr)%positive⌝ else True) -∗
    resBE IP IW l γF μ C tx' Va -∗
    resD IP IW l γF μ C tx' Va -∗
    gps_ownA_E γ tx' ={E}=∗
    resBE IP IW l γF μ C tx' Va ∗
    resD IP IW l γF μ C tx' Va ∗
    gps_ownA_E γ tx' ∗
    (if isEx then exwr γ tx q ∗ ⌜(tx ≤ tr)%positive⌝ else True) ∗
    (R t' (μ t') v) (default ∅ (mrel m') ⊔ Va).
  Proof.
    iIntros "VP ex rBE rD oE".
    rewrite (monPred_mono _ _ _ (lat_bottom_sqsubseteq (default ∅ (mrel m') ⊔ Va))).
    destruct ((cbends C) !! t') as [mm'|] eqn: Eqmm'.
    - iFrame "rD".
      have ?: mm' = m'.
      { move : Eqmm' => /map_filter_lookup_Some [Eq' _]. rewrite Eqt' in Eq'.
        by inversion Eq'. } subst mm'.
      rewrite bi.and_elim_r /resBE /gps_res (big_sepM_delete _ _ _ _ Eqmm').
      iDestruct "rBE" as "(Hm & $)". destruct isEx.
      + iDestruct "ex" as "[ex Letx]". iDestruct "Letx" as %Letx.
        iDestruct (ExWrite_agree with "oE ex") as %?. subst tx'.
        iFrame "oE ex". rewrite bi.and_elim_l.
        have Letx': (tx ≤ t')%positive by etrans. case (decide (tx ≤ t')%positive) => [?|//].
        inversion Eqv' as [? Eqv''|Eqv'']; by iMod ("VP" with "Hm") as "[$$]".
      + iFrame "oE".
        case (decide (tx' ≤ t')%positive) => Letx;
          [rewrite bi.and_elim_l|rewrite bi.and_elim_r];
          inversion Eqv' as [? Eqv''|Eqv'']; simpl;
          by iMod ("VP" with "Hm") as "[$$]".
    - iFrame "rBE".
      have Eqt2: (C ∖ (cbends C)) !! t' = Some m' by apply lookup_difference_Some.
      rewrite /resD /gps_res. rewrite (big_sepM_delete _ _ _ _ Eqt2).
      iDestruct "rD" as "(Hm & $)". destruct isEx.
      + iDestruct "ex" as "[ex Letx]". iDestruct "Letx" as %Letx.
        iDestruct (ExWrite_agree with "oE ex") as %?. subst tx'.
        iFrame "oE ex".
        have Letx': (tx ≤ t')%positive by etrans. case (decide (tx ≤ t')%positive) => [?|//].
        inversion Eqv' as [? Eqv''|Eqv'']; iDestruct "Hm" as "[Hm|Hm]";
          [rewrite bi.and_elim_r bi.and_elim_l|rewrite bi.and_elim_l|
           rewrite bi.and_elim_r bi.and_elim_l|rewrite bi.and_elim_l];
          by iMod ("VP" with "Hm") as "[$$]".
      + iFrame "oE".
        case (decide (tx' ≤ t')%positive) => Letx;
          [inversion Eqv' as [? Eqv''|Eqv''];iDestruct "Hm" as "[Hm|Hm]"|
           inversion Eqv' as [? Eqv''|Eqv''];iDestruct "Hm" as "[Hm|[Hm|Hm]]"];
          [rewrite bi.and_elim_r bi.and_elim_l|rewrite bi.and_elim_l|
           rewrite bi.and_elim_r bi.and_elim_l|rewrite bi.and_elim_l|
           rewrite bi.and_elim_r bi.and_elim_l|rewrite bi.and_elim_l|
           rewrite bi.and_elim_r bi.and_elim_r|rewrite bi.and_elim_r bi.and_elim_l|
           rewrite bi.and_elim_l|rewrite bi.and_elim_r bi.and_elim_r];
          by iMod ("VP" with "Hm") as "[$$]".
  Qed.

  Lemma GPSRaw_Read
    IW (isEx: bool) tx q R (tr: time) sr vVr o γ γF bs Vo Va rs 𝓥 E
    (Extr : seen_local l tr 𝓥.(cur)) :
    ↑histN ⊆ E → o = Relaxed ∨ o = AcqRel →
    {{{ seen 𝓥
        ∗ ▷ (∀ t' s' v, ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
               ((F_read γF t' s' v ={E}=∗ F_read γF t' s' v ∗ R t' s' v) ∧
                (F γF t' s' v ={E}=∗ F γF t' s' v ∗ R t' s' v) ∧
                (if isEx then True
                 else IW l γF t' s' v ={E}=∗ IW l γF t' s' v ∗ R t' s' v)) ∅)
        ∗ ▷ (gps_inv IP IW l γ γF bs Vo)
        ∗ (if isEx then exwr γ tx q ∗ ⌜(tx ≤ tr)%positive⌝ else True)
        ∗ Reader γ {[tr := (sr, vVr)]}
        ∗ naViewSeen l γ Va rs 𝓥.(cur) }}}
      Read o (Lit l) at 𝓥 @ E
    {{{ t' s' V' 𝓥' v, RET v at 𝓥';
        seen 𝓥'
        ∗ ⌜𝓥 ⊑ 𝓥' ∧ sr ⊑ s' ∧ (tr ≤ t')%positive ∧ seen_local l t' 𝓥'.(cur)
          ∧ V' ⊑ if (decide (AcqRel ⊑ o)) then 𝓥'.(cur) else 𝓥'.(acq)⌝
        ∗ gps_inv IP IW l γ γF bs (Vo ⊔ 𝓥'.(cur))
        ∗ (if isEx then exwr γ tx q else True)
        ∗ R t' s' v V'
        ∗ ∃ Vt, Reader γ {[t' := (s', (v, Vt))]} ∧ ⌜Vt ⊔ Va = V'⌝ }}}.
  Proof.
    iIntros (SUB OR Φ) "(s & VP & Inv & ex & R & naV) Post".
    iDestruct "Inv" as (μ C tx' Va' rsa rs' ws)
                        "(>hist & >at & >aw & >nar & >oA & rBE & rD & >H)".
    iDestruct (naViewSeen_local_seen with "oA naV") as %(?&?&LeVaC&LNA&LNA2).
    subst Va' rs'. iClear "naV".
    iDestruct "H" as %(FINAL&SORTED&GOOD&[AT [AW JV]]&LeVa&NSL&IS&ND&HIW).
    iDestruct (gps_ownA_Reader_singleton_sub with "oA R") as %EQL.
    have Eqsr := toState_lookup_state _ _ _ _ _ EQL.
    destruct (toState_lookup_message _ _ _ _ _ EQL) as [mr [Eqr [Eqvr EqVr]]].

    have ?: Relaxed ⊑ o by destruct OR; subst o.
    iApply wp_fupd.
    iApply (iwp_read_atomic with "[$s $hist $at]"); [done| |done|done|].
    { exists tr, mr. split; [done|]. split; [|done]. by inversion Eqvr. }

    iNext. iIntros (𝓥' v ?) "(s' & hist & at & FACTS)".
    iDestruct "FACTS" as % (Ext & WFC & t' & m' & Eqt' & Eqv' & RH & LNA').
    have Eqs' := toState_lookup_state_inv μ _ _ _ _ Eqv' Eqt'.
    have Letr: tr ⊑ t'.
    { change (Some tr ⊑ Some t'). etrans; [apply Extr|by inversion RH]. }
    have Exts: sr ⊑ μ t'.
    { rewrite -Eqsr. apply SORTED; [..|done].
      - exists mr. split; [done|]. by inversion Eqvr.
      - exists m'. split; [done|]. by inversion Eqv'. }
    have ?: seen_local l t' (cur 𝓥') by eapply read_helper_seen_local.
    have ?: default ∅ (mrel m') ⊔ Va ⊑ (if decide (AcqRel ⊑ o) then cur 𝓥' else acq 𝓥').
    { apply lat_join_lub; [|etrans; [apply LeVaC|]];
        destruct OR; subst o; simpl;
        [by eapply read_helper_view_relaxed_1|by eapply read_helper_view_acq_1
        |etrans; [apply cur_acq|apply Ext]|apply Ext]. }
    iDestruct (gps_ownA_split with "oA") as "[oA [oE naE]]".
    iMod (gps_ownA_M_fork_Reader_singleton _ _ _ _ Eqs' with "oA") as "[oA R']".

    iSpecialize ("VP" $! t' (μ t') v with "[%]"); [done|].

    iAssert (∃ Vt, Reader γ {[t' := (μ t', (v, Vt))]}
            ∧ ⌜Vt ⊔ Va = default ∅ (mrel m') ⊔ Va⌝)%I with "[R']" as "R'".
    { iExists _. by iFrame "R'". }

    iMod (GPSRaw_read_resources IW isEx R μ C q tx tx' tr t' m' v _ _ Va E
                Eqt' Eqv' Letr with "VP ex rBE rD oE") as "(rBE & rD & oE & ex & Rs)".
    iApply ("Post" $! t' (μ t') (default ∅ m'.(mrel) ⊔ Va)
              with "[- $s' $Rs $R']"). iSplit; [done|].
    iDestruct (gps_ownA_split with "[$oA $oE $naE]") as "oA". iSplitR "ex".
    - iExists μ, C, tx', Va, _,_,_. iFrame. iPureIntro.
      do 3 (split; [done|]). split; [|split; [solve_lat|done]].
      split; [done|]. split. { eapply atw_local_mono; [done| |done]. solve_lat. }
      move => ?? /JV [??]. split; [|solve_lat].
      eapply seen_local_mono; [|done]. solve_lat.
    - destruct isEx; by [iDestruct "ex" as "[$_]"|].
  Qed.

  Lemma GPSRaw_Read_ex
    IW R (tx tr: time) q sr vVr o γ γF bs Vo Va rs 𝓥 E
    (Extr : seen_local l tr 𝓥.(cur)) (Letx: (tx ≤ tr)%positive):
    ↑histN ⊆ E → o = Relaxed ∨ o = AcqRel →
    {{{ seen 𝓥
        ∗ ▷ (∀ t' s' v, ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
               ((F_read γF t' s' v ={E}=∗ F_read γF t' s' v ∗ R t' s' v) ∧
                (F γF t' s' v ={E}=∗ F γF t' s' v ∗ R t' s' v)) ∅)
        ∗ ▷ (gps_inv IP IW l γ γF bs Vo)
        ∗ exwr γ tx q
        ∗ Reader γ {[tr := (sr, vVr)]}
        ∗ naViewSeen l γ Va rs 𝓥.(cur) }}}
      Read o (Lit l) at 𝓥 @ E
    {{{ t' s' V' 𝓥' v, RET v at 𝓥';
        seen 𝓥'
        ∗ ⌜𝓥 ⊑ 𝓥' ∧ sr ⊑ s' ∧ (tr ≤ t')%positive ∧ seen_local l t' 𝓥'.(cur)
          ∧ V' ⊑ if (decide (AcqRel ⊑ o)) then 𝓥'.(cur) else 𝓥'.(acq)⌝
        ∗ gps_inv IP IW l γ γF bs (Vo ⊔ 𝓥'.(cur))
        ∗ exwr γ tx q
        ∗ R t' s' v V'
        ∗ ∃ Vt, Reader γ {[t' := (s', (v, Vt))]} ∧ ⌜Vt ⊔ Va = V'⌝ }}}.
  Proof.
    iIntros (SUB OR Φ) "(s & VP & Inv & ex & R & naV) Post".
    iApply (GPSRaw_Read IW true tx q R tr sr vVr o γ γF bs Vo Va rs
              with "[VP $s $Inv $ex $R $naV] Post"); [done..|].
    iFrame (Letx). iIntros "!>" (???). rewrite bi.and_True. by iApply "VP".
  Qed.

  Lemma GPSRaw_Read_non_ex
    IW R (tr: time) sr vVr o γ γF bs Vo Va rs 𝓥 E
    (Extr : seen_local l tr 𝓥.(cur)) :
    ↑histN ⊆ E → o = Relaxed ∨ o = AcqRel →
    {{{ seen 𝓥
        ∗ ▷ (∀ t' s' v, ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
               ((F_read γF t' s' v ={E}=∗ F_read γF t' s' v ∗ R t' s' v) ∧
                (F γF t' s' v ={E}=∗ F γF t' s' v ∗ R t' s' v) ∧
                (IW l γF t' s' v ={E}=∗ IW l γF t' s' v ∗ R t' s' v)) ∅)
        ∗ ▷ (gps_inv IP IW l γ γF bs Vo)
        ∗ Reader γ {[tr := (sr, vVr)]}
        ∗ naViewSeen l γ Va rs 𝓥.(cur) }}}
      Read o (Lit l) at 𝓥 @ E
    {{{ t' s' V' 𝓥' v, RET v at 𝓥';
        seen 𝓥'
        ∗ ⌜𝓥 ⊑ 𝓥' ∧ sr ⊑ s' ∧ (tr ≤ t')%positive ∧ seen_local l t' 𝓥'.(cur)
          ∧ V' ⊑ if (decide (AcqRel ⊑ o)) then 𝓥'.(cur) else 𝓥'.(acq)⌝
        ∗ gps_inv IP IW l γ γF bs (Vo ⊔ 𝓥'.(cur))
        ∗ R t' s' v V'
        ∗ ∃ Vt, Reader γ {[t' := (s', (v, Vt))]} ∧ ⌜Vt ⊔ Va = V'⌝ }}}.
  Proof.
    iIntros (SUB OR Φ) "(s & VP & Inv & R & naV) Post".
    iApply (GPSRaw_Read IW false 1 1%Qp R tr sr vVr o γ γF bs Vo Va rs
              with "[$s $VP $Inv $R $naV]"); [done..|].
    iIntros "!>" (?????) "(?&?&?&_&?&?)". iApply "Post". by iFrame.
  Qed.

  (* FIXME: this rule is weak. We don't need to acquire the Writer (in Wv)
    to get the latest write event. We just need to know that we have seen Vo *)
  Lemma GPSRaw_Read_val_dealloc
    IW (isEx: bool) vd tx q (P : vProp) R R' (tr: time) sr vVr γ γF bs Vo Va rs 𝓥 E
    (Extr : seen_local l tr 𝓥.(cur)) :
    ↑histN ⊆ E →
    let Wv t s v γ : vProp :=
      (∃ ζ, PrtSeenView l t s v γ ∗ ⎡ Writer γ ζ ⎤ ∗
        ⌜∀ t', is_Some (ζ !! t') → t' ⊑ t⌝)%I in
    let Rs q t : vProp :=
      (∃ (tx: time), ⎡ exwr γ tx q ⎤ ∗ ⌜(tx ≤ t)%positive⌝)%I in
    {{{ hist_inv ∗ seen 𝓥 ∗ P 𝓥.(cur)
        ∗ ▷ (∀ t' s', ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
             (P -∗ (if isEx then Rs q tr else True%I) -∗
              (F γF t' s' vd ={E}=∗ monPred_in Vo ∗ Wv t' s' vd γ ∗ R t' s') ∧
              (F_read γF t' s' vd ={E}=∗ False) ∧
              (if isEx then True else IW l γF t' s' vd ={E}=∗ False)) 𝓥.(cur) ∧
             (∀ v, (F_read γF t' s' v ={E}=∗ F_read γF t' s' v ∗ R' t' s' v) ∧
              (F γF t' s' v ={E}=∗ F γF t' s' v ∗ R' t' s' v) ∧
              (if isEx then True
               else IW l γF t' s' v ={E}=∗ IW l γF t' s' v ∗ R' t' s' v)) ∅)
        ∗ ▷ (gps_inv IP IW l γ γF bs Vo)
        ∗ (if isEx then exwr γ tx q ∗ ⌜(tx ≤ tr)%positive⌝ else True)
        ∗ Reader γ {[tr := (sr, vVr)]}
        ∗ naViewSeen l γ Va rs 𝓥.(cur) }}}
      Read AcqRel (Lit l) at 𝓥 @ E
    {{{ t' s' V' 𝓥' v, RET v at 𝓥';
        seen 𝓥'
        ∗ ⌜𝓥 ⊑ 𝓥' ∧ sr ⊑ s' ∧ (tr ≤ t')%positive ∧ seen_local l t' 𝓥'.(cur)
          ∧ V' ⊑ 𝓥'.(cur)⌝
        ∗ if (decide (v = vd)) then
            (R t' s' ∗ l ↦ v) V'
          else
             (P 𝓥'.(cur) ∗ R' t' s' v V'
             ∗ gps_inv IP IW l γ γF bs (Vo ⊔ 𝓥'.(cur))
             ∗ (if isEx then exwr γ tx q else True)
             ∗ ∃ Vt, Reader γ {[t' := (s', (v, Vt))]} ∧ ⌜Vt ⊔ Va = V'⌝)  }}}.
  Proof.
    iIntros (SUB Wv Rs Φ) "(#HInv & s & P & VP & Inv & ex & R & naV) Post".
    iDestruct "Inv" as (μ C tx' Va' rsa rs' ws)
                        "(>hist & >at & >aw & >nar & >oA & rBE & rD & >H)".
    iDestruct (naViewSeen_local_seen with "oA naV") as %(?&?&LeVaC&LNA&LNA2).
    subst Va' rs'. iClear "naV".
    iDestruct "H" as %(FINAL&SORTED&GOOD&[AT [AW JV]]&LeVa&NSL&IS&ND&HIW).
    iDestruct (gps_ownA_Reader_singleton_sub with "oA R") as %EQL.
    have Eqsr := toState_lookup_state _ _ _ _ _ EQL.
    destruct (toState_lookup_message _ _ _ _ _ EQL) as [mr [Eqr [Eqvr EqVr]]].

    iApply wp_fupd.
    iApply (iwp_read_atomic with "[$s $hist $at]"); [done| |done..|].
    { exists tr, mr. split; [done|]. split; [|done]. by inversion Eqvr. }

    iNext. iIntros (𝓥' v ?) "(s' & hist & at & FACTS)".
    iDestruct "FACTS" as % (Ext & WFC & t' & m' & Eqt' & Eqv' & RH & AT').

    have Eqs' := toState_lookup_state_inv μ _ _ _ _ Eqv' Eqt'.
    have Letr: tr ⊑ t'.
    { change (Some tr ⊑ Some t'). etrans; [apply Extr|by inversion RH]. }
    have Exts: sr ⊑ μ t'.
    { rewrite -Eqsr. apply SORTED; [..|done].
      - exists mr. split; [done|]. by inversion Eqvr.
      - exists m'. split; [done|]. by inversion Eqv'. }
    have ?: seen_local l t' 𝓥'.(cur) by eapply read_helper_seen_local; [|eauto].
    have Le': default ∅ (mrel m') ⊔ Va ⊑ 𝓥'.(cur).
    { apply lat_join_lub;
      [by eapply read_helper_view_acq_1|etrans; [apply LeVaC|apply Ext]]. }
    have ?: 𝓥.(cur) ⊔ (default ∅ (mrel m') ⊔ Va) ⊑ 𝓥'.(cur).
    { apply lat_join_lub; [by rewrite Ext|done]. }

    iDestruct (gps_ownA_split with "oA") as "[oA [oE naE]]".
    iMod (gps_ownA_M_fork_Reader_singleton _ _ _ _ Eqs' with "oA") as "[oA R']".
    case (decide (v = vd)) => Eqvd; [subst vd|]; last first.
    { iAssert (∃ Vt, Reader γ {[t' := (μ t', (v, Vt))]}
            ∧ ⌜Vt ⊔ Va = default ∅ (mrel m') ⊔ Va⌝)%I with "[R']" as "R'".
      { iExists _. by iFrame "R'". }
      iSpecialize ("VP" $! t' (μ t') with "[%]"); [done|]. rewrite bi.and_elim_r.
      iSpecialize ("VP" $! v).
      iMod (GPSRaw_read_resources IW isEx R' μ C q tx tx' tr t' m' v _ _ Va E
                Eqt' Eqv' Letr with "VP ex rBE rD oE") as "(rBE & rD & oE & ex & Rs)".
      iApply ("Post" $! t' (μ t') (default ∅ m'.(mrel) ⊔ Va)
              with "[- $s']"). iSplit; [done|].
      rewrite (decide_False _ _ Eqvd). iFrame "R' Rs P".
      iSplitR "ex"; last first.
      { destruct isEx; [|done]. by iDestruct "ex" as "[$ ?]". }
      iDestruct (gps_ownA_split with "[$oA $oE $naE]") as "oA".
      iExists μ, C, tx', Va, _, rs, ws. iFrame. iPureIntro.
      do 3 (split; [done|]). split; [|split; [solve_lat|done]].
      split; [done|]. split. { eapply atw_local_mono; [done| |done]. solve_lat. }
      move => ?? /JV [??]. split; [|solve_lat].
      eapply seen_local_mono; [|done]. solve_lat. }

    iSpecialize ("VP" $! t' (μ t') with "[%] P"); [done|].

    iAssert ((if isEx then exwr γ tx q ∗ ⌜(tx ≤ tr)%positive⌝ else True) ∗
              gps_ownA_E γ tx' ∗
             (F γF t' (μ t') v ∨ F_read γF t' (μ t') v ∨
              if isEx then False else IW l γF t' (μ t') v) (default ∅ (mrel m') ⊔ Va))%I
      with "[rBE rD ex oE]" as "[ex [oE F]]".
    { destruct ((cbends C) !! t') as [mm'|] eqn: Eqmm'.
      - have ? : mm' = m'. { move : Eqmm' => /map_filter_lookup_Some. rewrite Eqt'.
        move => [[<- //]]. } subst mm'.
        rewrite /resBE /gps_res (big_sepM_lookup _ _ _ _ Eqmm').
        destruct isEx; [|case decide => NEq].
        + iDestruct "ex" as "[ex %]";
          iDestruct (ExWrite_agree with "oE ex") as %?; subst tx'.
          inversion Eqv' as [? Eqv|]; (rewrite decide_True; [|by etrans]); by iFrame.
        + inversion Eqv'; by iFrame.
        + inversion Eqv'; by iFrame.
      - have Eqt2: (C ∖ (cbends C)) !! t' = Some m' by apply lookup_difference_Some.
        rewrite /resD /gps_res. rewrite (big_sepM_lookup _ _ _ _ Eqt2).
        destruct isEx; [|case decide => NEq].
        + iDestruct "ex" as "[ex %]";
          iDestruct (ExWrite_agree with "oE ex") as %?; subst tx'.
          inversion Eqv' as [? Eqv|]; (rewrite decide_True; [|by etrans]);
          iDestruct "rD" as "[$|$]"; by iFrame.
        + inversion Eqv'; by iDestruct "rD" as "[$|$]".
        + inversion Eqv'; by iDestruct "rD" as "[$|$]". }

    iSpecialize ("VP" with "[ex]"). { destruct isEx; [iExists _; by iFrame "ex"|done]. }
    rewrite 3!(monPred_mono _ _ _
                (lat_join_sqsubseteq_l 𝓥.(cur) (default ∅ (mrel m') ⊔ Va))).
    iDestruct "F" as "[F|[Fr|IW]]".
    - rewrite bi.and_elim_l. iMod ("VP" with "F") as "(mIn & W & Rs)".
      iDestruct "mIn" as %VoIn.
      iDestruct "W" as (ζ) "(Prt & W & MAX)". iDestruct "MAX" as %MAX.
      iDestruct (Writer_exact with "oA W") as %?. subst ζ.

      iInv histN as (σ) ">[Hσ ctx]" "HClose".
      iMod (hist_ctx_hist_drop_singleton _ _ C t' m' with "ctx [hist]")
        as "[ctx hist]"; [|by inversion Eqv'|by rewrite (insert_id _ _ _ Eqt')|].
      { move => t0 [m0 Eq0]. apply (MAX t0). specialize (ND _ _ Eq0).
        destruct m0.(mval) eqn:Eqv0; [|done|];
          eexists; (apply toState_lookup_state_inv; [|done]); rewrite Eqv0;
          by constructor. }
      iDestruct (hist_ctx_hist_good with "ctx hist") as %GH'.
      iMod ("HClose" with "[ctx Hσ]") as "_". { iExists _. by iFrame. }
      iModIntro.
      iApply ("Post" $! t' (μ t') (𝓥'.(cur) ⊔ (default ∅ (mrel m') ⊔ Va))).
      rewrite decide_True; [|done].
      iFrame "s' Rs". iSplit. { iPureIntro. do 4 (split; [done|]). solve_lat. }
      rewrite own_loc_na_eq.
      iExists t', m'. iSplit; last first.
      { iPureIntro. split; [done|]. destruct m'.(mrel); [|done]. solve_lat. }
      iExists _,_,_. iFrame "hist at aw nar". iPureIntro.
      split; last split; [|done|split]; last split.
      + exists t',m'. rewrite lookup_insert. repeat split; [by inversion Eqv'|].
        eapply seen_local_mono; [|done]. solve_lat.
      + simpl in AT'. eapply atr_local_mono; [done| |done]. solve_lat.
      + eapply atw_local_mono; [done| |done]. solve_lat.
      + eapply na_local_mono; [done| |apply LNA]. solve_lat.
    - rewrite bi.and_elim_r bi.and_elim_l. by iMod ("VP" with "Fr") as %[].
    - rewrite bi.and_elim_r bi.and_elim_r. destruct isEx.
      + by iDestruct "IW" as %[].
      + by iMod ("VP" with "IW") as %[].
  Qed.

  Lemma GPSRaw_Read_val_dealloc_ex
    IW (P : vProp) R R' vd q (tx tr: time) sr vVr γ γF bs Vo Va rs 𝓥 E
    (Extr : seen_local l tr 𝓥.(cur)) (LE: (tx ≤ tr)%positive):
    ↑histN ⊆ E →
    let Wv t s v γ : vProp :=
      (∃ ζ, PrtSeenView l t s v γ ∗ ⎡ Writer γ ζ ⎤ ∗
        ⌜∀ t', is_Some (ζ !! t') → t' ⊑ t⌝)%I in
    let Rs q t : vProp :=
      (∃ (tx: time), ⎡ exwr γ tx q ⎤ ∗ ⌜(tx ≤ t)%positive⌝)%I in
    {{{ hist_inv ∗ seen 𝓥 ∗ P 𝓥.(cur)
        ∗ ▷ (∀ t' s', ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
               (P -∗ Rs q tr -∗ (F γF t' s' vd ={E}=∗
                                 monPred_in Vo ∗ Wv t' s' vd γ ∗ R t' s') ∧
                           (F_read γF t' s' vd ={E}=∗ False)) 𝓥.(cur) ∧
               (∀ v, (F_read γF t' s' v ={E}=∗ F_read γF t' s' v ∗ R' t' s' v) ∧
                     (F γF t' s' v ={E}=∗ F γF t' s' v ∗ R' t' s' v)) ∅)
        ∗ ▷ (gps_inv IP IW l γ γF bs Vo)
        ∗ exwr γ tx q
        ∗ Reader γ {[tr := (sr, vVr)]}
        ∗ naViewSeen l γ Va rs 𝓥.(cur) }}}
      Read AcqRel (Lit l) at 𝓥 @ E
    {{{ t' s' V' 𝓥' v, RET v at 𝓥';
        seen 𝓥'
        ∗ ⌜𝓥 ⊑ 𝓥' ∧ sr ⊑ s' ∧ (tr ≤ t')%positive ∧ seen_local l t' 𝓥'.(cur)
          ∧ V' ⊑ 𝓥'.(cur)⌝
         ∗ if (decide (v = vd)) then
            (R t' s' ∗ l ↦ v) V'
          else
             (P 𝓥'.(cur) ∗ R' t' s' v V' ∗ gps_inv IP IW l γ γF bs (Vo ⊔ 𝓥'.(cur))
             ∗ exwr γ tx q
             ∗ ∃ Vt, Reader γ {[t' := (s', (v, Vt))]} ∧ ⌜Vt ⊔ Va = V'⌝) }}}.
  Proof.
    iIntros (SUB Wv Rs Φ) "(#HInv & s & P & VP & Inv & ex & R & naV) Post".
    iApply (GPSRaw_Read_val_dealloc IW true vd tx q P R R' tr sr _ γ γF bs Vo Va
            with "[$HInv $s VP Inv ex $P R naV] Post"); [done..|].
    iFrame "Inv ex R naV". iFrame (LE).
    iIntros "!>" (t' s' Ext). iSpecialize ("VP" $! t' s' Ext). iSplit.
    - rewrite bi.and_elim_l. iIntros (? ->) "P". iIntros (??) "ex".
      iSpecialize ("VP" with "P [$ex]"). iSplit.
      + by rewrite bi.and_elim_l.
      + rewrite bi.and_elim_r. by iSplit.
    - rewrite bi.and_elim_r. iIntros (v). iSpecialize ("VP" $! v).
      iSplit; last iSplit; last done. by rewrite bi.and_elim_l.
      by rewrite bi.and_elim_r.
  Qed.

  Lemma GPSRaw_Read_val_dealloc_non_ex
    IW (P: vProp) R R' vd (tr: time) sr vVr γ γF bs Vo Va rs 𝓥 E
    (Extr : seen_local l tr 𝓥.(cur)) :
    ↑histN ⊆ E →
    let Wv t s v γ : vProp :=
      (∃ ζ, PrtSeenView l t s v γ ∗ ⎡ Writer γ ζ ⎤ ∗
        ⌜∀ t', is_Some (ζ !! t') → t' ⊑ t⌝)%I in
    {{{ hist_inv ∗ seen 𝓥 ∗ P 𝓥.(cur)
        ∗ ▷ (∀ t' s', ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
                (P -∗ (F γF t' s' vd ={E}=∗
                    monPred_in Vo ∗ Wv t' s' vd γ ∗ R t' s') ∧
                (F_read γF t' s' vd ={E}=∗ False) ∧
                (IW l γF t' s' vd ={E}=∗ False)) 𝓥.(cur) ∧
               (∀ v,
                (F_read γF t' s' v ={E}=∗ F_read γF t' s' v ∗ R' t' s' v) ∧
                (F γF t' s' v ={E}=∗ F γF t' s' v ∗ R' t' s' v) ∧
                (IW l γF t' s' v ={E}=∗ IW l γF t' s' v ∗ R' t' s' v)) ∅)
        ∗ ▷ (gps_inv IP IW l γ γF bs Vo)
        ∗ Reader γ {[tr := (sr, vVr)]}
        ∗ naViewSeen l γ Va rs 𝓥.(cur) }}}
      Read AcqRel (Lit l) at 𝓥 @ E
    {{{ t' s' V' 𝓥' v, RET v at 𝓥';
        seen 𝓥'
        ∗ ⌜𝓥 ⊑ 𝓥' ∧ sr ⊑ s' ∧ (tr ≤ t')%positive ∧ seen_local l t' 𝓥'.(cur)
          ∧ V' ⊑ 𝓥'.(cur)⌝
        ∗ if (decide (v = vd)) then
            (R t' s' ∗ l ↦ v) V'
          else
            (P 𝓥'.(cur) ∗ R' t' s' v V' ∗ gps_inv IP IW l γ γF bs (Vo ⊔ 𝓥'.(cur))
             ∗ ∃ Vt, Reader γ {[t' := (s', (v, Vt))]} ∧ ⌜Vt ⊔ Va = V'⌝) }}}.
  Proof.
    iIntros (SUB Wv Φ) "(#HInv & s & P & VP & Inv & R & naV) Post".
    iApply (GPSRaw_Read_val_dealloc IW false vd 1 1%Qp P R R' tr sr _ γ γF
              bs Vo Va with "[$HInv $s $P VP Inv R naV]"); [done..| |].
    { iFrame "Inv R naV". iIntros "!>" (t' s' Ext).
      iSpecialize ("VP" $! t' s' Ext). iSplit.
      - rewrite bi.and_elim_l. iIntros (? ->) "P". iIntros (? ExtV) "_".
        iSpecialize ("VP" with "P"). rewrite ExtV. iSplit; last iSplit.
        + by rewrite bi.and_elim_l.
        + by rewrite bi.and_elim_r bi.and_elim_l.
        + by rewrite bi.and_elim_r bi.and_elim_r.
      - rewrite bi.and_elim_r. iIntros (v). iSpecialize ("VP" $! v).
        iSplit; last iSplit. by rewrite bi.and_elim_l.
        by rewrite bi.and_elim_r bi.and_elim_l.
        by rewrite bi.and_elim_r bi.and_elim_r. }
    iIntros "!>" (t' s' V' 𝓥' v) "(s' & FACTS & Rs)".
    iApply ("Post" $! t' s' V' 𝓥' v with "[$s' $FACTS Rs]").
    case decide => ?; [done|by iDestruct "Rs" as "($ & $ & $ & _ & $)"].
  Qed.

  (* SingleWriter reads *)
  Lemma GPSRaw_ReadSW
    IW (R: vProp) (tr: time) sr vr Vr o ζ γ γF q bs Vo Va rs 𝓥 E
    (Extr : seen_local l tr 𝓥.(cur)) (LE: Vr ⊑ 𝓥.(cur))
    (MAX: ∀ t', is_Some (ζ !! t') → t' ⊑ tr) :
    ↑histN ⊆ E → o = Relaxed ∨ o = AcqRel →
    {{{ seen 𝓥
        ∗ ▷ ((F γF tr sr vr ={E}=∗ F γF tr sr vr ∗ R) ∅)
        ∗ ▷ (gps_inv IP IW l γ γF bs Vo)
        ∗ Reader γ {[tr := (sr, (vr, Vr))]}
        ∗ naViewSeen l γ Va rs 𝓥.(cur)
        ∗ exwr γ tr q ∗ Writer γ ζ }}}
      Read o #l at 𝓥 @ E
    {{{ 𝓥', RET vr at 𝓥';
        seen 𝓥' ∗ ⌜𝓥 ⊑ 𝓥'⌝
        ∗ gps_inv IP IW l γ γF bs (Vo ⊔ 𝓥'.(cur))
        ∗ exwr γ tr q ∗ Writer γ ζ
        ∗ R 𝓥'.(cur) }}}.
  Proof.
    iIntros (SUB OR Φ) "(s & VP & Inv & R & naV & ex & W) Post".
    iDestruct "Inv" as (μ C tx Va' rsa rs' ws)
                        "(>hist & >at & >aw & >nar & >oA & rBE & rD & >H)".
    iDestruct (naViewSeen_local_seen with "oA naV") as %(?&?&LeVaC&LNA&LNA2).
    subst Va' rs'. iClear "naV".
    iDestruct "H" as %(FINAL&SORTED&GOOD&[AT [AW JV]]&LeVa&NSL&IS&ND&HIW).
    iDestruct (gps_ownA_Reader_singleton_sub with "oA R") as %EQL.
    have Eqsr := toState_lookup_state _ _ _ _ _ EQL.
    destruct (toState_lookup_message _ _ _ _ _ EQL) as [mr [Eqr [Eqvr EqVr]]].

    have ?: Relaxed ⊑ o by destruct OR; subst o.
    iApply wp_fupd.
    iApply (iwp_read_atomic with "[$s $hist $at]"); [done| |done..|].
    { exists tr, mr. split; [done|]. split; [|done]. by inversion Eqvr. }

    iNext. iIntros (𝓥' v ?) "(s' & hist & at & FACTS)".
    iDestruct "FACTS" as % (Ext & WFC & t' & m' & Eqt' & Eqv' & RH & AT').

    have Eqs' := toState_lookup_state_inv μ _ _ _ _ Eqv' Eqt'.
    iDestruct (gps_ownA_split with "oA") as "[oA [oE oN]]".
    iDestruct (ExWrite_agree with "oE ex") as %?. subst tx.
    iDestruct (Writer_exact with "oA W") as %?. subst ζ.
    iDestruct (gps_ownA_split with "[$oA $oE $oN]") as "oA".

    have ?: tr = t'.
    { apply : anti_symm.
      - change (Some tr ⊑ Some t'). etrans; first apply Extr. by inversion RH.
      - apply MAX. by eexists. } subst t'.
    have ?: vr = v. { rewrite EQL in Eqs'. by inversion Eqs'. } subst vr.
    have ?: mr = m'. { rewrite Eqr in Eqt'. by inversion Eqt'. } subst m'.
    clear Eqt'.

    (* the top write is certainly a block-end *)
    have HLa: (cbends C) !! tr = Some mr.
    { apply map_filter_lookup_Some. split; [done|]. rewrite /cell_adj.
      move => t0 m0 Eq0 /= NEq0 ?. subst t0.
      apply (irreflexive_eq (R:=(⊏)) tr tr); [done|].
      eapply (strict_transitive_l _ (tr+1)%positive).
      - split. apply Pos.lt_le_incl. lia. apply Pos.lt_nle. lia.
      - apply MAX. specialize (ND _ _ Eq0).
        destruct (m0.(mval)) as [| |v0] eqn:Eqv0; [|done|];
          eexists; (apply toState_lookup_state_inv; [|done]);
          rewrite Eqv0; by constructor. }

    rewrite (monPred_mono _ _ _ (lat_bottom_sqsubseteq (default ∅ (mrel mr) ⊔ Va))).
    iAssert (|={E}=> R (default ∅ (mrel mr) ⊔ Va) ∗
                     resBE IP IW l γF μ C tr Va)%I with "[VP rBE]" as ">[Rs rBE]".
    { rewrite /resBE /gps_res (big_sepM_delete _ _ _ _ HLa) Eqsr.
      iDestruct "rBE" as "(Hm & $)".
      inversion Eqv' as [? Eqvr'|Eqvr']; subst v; (rewrite decide_True; [|done]);
      by iMod ("VP" with "Hm") as "[$ $]". }

    iApply ("Post" with "[- $s' $ex $W]"). iSplit; [done|].
    iSplitR "Rs"; last first.
    { rewrite -(_: default ∅ mr.(mrel) ⊔ Va ⊑ 𝓥'.(cur)); [done|].
      apply lat_join_lub.
      - simpl in EqVr. by rewrite -EqVr LE Ext.
      - by rewrite LeVaC Ext. }
    iExists μ, C, tr, Va, _, rs, ws. iFrame "hist at aw nar oA rD rBE".
    iPureIntro. do 3 (split; [done|]). split; [|split; [solve_lat|done]].
    split; [done|].  split. { eapply atw_local_mono; [done| |done]. solve_lat. }
    move => ?? /JV [??]. split; [|solve_lat].
    eapply seen_local_mono; [|done]. solve_lat.
  Qed.

  Lemma resBE_full_impl IW C tx γF μ t' s' m' v Va
    (Lt : tx ⊏ t') (FRESH: C !! t' = None) (Eqv: memval_val_rel m'.(mval) v) :
    resBE IP IW l γF μ C tx Va -∗ resD IP IW l γF μ C tx Va -∗
    F γF t' s' v (default ∅ (mrel m') ⊔ Va) -∗
      resBE IP IW l γF (<[t':=s']> μ) (<[t':=m']> C) tx Va ∗
      resD IP IW l γF (<[t':=s']> μ) (<[t':=m']> C) tx Va.
  Proof.
    iIntros "rBE rD F".
    have Letx: tx ⊑ t' by apply strict_include.
    have INS := cbends_ins_update t' m' C FRESH.
    inversion INS; subst; simpl in *.
    - iSplitL "rBE".
      + rewrite /resBE /gps_res {2}/cbends -H4.
        iApply (big_sepM_mono with "rBE") => t0 ? /map_filter_lookup_Some [Eq ?]/=.
        rewrite fn_lookup_insert_ne; [done|] =>?. subst t0. by rewrite FRESH in Eq.
      + rewrite /resD /gps_res.
        iApply (big_sepM_subseteq _ (<[t':=m']> (C ∖ cbends C))).
        { rewrite {1}/cbends -H4. rewrite map_difference_insert_1; [done|].
          apply map_filter_lookup_None; by left. }
        rewrite big_sepM_insert; last by (apply lookup_difference_None; left).
        rewrite fn_lookup_insert. iSplitL "F".
        { inversion Eqv; rewrite /= (decide_True _ _ Letx). iFrame "F". by iLeft. }
        iApply (big_sepM_mono with "rD") => t0 ? /lookup_difference_Some [Eq _]/=.
        rewrite fn_lookup_insert_ne; [done|] =>?. subst t0. by rewrite FRESH in Eq.
    - have NEqk': k' ≠ t'.
      { move => ?. subst k'. by rewrite FRESH in In. }
      have IN: cbends C !! k' = Some v'.
      { apply map_filter_lookup_Some. split; [done|].
        move => t0 m0 Eq0 /= NEqt0. rewrite /cell_adj /= => Eqk'.
        rewrite /cell_adj /= -Eqk' in sL. subst t'. by rewrite FRESH in Eq0. }
      rewrite {1}/resBE /gps_res (big_sepM_delete _ _ _ _ IN).
      iDestruct "rBE" as "[Hm' rBE]". iSplitL "rBE F".
      + rewrite /resBE /gps_res {2}/cbends -H4.
        rewrite big_sepM_insert; last first.
        { rewrite lookup_delete_ne; [|done]. apply map_filter_lookup_None. by left. }
        rewrite fn_lookup_insert.
        iSplitL "F". { inversion Eqv; by rewrite /= (decide_True _ _ Letx). }
        iApply (big_sepM_mono with "rBE") => t0 ? /lookup_delete_Some
          [_ /map_filter_lookup_Some [Eq _]]. rewrite fn_lookup_insert_ne; [done|].
        move => ?. subst t0. by rewrite FRESH in Eq.
      + rewrite /resD /gps_res {2}/cbends -H4.
        iApply (big_sepM_subseteq _ (<[k':=v']> (C ∖ cbends C))).
        { etrans; [apply map_difference_insert_2|by rewrite -map_difference_delete]. }
        rewrite big_sepM_insert; last first.
        { apply lookup_difference_None. right. by eexists. }
        rewrite fn_lookup_insert_ne; [|done]. iSplitL "Hm'".
        { destruct v'.(mval); [|done|]; case decide => ?; by iFrame "Hm'". }
        iApply (big_sepM_mono with "rD") => t0 ? /lookup_difference_Some [Eq _]/=.
        rewrite fn_lookup_insert_ne; [done|] =>?. subst t0. by rewrite FRESH in Eq.
    - iSplitR "rD".
      + rewrite /resBE /gps_res /cbends -H4 big_sepM_insert; last first.
        { apply map_filter_lookup_None. by left. } iSplitL "F".
        { inversion Eqv; by rewrite /= fn_lookup_insert (decide_True _ _ Letx). }
        iApply (big_sepM_mono with "rBE") => t0 ? /map_filter_lookup_Some [Eq _]/=.
        rewrite fn_lookup_insert_ne; [done|] =>?. subst t0. by rewrite FRESH in Eq.
      + rewrite /resD /gps_res {2}/cbends -H4.
        iApply (big_sepM_subseteq _ (C ∖ cbends C)).
        { by apply map_difference_insert_2. }
        iApply (big_sepM_mono with "rD") => t0 ? /lookup_difference_Some [Eq _]/=.
        rewrite fn_lookup_insert_ne; [done|] =>?. subst t0. by rewrite FRESH in Eq.
    - have NEqk': k' ≠ t'.
      { move => ?. subst k'. by rewrite FRESH in In. }
      have IN: cbends C !! k' = Some v'.
      { apply map_filter_lookup_Some. split; [done|].
        move => t0 m0 Eq0 /= NEqt0. rewrite /cell_adj /= => Eqk'.
        rewrite /cell_adj /= -Eqk' in sL.  subst t'. by rewrite FRESH in Eq0. }
      rewrite {1}/resBE /gps_res (big_sepM_delete _ _ _ _ IN).
      iDestruct "rBE" as "[Hm' rBE]". iSplitL "rBE".
      + rewrite /resBE /gps_res {2}/cbends -H4.
        iApply (big_sepM_mono with "rBE") => t0 ? /lookup_delete_Some
          [_ /map_filter_lookup_Some [Eq _]]. rewrite fn_lookup_insert_ne; [done|].
        move => ?. subst t0. by rewrite FRESH in Eq.
      + rewrite /resD /gps_res {2}/cbends -H4.
        iApply (big_sepM_subseteq _ (<[t':=m']> (C ∖ delete k' (cbends C)))).
        { rewrite /cbends. rewrite map_difference_insert_1; [done|].
          apply lookup_delete_None. right. apply map_filter_lookup_None. by left. }
        rewrite big_sepM_insert; last first.
        { apply lookup_difference_None. by left. }
        rewrite fn_lookup_insert. iSplitL "F".
        { inversion Eqv; rewrite /= (decide_True _ _ Letx); by iFrame "F". }
        iApply (big_sepM_subseteq _ (<[k':=v']> (C ∖ cbends C))).
        { by rewrite -map_difference_delete. }
        rewrite big_sepM_insert; last first.
        { apply lookup_difference_None. right. by eexists. }
        rewrite fn_lookup_insert_ne; [|done]. iSplitL "Hm'".
        { destruct v'.(mval); [|done|]; case decide => ?; by iFrame "Hm'". }
        iApply (big_sepM_mono with "rD") => t0 ? /lookup_difference_Some [Eq _]/=.
        rewrite fn_lookup_insert_ne; [done|] =>?. subst t0. by rewrite FRESH in Eq.
  Qed.

  (** WRITE *)
  (* Normal protocol writes *)
  Lemma GPSRaw_Write
    IW (tr tx: time) sr vVr o v s' ζ q γ γF Vo Va rs 𝓥 E
    (Extr : seen_local l tr 𝓥.(cur))
    (FIN: final_st s') (LEx: (tx ≤ tr)%positive) :
    ↑histN ⊆ E → o = Relaxed ∨ o = AcqRel →
    let Vs V1 V2 := if decide (AcqRel ⊑ o) then V1 else V2 in
    {{{ seen 𝓥
        ∗ ▷ (gps_inv IP IW l γ γF false Vo)
        ∗ Reader γ {[tr := (sr, vVr)]}
        ∗ naViewSeen l γ Va rs 𝓥.(cur)
        ∗ exwr γ tx q ∗ Writer γ ζ }}}
      Write o #l v at 𝓥 @ E
    {{{ (t': time) (V': view) 𝓥', RET #☠ at 𝓥';
        seen 𝓥'
        ∗ ⌜𝓥 ⊑ 𝓥' ∧ 𝓥.(cur) !!w l ⊏ Some t' ∧ (tr < t')%positive
          ∧ seen_local l t' 𝓥'.(cur)
          ∧ seen_local l t' V'
          ∧ 𝓥.(rel) !! l ⊑ Some V'
          ∧ V' ⊑ default ∅ (𝓥'.(rel) !! l) ⊔ 𝓥'.(frel)
          ∧ Vs 𝓥'.(cur) 𝓥'.(frel) ⊑ V'⌝
        ∗ (F γF t' s' v (V' ⊔ Va) -∗ gps_inv IP IW l γ γF false (Vo ⊔ 𝓥'.(cur)))
        ∗ Reader γ {[t' := (s', (v, V'))]}
        ∗ exwr γ tx q ∗ Writer γ (<[t' := (s', (v, V'))]> ζ) }}}.
  Proof.
    iIntros (SUB OW Vs Φ) "(s & Inv & R & naV & ox & W) Post".
    iDestruct "Inv" as (μ C tx' Va' rsa rs' ws)
      "(>hist & >at & >aw & >nar & >oA & resBE & resD & >H)".
    iDestruct (naViewSeen_local_seen with "oA naV") as %(?&?&LeVaC&LNA&LNA2).
    subst Va' rs'. iClear "naV".
    iDestruct "H" as %(FINAL&SORTED&GOOD&[AT [AW JV]]&LeVa& NAL & IS & ND & _).

    iDestruct (gps_ownA_Reader_singleton_sub with "oA R") as %EQL.
    have Eqsr := toState_lookup_state _ _ _ _ _ EQL.
    destruct (toState_lookup_message _ _ _ _ _ EQL) as [mr [Eqr [Eqvr EqVr]]].

    have RLX: Relaxed ⊑ o by destruct OW; subst o.

    iApply wp_fupd.
    iApply (iwp_write_atomic with "s hist nar aw"); [done| |done..|].
    { exists tr, mr; by inversion Eqvr. }
    iIntros (𝓥' C' t' (Ext & WFC & m' & EqC' & Eqt' & Eqv' & ADD & WH))
            "!> !> !> s' hist' na aw". iIntros (ATW' WFC'). subst C'.

    iDestruct (Writer_exact_2 with "oA W") as %?. subst ζ.
    iDestruct (gps_ownA_split with "oA")  as "[oA [oE oN]]".
    iDestruct (ExWrite_agree with "oE ox") as %?. subst tx'.

    have HN: toState μ C !! t' = None by rewrite /toState lookup_imap Eqt'.
    iMod (Writer_update_insert_Reader _ _ _ _ (s', (v, default ∅ m'.(mrel))) HN
            with "oA W") as "(oA & W' & R')".

    have Ext': 𝓥.(cur) !!w l ⊏ Some t' by inversion WH.
    have Ext2: tr ⊏ t'.
    { change (Some tr ⊏ Some t').
      eapply strict_transitive_r; last apply Ext'.
      etrans; [apply Extr|by apply view_sqsubseteq]. }
    have ?: tx ⊏ t'. { eapply strict_transitive_r; last apply Ext2. done. }
    have ?: (tr < t')%positive. { by apply Pos.lt_nle, Ext2. }

    iApply ("Post" $! t' with "[- $s' $R' $ox $W']"). iSplit.
    { iPureIntro. split; [by solve_lat|].
      do 2 (split; [done|]).
      split; [by eapply write_helper_seen_local|].
      split; [by eapply write_helper_seen_local_write|].
      split; [|split].
      - inversion WH. rewrite /write_Rw decide_True // right_id_L.
        case ?:(𝓥.(rel) !! l); solve_lat.
      - inversion WH. rewrite lookup_insert /write_Rw decide_True // right_id_L.
        solve_lat.
      - destruct OW; subst o; simpl.
        + have ?:= (write_helper_relaxed_mrel_frel _ _ _ _ _ _ WH).
          by destruct m'.(mrel).
        + have ?:= write_helper_release_mrel_cur _ _ _ _ _ _ WH.
          by destruct m'.(mrel). }

    iIntros "F".
    iExists (<[t' := s']> μ), (<[t':=m']> C), tx, Va, rsa, rs, _.
    rewrite -toState_insert; last by rewrite Eqv'; constructor.
    rewrite gps_ownA_split. iFrame "hist' at aw na oA oE oN".

    iDestruct (resBE_full_impl with "resBE resD F") as "[$ $]"; [done|done|..].
    { by rewrite Eqv'; constructor. }

    iPureIntro. split; last split; [..|split; first done]; last split; last split;
      last split; last split.
    - move => t [m [Eqm Eqv]]. move : Eqm.
      case (decide (t = t')) => [->|?];
        [by rewrite lookup_insert fn_lookup_insert /=
        |rewrite lookup_insert_ne; last done].
      rewrite fn_lookup_insert_ne; last done.
      move => SOME NC. apply FINAL; first by eexists.
      destruct NC as (td & ? & ? & NE).
      exists td. repeat split; [done..|].
      case (decide (t' = td)) => ?; first by subst t'. move : NE.
      by rewrite lookup_insert_ne.
    - move => t1 t2 [m1 [Eq1 Eqv1]] [m2 [Eq2 Eqv2]].
      move : Eq1 Eq2.
      case (decide (t1 = t')) => [->|?];
        [rewrite lookup_insert fn_lookup_insert
        |rewrite lookup_insert_ne; last done;
         rewrite fn_lookup_insert_ne; last done];
      (case (decide (t2 = t')) => [->|?];
        [rewrite lookup_insert fn_lookup_insert
        |rewrite lookup_insert_ne; last done;
         rewrite fn_lookup_insert_ne; last done]); [done| |done|].
      + move => _ St2 Le2. apply FINAL; first by eexists.
        exists t'. repeat split; [by apply strict_include|done|by inversion ADD].
      + move => St1 St2. apply SORTED; by eexists.
    - eapply inv_view_mono; last by apply inv_view_insert_hist.
      rewrite -lat_join_assoc. apply lat_join_mono; [done|].
      clear -WH OW RLX. apply lat_join_lub.
      + inversion WH. rewrite /= (decide_True _ _ RLX). solve_lat.
      + have Ext : mrel m' ⊑ Some (cur 𝓥')
          by apply (write_helper_cur_tview_include _ _ _ _ _ _ _ WH).
        rewrite Ext. solve_lat.
    - solve_lat.
    - done.
    - rewrite (toState_insert _ _ _ _ v); last by rewrite Eqv'; constructor.
      rewrite lookup_insert_ne // => ?. subst tx.
      by apply (irreflexive_eq (R:= (⊏)) t' t').
    - split; [|done].
      move => t ?. case (decide (t = t')) => [->|?].
      + rewrite lookup_insert => [[<-]]. by rewrite Eqv'.
      + rewrite lookup_insert_ne; [|done]. by apply ND.
  Qed.

  (* SingleWriter writes *)
  Lemma GPSRaw_SWWrite_resources IW γF μ C (tx: time) sx vx Vx mx Va t' s' v m'
    (MAX: map_Forall (λ (t0: time) m0, (t0 ≤ tx)%positive) C)
    (Lttx: (tx < t')%positive)
    (Eqt' : C !! t' = None) (Eqv' : mval m' = VVal v)
    (ADD : cell_addins t' m' C (<[t':=m']> C))
    (Eqx : C !! tx = Some mx)
    (Eqvx : memval_val_rel (mval mx) vx)
    (EqVx : Vx = default ∅ (mrel mx))
    (Eqsx: sx = μ tx) :
    resBE IP IW l γF μ C tx Va -∗
    resD IP IW l γF μ C tx Va -∗
    (F γF t' s' v (default ∅ (mrel m') ⊔ Va) -∗
      (IW l γF tx sx vx) (Vx ⊔ Va) -∗
      resBE IP IW l γF (<[t':=s']> μ) (<[t':=m']> C) t' Va ∗
      resD IP IW l γF (<[t':=s']> μ) (<[t':=m']> C) t' Va) ∗
    F γF tx sx vx (Vx ⊔ Va).
  Proof.
    have In: (cbends C) !! tx = Some mx.
    { apply map_filter_lookup_Some. split; first done.
      move => t m Eqt NEqt. rewrite /cell_adj /= => Eqtx.
      apply (irreflexive_eq (R:= (Pos.lt)) t t); [done|].
      eapply Pos.le_lt_trans; [by apply (MAX _ m)|]. rewrite Eqtx. lia. }
    have ?: cbends C !! t' = None by apply map_filter_lookup_None; left.
    have NLe: ¬ (t' ≤ tx)%positive by apply Pos.lt_nle.
    have NEqx: t' ≠ tx. { move => ?. subst t'. by eapply (irreflexive_eq (R:= Pos.lt)). }
    have INS := cbends_ins_update t' m' C Eqt'.
    iIntros "rBE rD".
    rewrite /resBE /gps_res (big_sepM_delete _ _ _ _ In) EqVx.
    iDestruct "rBE" as "[F rBE]".
    iAssert (F γF tx sx vx (default ∅ (mrel mx) ⊔ Va))
      with "[F]" as "$". { rewrite Eqsx. by inversion Eqvx; rewrite decide_True. }
    iIntros "F IW".
    inversion INS; subst; simpl in *.
    - exfalso. destruct sR as [t0 [m0 [Eqt0 Eqm0]]].
      have Lt': (t' < t0)%positive.
      { rewrite /cell_adj /= in Eqm0. rewrite Eqm0. lia. }
      apply (irreflexive_eq (R:= Pos.lt) t' t'); [done|]. etrans; [exact Lt'|].
      eapply Pos.le_lt_trans; [by eapply MAX|exact Lttx].
    - rewrite /cell_adj /= in sL. clear nR.
      have Eqtx: k' = tx.
      { apply : anti_symm; [by eapply MAX|]. apply Pos.lt_succ_r.
        move : NLe => /Pos.lt_nle. rewrite sL. lia. } subst k'. clear In0 INS.
     iSplitL "rBE F".
      + iApply (big_sepM_subseteq _ (<[t' := m']> (delete tx (cbends C)))).
        { eapply cbends_ins_extend_subseteq; eauto. }
        rewrite big_sepM_insert; last first.
        { apply lookup_delete_None. by right. }
        rewrite Eqv' decide_True; [|done]. rewrite fn_lookup_insert. iFrame "F".
        iApply (big_sepM_mono with "rBE") => t0 m0 /lookup_delete_Some
          [NEq /map_filter_lookup_Some [Eq0 ?]] /=.
        rewrite fn_lookup_insert_ne; last first.
        { move => ?. subst t0. by rewrite Eqt' in Eq0. }
        have Lt0: (t0 < tx)%positive.
        { apply MAX, Pos.le_lteq in Eq0 as [?|?]; done. }
        have NLe1: ¬ (tx ≤ t0)%positive by apply Pos.lt_nle.
        have NLe2: ¬ (t' ≤ t0)%positive. { apply Pos.lt_nle. by etrans. }
        destruct (m0.(mval)); [|done|];
          by rewrite (decide_False _ _ NLe1) (decide_False _ _ NLe2).
      + rewrite /resD /gps_res.
        iApply (big_sepM_subseteq _ (<[tx := mx]> (C ∖ cbends C))).
        { rewrite {1}/cbends -H4.
          etrans; first apply map_difference_insert_2.
          by rewrite  -map_difference_delete. }
        rewrite big_sepM_insert; last first.
        { apply lookup_difference_None. right. by eexists. }
        iSplitL "IW".
        { rewrite fn_lookup_insert_ne; [|done].
          inversion Eqvx; rewrite (decide_False _ _ NLe); by iFrame "IW". }
        iApply (big_sepM_mono with "rD") => t0 m0 /lookup_difference_Some
          [Eq0 NEq0] /=.
        rewrite fn_lookup_insert_ne; last first.
        { move => ?. subst t0. by rewrite Eqt' in Eq0. }
        have Lt0: (t0 < tx)%positive.
        { apply MAX, Pos.le_lteq in Eq0 as [?|?]; [done|].
          subst t0. by rewrite In in NEq0. }
        have NLe1: ¬ (tx ≤ t0)%positive by apply Pos.lt_nle.
        have NLe2: ¬ (t' ≤ t0)%positive. { apply Pos.lt_nle. by etrans. }
        destruct (m0.(mval)); [|done|];
          by rewrite (decide_False _ _ NLe1) (decide_False _ _ NLe2).
    - iSplitL "rBE F IW".
      + iApply (big_sepM_subseteq _ (<[t':=m']> (cbends C))).
        { by apply mblock_ends_ins_fresh_mono. }
        rewrite big_sepM_insert; [|done].
        rewrite fn_lookup_insert Eqv' decide_True; [|done]. iFrame "F".
        rewrite (big_sepM_delete _ _ _ _ In). iSplitL "IW".
        { rewrite fn_lookup_insert_ne; [|done].
          inversion Eqvx; rewrite (decide_False _ _ NLe); by iFrame "IW". }
        iApply (big_sepM_mono with "rBE") => t0 m0 /lookup_delete_Some
          [NEq /map_filter_lookup_Some [Eq0 ?]] /=.
        rewrite fn_lookup_insert_ne; last first.
        { move => ?. subst t0. by rewrite Eqt' in Eq0. }
        have Lt0: (t0 < tx)%positive.
        { apply MAX, Pos.le_lteq in Eq0 as [?|?]; done. }
        have NLe1: ¬ (tx ≤ t0)%positive by apply Pos.lt_nle.
        have NLe2: ¬ (t' ≤ t0)%positive. { apply Pos.lt_nle. by etrans. }
        destruct (m0.(mval)); [|done|];
          by rewrite (decide_False _ _ NLe1) (decide_False _ _ NLe2).
      + rewrite /resD /gps_res /=.
        iApply (big_sepM_subseteq _ (C ∖ cbends C)).
        { rewrite {1}/cbends -H4. by apply map_difference_insert_2. }
        iApply (big_sepM_mono with "rD") => t0 m0 /lookup_difference_Some [Eq0 NEq0]/=.
        rewrite fn_lookup_insert_ne; last first.
        { move => ?. subst t0. by rewrite Eqt' in Eq0. }
        have Lt0: (t0 < tx)%positive.
        { apply MAX, Pos.le_lteq in Eq0 as [?|?]; [done|].
          subst t0. by rewrite In in NEq0. }
        have NLe1: ¬ (tx ≤ t0)%positive by apply Pos.lt_nle.
        have NLe2: ¬ (t' ≤ t0)%positive. { apply Pos.lt_nle. by etrans. }
        destruct (m0.(mval)); [|done|];
          by rewrite (decide_False _ _ NLe1) (decide_False _ _ NLe2).
    - exfalso. destruct sR as [t0 [m0 [Eqt0 Eqm0]]].
      have Lt': (t' < t0)%positive.
      { rewrite /cell_adj /= in Eqm0. rewrite Eqm0. lia. }
      apply (irreflexive_eq (R:= Pos.lt) t' t'); [done|]. etrans; [exact Lt'|].
      eapply Pos.le_lt_trans; [by eapply MAX|exact Lttx].
  Qed.

  Lemma GPSRaw_WriteSW IW
    (tx: time) sx vx Vx o v s' ζ γ γF Vo Va rs 𝓥 E
    (Extr : seen_local l tx 𝓥.(cur)) (Exts : sx ⊑ s')
    (MAXt: ∀ t', is_Some (ζ !! t') → t' ⊑ tx) :
    ↑histN ⊆ E → o = Relaxed ∨ o = AcqRel →
    let Vs V1 V2 := if decide (AcqRel ⊑ o) then V1 else V2 in
    {{{ seen 𝓥
        ∗ ▷ (gps_inv IP IW l γ γF true Vo)
        ∗ Reader γ {[tx := (sx, (vx, Vx))]}
        ∗ naViewSeen l γ Va rs 𝓥.(cur)
        ∗ exwr γ tx 1%Qp ∗ Writer γ ζ }}}
      Write o #l v at 𝓥 @ E
    {{{ (t': time) V' Vt' 𝓥', RET #☠ at 𝓥';
        seen 𝓥'
        ∗ ⌜𝓥 ⊑ 𝓥' ∧ 𝓥.(cur) !!w l ⊏ Some t' ∧ (tx < t')%positive
          ∧ seen_local l t' 𝓥'.(cur)
          ∧ seen_local l t' V'
          ∧ 𝓥'.(rel) !! l ⊑ Some V'
          ∧ Vs 𝓥'.(cur) 𝓥'.(frel) ⊑ V'
          ∧ Vt' ⊔ Va = V' ∧ Vt' ⊑ 𝓥'.(cur)⌝
        ∗ (F γF t' s' v V' -∗ IW l γF tx sx vx (Vx ⊔ Va) -∗
            gps_inv IP IW l γ γF true (Vo ⊔ 𝓥'.(cur)))
        ∗ (F γF tx sx vx) (Vx ⊔ Va)
        ∗ exwr γ t' 1%Qp
        ∗ Reader γ {[t' := (s', (v, Vt'))]}
        ∗ Writer γ (<[t' := (s', (v, Vt'))]> ζ)
        ∗ ⌜V' = Vt' ⊔ Va
           ∧ ∀ t0, is_Some (<[t' := (s', (v, Vt'))]> ζ !! t0) → t0 ⊑ t'⌝ }}}.
  Proof.
    iIntros (SUB OW Vs Φ) "(s & Inv & R & naV & ox & W) Post".
    iDestruct "Inv" as (μ C tx' Va' rsa rs' ws)
      "(>hist & >at & >aw & >nar & >oA & resBE & resD & >H)".
    iDestruct (naViewSeen_local_seen with "oA naV") as %(?&?&LeVaC&LNA&LNA2).
    subst Va' rs'. iClear "naV".
    iDestruct "H" as %(FINAL&SORTED&GOOD&[AT[AW JV]]&LeVa& NAL & IS & ND & RC).

    iDestruct (gps_ownA_Reader_singleton_sub with "oA R") as %EQL.
    have Eqsx := toState_lookup_state _ _ _ _ _ EQL.
    destruct (toState_lookup_message _ _ _ _ _ EQL) as [mx [Eqx [Eqvx EqVx]]].

    have RLX: Relaxed ⊑ o by destruct OW; subst o.

    iApply wp_fupd.
    iApply (iwp_write_atomic with "s hist nar aw"); [done| |done..|].
    { exists tx, mx; by inversion Eqvx. }
    iIntros (𝓥' C' t' (Ext & WFC & m' & EqC' & Eqt' & Eqv' & ADD & WH))
            "!> !> !> s' hist' nar aw". iIntros (AW' WFC'). subst C'.

    iDestruct (Writer_exact_2 with "oA W") as %?. subst ζ.
    iDestruct (gps_ownA_split with "oA") as "[oA [oE oN]]".
    iDestruct (ExWrite_agree with "oE ox") as %?. subst tx'.

    have HN: toState μ C !! t' = None by rewrite /toState lookup_imap Eqt'.
    iMod (Writer_update_insert_Reader _ _ _ _ (s', (v, default ∅ m'.(mrel))) HN
            with "oA W") as "(oA & W' & R')".

    have Ext': 𝓥.(cur) !!w l ⊏ Some t' by inversion WH.
    have Ext2: tx ⊏ t'.
    { change (Some tx ⊏ Some t'). eapply strict_transitive_r; [|apply Ext'].
       etrans; [apply Extr|by apply view_sqsubseteq]. }
    have ?: (tx < t')%positive. { by apply Pos.lt_nle, Ext2. }

    iMod (exwr_update _ _ t' with "oE ox") as "[oE ox]".

    iApply ("Post" $! t' (default ∅ m'.(mrel) ⊔ Va) (default ∅ m'.(mrel))
                      with "[- $s' $ox]").
    iSplit.
    { iPureIntro. split; [solve_lat|].
      do 2 (split; [done|]).
      split; [by eapply write_helper_seen_local|].
      split; [|split]; last split; last split; [..|done|].
      - apply (seen_local_mono _ _ _ _ (lat_join_sqsubseteq_l _ _)).
        by eapply write_helper_seen_local_write.
      - inversion WH. rewrite /write_Rw decide_True // right_id_L lookup_insert.
        solve_lat.
      - etrans; last by apply lat_join_sqsubseteq_l.
        destruct OW; subst o; simpl.
        + have ?:= (write_helper_relaxed_mrel_frel _ _ _ _ _ _ WH).
          by destruct m'.(mrel).
        + have ?:= write_helper_release_mrel_cur _ _ _ _ _ _ WH.
          by destruct m'.(mrel).
      - by rewrite (write_helper_cur_tview_include _ _ _ _ _ _ _ WH). }

    have MAX': map_Forall (λ (t0: time) m0, (t0 ≤ tx)%positive) C.
    { move => t m Eqt. apply MAXt.
      destruct (m.(mval)) as [| |v0] eqn:Eqv0; [|by apply ND in Eqt|];
        eexists; (apply toState_lookup_state_inv; [|done]); rewrite Eqv0; constructor. }

    iFrame "R' W'". simpl in EqVx, Eqvx.

    iDestruct (GPSRaw_SWWrite_resources IW γF μ C tx sx vx Vx mx Va t' s' v m'
                  MAX' with "resBE resD") as "[VS $]"; [done..|].
    iSplitR ""; last first.
    { iPureIntro. split; [solve_lat|].
      move => t0 [m0]. case (decide (t0 = t')) => [->|?]; [done|].
      rewrite lookup_insert_ne; last done. move => ?.
      etrans; [|apply strict_include, Ext2]. apply MAXt. by eexists. }

    iIntros "F IW". iExists (<[t' := s']> μ), (<[t':=m']> C), t', Va, rsa, rs, _.
    rewrite -toState_insert; [|by rewrite Eqv'; constructor].
    rewrite gps_ownA_split. iFrame "hist' at aw nar oA oE oN".
    iDestruct ("VS" with "F IW") as "[$ $]".

    iPureIntro. split; last split; [..|split; first done]; last split;
      last split; last split; last split; last split.
    - move => t [m [Eqm Eqv]]. move : Eqm.
      case (decide (t = t')) => [->|?];
        [rewrite lookup_insert fn_lookup_insert /=
        |rewrite lookup_insert_ne; last done].
      + move => ? [td [Le1 [Le2]]]. have ?: td = t' by apply : anti_symm.
        subst td. by rewrite lookup_insert.
      + rewrite fn_lookup_insert_ne; last done.
        move => SOME [t0 [Le1 [Le2 Eq0]]].
        exfalso. apply (irreflexive_eq (R:= (Pos.lt)) t t); [done|].
        eapply Pos.le_lt_trans; first apply MAXt.
        { destruct (mval m) eqn:EQ; [|done|];
            eexists; (apply toState_lookup_state_inv; last exact SOME);
            rewrite EQ; constructor. }
        eapply Pos.lt_le_trans; [eauto|by etrans].
    - move => t1 t2 [m1 [Eq1 Eqv1]] [m2 [Eq2 Eqv2]].
      move : Eq1 Eq2.
      case (decide (t1 = t')) => [->|?];
        [rewrite lookup_insert fn_lookup_insert
        |rewrite lookup_insert_ne; last done;
         rewrite fn_lookup_insert_ne; last done];
      (case (decide (t2 = t')) => [->|?];
        [rewrite lookup_insert fn_lookup_insert
        |rewrite lookup_insert_ne; last done;
         rewrite fn_lookup_insert_ne; last done]); [done| | |].
      + move => _ St2 Le2. apply FINAL; first by eexists.
        exists t'. repeat split; [by apply strict_include|done|by inversion ADD].
      + move => St1 _ Le1. etrans; last apply Exts. rewrite -Eqsx.
        apply SORTED; [by eexists|exists mx; by inversion Eqvx|..].
        apply MAXt. destruct (mval m1) eqn:EQ; [|done|];
          eexists; (apply toState_lookup_state_inv; last exact St1);
          rewrite EQ; constructor.
      + move => St1 St2. apply SORTED; by eexists.
    - eapply inv_view_mono; last by apply inv_view_insert_hist.
      rewrite -lat_join_assoc. apply lat_join_mono; [done|].
      clear -WH OW RLX. apply lat_join_lub.
      + inversion WH. rewrite /= (decide_True _ _ RLX). solve_lat.
      + have Ext : mrel m' ⊑ Some (cur 𝓥')
          by apply (write_helper_cur_tview_include _ _ _ _ _ _ _ WH).
        rewrite Ext. solve_lat.
    - solve_lat.
    - done.
    - rewrite (toState_insert _ _ _ _ v) ?lookup_insert; [by eauto|].
      rewrite Eqv'. constructor.
    - move => t ?. case (decide (t = t')) => [->|?].
      + rewrite lookup_insert => [[<-]]. by rewrite Eqv'.
      + rewrite lookup_insert_ne; [|done]. by apply ND.
    - move => t0 [m0 Eqt0] Let0.
      have ?: t' = t0.
      { apply : anti_symm; [done|]. case (decide (t0 = t')) => [->//|?].
        rewrite lookup_insert_ne in Eqt0; [|done]. apply strict_include.
        eapply strict_transitive_r; [by eapply MAX'|apply Ext2]. }
      subst t0. move => [td [? [? NIN]]].
      have ?: td = t' by apply : anti_symm. subst td. by rewrite NIN in Eqt0.
  Qed.

  (* CAS for single writer *)
  Lemma GPSRaw_CAS_resources (IW: loc → gname → time → Prtcl → val → vProp)
    γF μ C (tx: time) Va t' v' m' t s v m
    (Letx': (tx ≤ t')%positive) (Let': (t' < t)%positive)
    (Eqt: C !! t = None) (Eqt': C !! t' = Some m') (Eqm': cbends C !! t' = Some m')
    (Eqv : mval m = VVal v) (Eqv' : mval m' = VVal v') (ADJ: t = (t'+1)%positive) :
    gps_res (λ t v , if decide (tx ≤ t)%positive
                then IP true l γF t (μ t) v else IW l γF t (μ t) v)%I
            (delete t' (cbends C)) Va -∗
    resD IP IW l γF μ C tx Va -∗
    F γF t s v (default ∅ (mrel m) ⊔ Va) -∗
    F_read γF t' (μ t') v' (default ∅ (mrel m') ⊔ Va) -∗
      resBE IP IW l γF (<[t:=s]> μ) (<[t:=m]> C) tx Va ∗
      resD IP IW l γF (<[t:=s]> μ) (<[t:=m]> C) tx Va.
  Proof.
    iIntros "rBE rD F Fr".
    have ?: t' ≠ t. { move => ?. subst t. lia. }
    have HLN: delete t' (cbends C) !! t = None.
    { rewrite lookup_delete_ne; last done. by apply map_filter_lookup_None; left. }
    have Letx: (tx ≤ t)%positive by lia.
    have INS := cbends_ins_update t m C Eqt.
    inversion INS; subst; simpl in *.
    - exfalso. by apply (nL _ _ Eqt').
    - rewrite /cell_adj /= in sL. have ?: k' = t' by lia. subst k'.
      rewrite Eqt' in In. inversion In; subst v'0; clear In. iSplitL "rBE F".
      + rewrite /resBE /gps_res {2}/cbends -H4 big_sepM_insert; [|done].
        rewrite Eqv (decide_True _ _ Letx) fn_lookup_insert. iFrame "F".
        iApply (big_sepM_mono with "rBE") => t0 ? /lookup_delete_Some
          [_ /map_filter_lookup_Some [Eq _]].
        rewrite fn_lookup_insert_ne; [done|]=>?. subst t0. by rewrite Eqt in Eq.
      + rewrite /resD /gps_res.
        iApply (big_sepM_subseteq _ (<[t':=m']> (C ∖ cbends C))).
        { rewrite {1}/cbends -H4. etrans; [by apply map_difference_insert_2|].
          by rewrite -map_difference_delete. }
        rewrite big_sepM_insert; last first.
        { apply lookup_difference_None. right. by eexists. }
        rewrite Eqv' (decide_True _ _ Letx') fn_lookup_insert_ne; [|done].
        iSplitL "Fr"; [by iFrame "Fr"|].
        iApply (big_sepM_mono with "rD") => t0 ? /lookup_difference_Some [Eq _].
        rewrite fn_lookup_insert_ne; [done|]=>?. subst t0. by rewrite Eqt in Eq.
    - exfalso. by apply (nL _ _ Eqt').
    - rewrite /cell_adj /= in sL. have ?: k' = t' by lia. subst k'.
      rewrite Eqt' in In. inversion In; subst v'0; clear In. iSplitL "rBE".
      + rewrite /resBE /gps_res {2}/cbends -H4.
        iApply (big_sepM_mono with "rBE") => t0 ? /lookup_delete_Some
          [_ /map_filter_lookup_Some [Eq _]].
        rewrite fn_lookup_insert_ne; [done|]=>?. subst t0. by rewrite Eqt in Eq.
      + rewrite /resD /gps_res {2}/cbends -H4.
        rewrite (map_difference_delete _ m'); last by rewrite lookup_insert_ne.
        rewrite big_sepM_insert; last first.
        { apply lookup_difference_None. right. by eexists. }
        rewrite Eqv' (decide_True _ _ Letx') fn_lookup_insert_ne; [|done].
        iSplitL "Fr"; [by iFrame "Fr"|].
        rewrite map_difference_insert_1; last first.
        { apply map_filter_lookup_None. by left. }
        rewrite big_sepM_insert; last first.
        { apply lookup_difference_None. by left. }
        rewrite Eqv (decide_True _ _ Letx) fn_lookup_insert.
        iSplitL "F"; [by iFrame "F"|].
        iApply (big_sepM_mono with "rD") => t0 ? /lookup_difference_Some [Eq _].
        rewrite fn_lookup_insert_ne; [done|]=>?. subst t0. by rewrite Eqt in Eq.
  Qed.

  Lemma GPSRaw_CAS_int_write_view
    IW (isEx: bool) (tr tx: time) sr vVr orf or ow (vr: Z) (vw: val) γ γF q Vo Va rs 𝓥 E E'
    (P: vProp) (R : time → pr_stateT Prtcl → lit → vProp)
    (Extr : seen_local l tr 𝓥.(cur)) (SUB: E' ⊆ E) :
    ↑histN ⊆ E →
    orf = Relaxed ∨ orf = AcqRel →
    or = Relaxed ∨ or = AcqRel → ow = Relaxed ∨ ow = AcqRel →
    let Vs V1 V2 o := if decide (AcqRel ⊑ o) then V1 else V2 in
    let Wv t s v γ : vProp :=
      (∃ ζ, PrtSeenView l t s v γ ∗ ⎡ Writer γ ζ ⎤ ∗
        ⌜∀ t', is_Some (ζ !! t') → t' ⊑ t⌝)%I in
    let oEX := (if isEx then exwr γ tx q ∗ ⌜(tx ≤ tr)%positive⌝ else True)%I in
    {{{ seen 𝓥
        ∗ ▷ (∀ t' s' v, ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
              (if isEx then (F γF t' s' v ∨ F_read γF t' s' v)
               else (F γF t' s' v ∨ F_read γF t' s' v ∨ IW l γF t' s' v)) -∗
              ⌜∃ vl, v = #vl ∧ lit_comparable vr vl⌝) ∅
        ∗ (∀ t' s', ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
            (▷ if isEx then True else (P -∗ IW l γF t' s' #vr ={E}=∗ False)) ∧
            ∀ (v: lit), ⌜lit_neq vr v⌝ -∗
            ▷ <obj> ((F_read γF t' s' #v ={E}=∗ F_read γF t' s' #v ∗ R t' s' v) ∧
                      (F γF t' s' #v ={E}=∗ F γF t' s' #v ∗ R t' s' v) ∧
                      (if isEx then True else
                        IW l γF t' s' #v ={E}=∗ IW l γF t' s' #v ∗ R t' s' v)))
             (Vs 𝓥.(cur) 𝓥.(frel) ow)
        ∗ ▷ (gps_inv IP IW l γ γF true Vo)
        ∗ Reader γ {[tr := (sr, vVr)]}
        ∗ naViewSeen l γ Va rs 𝓥.(cur)
        ∗ oEX
        ∗ P (Vs 𝓥.(cur) 𝓥.(frel) ow) }}}
      CAS #l #vr vw orf or ow at 𝓥 @ E
    {{{ b (v': lit) s' (t': time) V' Vt' 𝓥', RET #b at 𝓥';
        seen 𝓥'
        ∗ ⌜𝓥 ⊑ 𝓥' ∧ sr ⊑ s' ∧ seen_local l t' 𝓥'.(cur) ∧ Vt' ⊔ Va = V'⌝
        ∗ oEX
        ∗ P (Vs 𝓥'.(cur) 𝓥'.(frel) ow)
        ∗ ( (∃ (t: time) Vt,
            ⌜b = true ∧ v' = LitInt vr ∧ (tr ≤ t)%positive ∧ (t < t')%positive
              ∧ Vt ⊑ Vt'
              ∧ 𝓥.(cur) !!w l ⊏ Some t'
              ∧ 𝓥.(rel) !! l ⊑ Some V'
              ∧ Vs 𝓥'.(cur) 𝓥'.(frel) ow ⊑ V'
              ∧ V' ⊑ Vs 𝓥'.(cur) 𝓥'.(acq) or⌝
            ∗ (Wv t s' #v' γ V' -∗ ∀ s'', ⌜s' ⊑ s''⌝ ={E}=∗
                Wv t' s'' vw γ V' ∗ Reader γ {[t' := (s'', (vw, Vt'))]} ∗
                (F γF t' s'' vw V' -∗ F_read γF t s' #v' (Vt ⊔ Va)
                    ={E}=∗ gps_inv IP IW l γ γF true (Vo ⊔ 𝓥'.(cur))))
            ∗ (F γF t s' #v') (Vt ⊔ Va)
            ∗ Reader γ {[t := (s', (#v', Vt))]})
        ∨ ( ⌜b = false ∧ lit_neq vr v' ∧ (tr ≤ t')%positive
              ∧ V' ⊑ Vs 𝓥'.(cur) 𝓥'.(acq) orf⌝
            ∗ gps_inv IP IW l γ γF true (Vo ⊔ 𝓥'.(cur))
            ∗ Reader γ {[t' := (s', (#v', Vt'))]}
            ∗ R t' s' v' (Vs 𝓥'.(cur) 𝓥'.(acq) orf)))
        }}}.
  Proof.
    iIntros (? ORF OR OW Vs Wv oEX Φ)
            "(s & Comp & VS & Inv & #R & naV & ox & P) Post".
    iDestruct "Inv" as (μ C tx' Va' rsa rs' ws)
                       "(>hist & >at & >aw & >nar & >oA & rBE & rD & >H)".
    iDestruct (naViewSeen_local_seen with "oA naV") as %(?&?&LeVaC&LNA&LNA2).
    subst Va' rs'. iClear "naV".
    iDestruct "H" as %(FINAL&SORTED&GOOD&[AT [AW JV]]&LeVa & NAL & IS & ND & RC).

    iDestruct (gps_ownA_Reader_singleton_sub with "oA R") as %EQL.
    have Eqsr := toState_lookup_state _ _ _ _ _ EQL.
    destruct (toState_lookup_message _ _ _ _ _ EQL) as [mr [Eqr [Eqvr EqVr]]].

    iDestruct (gps_ownA_split with "oA") as "[oA [oE #oN]]".

    iAssert (▷ ⌜∀ t m v, C !! t = Some m → 𝓥.(cur) !!w l ⊑ Some t
                → memval_val_rel m.(mval) v
                → ∃ vl : lit, v = #vl ∧ lit_comparable vr vl⌝)%I
    with "[rBE rD Comp ox oE]" as "#> COMP".
    { iNext. iIntros (t m v Eqt Le Eqv).
      have Eqs := toState_lookup_state_inv μ _ _ _ _ Eqv Eqt.
      have Letr: (tr ≤ t)%positive.
      { change (Some tr ⊑ Some t). by etrans; [apply Extr|apply Le]. }
      have Les: sr ⊑ μ t.
      { rewrite -Eqsr. apply SORTED; [exists mr; by inversion Eqvr
                                      |exists m; by inversion Eqv|done]. }
      rewrite (monPred_mono _ _ _ (lat_bottom_sqsubseteq (default ∅ m.(mrel) ⊔ Va))).
      iApply ("Comp" $! t (μ t) v with "[%] [-]"); [done|..].
      destruct ((cbends C) !! t) as [tm|] eqn:Eqtm.
      - rewrite /resBE /gps_res (big_sepM_lookup _ _ _ _ Eqtm).
        have ?: tm = m.
        { move : Eqtm => /map_filter_lookup_Some [Eqtm _]. rewrite Eqt in Eqtm.
          by inversion Eqtm. } subst tm.
        destruct isEx.
        + iDestruct "ox" as "[ox %]".
          iDestruct (ExWrite_agree with "oE ox") as %?. subst tx'.
          have Letx: (tx ≤ t)%positive by etrans.
          inversion Eqv; rewrite (decide_True _ _ Letx); by iFrame "rBE".
        + inversion Eqv; case decide => ?; by iFrame "rBE".
      - have Eqtm2: (C ∖ cbends C) !! t = Some m by apply lookup_difference_Some.
        rewrite /resD /gps_res (big_sepM_lookup _ _ _ _ Eqtm2). destruct isEx.
        + iDestruct "ox" as "[ox %]".
          iDestruct (ExWrite_agree with "oE ox") as %?. subst tx'.
          have Letx: (tx ≤ t)%positive by etrans.
          inversion Eqv; rewrite (decide_True _ _ Letx);
            by iDestruct "rD" as "[$|$]".
        + inversion Eqv; case decide => ?; iDestruct "rD" as "[$|$]". }
    iClear "Comp". iDestruct "COMP" as %COMP.

    have RRLX: Relaxed ⊑ or by destruct OR; subst or.
    have FRLX: Relaxed ⊑ orf by destruct ORF; subst orf.
    have WRLX: Relaxed ⊑ ow by destruct OW; subst ow.
    iApply wp_fupd.
    iApply (iwp_cas_int l vr vw orf or ow C _ _ _ _ _ _ E E'
              with "s hist at nar aw");
      [done|done|done|exists tr, mr; by inversion Eqvr|done..|].

    iIntros (b rd v' 𝓥' C' t' (WFC&m'&𝓥x&Eqt'&Eqv'&Ext1&Ext2&AT'&CCase)).

    have Eqv'2: memval_val_rel m'.(mval) #v' by rewrite Eqv'; constructor.
    have Eqs' := toState_lookup_state_inv μ _ _ _ _ Eqv'2 Eqt'.
    have Ext3: 𝓥 ⊑ 𝓥' by etrans.

    iMod (gps_ownA_M_fork_Reader_singleton _ _ _ _ Eqs' with "oA") as "[oA #R']".

    have ?: Vs 𝓥.(cur) 𝓥.(frel) ow ⊑ Vs 𝓥'.(cur) 𝓥'.(frel) ow.
      { rewrite /Vs. destruct OW; subst ow; simpl; by apply Ext3. }

    destruct CCase as [(? & ? & ? & ATW' & RH)
    |(? & RH & ? & m & EqC' & Eqt & Eqv
      & LeRel & Let1 & Let2 & ADD & HACQ & CUR & ATW' & WH)].
    - (* CAS fails *)
      have ?: tr ⊑ t'.
      { change (Some tr ⊑ Some t'). etrans; first apply Extr. by inversion RH. }
      have Exts: sr ⊑ μ t'. { rewrite -Eqsr. apply SORTED;
          [exists mr; by inversion Eqvr|by exists m'; rewrite Eqv'|done]. }

      iSpecialize ("VS" $! t' (μ t') with "[%]"); [done|].
      rewrite bi.and_elim_r. iSpecialize ("VS" $! v' with "[%]"); [done|].
      subst b C'.
      iApply step_fupd_intro; [done|]. iIntros "!> s' hist' at nar aw" (WFC').

      have LeV': default ∅ m'.(mrel) ⊔ Va ⊑ Vs 𝓥'.(cur) 𝓥'.(acq) orf.
      { destruct ORF; subst orf; simpl; (etrans; last by apply Ext2);
          apply lat_join_lub.
        - by eapply read_helper_view_relaxed.
        - etrans; [apply LeVaC|]. etrans; [apply cur_acq|apply Ext1].
        - by eapply read_helper_view_acq.
        - etrans; [apply LeVaC|apply Ext1]. }

      iAssert (|={E}=> R t' (μ t') v' (Vs 𝓥'.(cur) 𝓥'.(acq) orf) ∗
                        resBE IP IW l γF μ C tx' Va ∗
                        resD IP IW l γF μ C tx' Va ∗
                        oEX ∗ gps_ownA_E γ tx')%I
        with "[VS rBE rD ox oE]" as ">(Rd & rBE & rD & ox & oE)".
      { iSpecialize ("VS" $! (default ∅ (mrel m') ⊔ Va)).
        destruct ((cbends C) !! t') as [m2|] eqn:Eq2.
        - iFrame "rD".
          have ?: m2 = m'. { move : Eq2 => /map_filter_lookup_Some [Eq2 _].
              rewrite Eqt' in Eq2. by inversion Eq2. } subst m2.
          rewrite bi.and_elim_r /resBE /gps_res (big_sepM_delete _ _ _ _ Eq2) Eqv'.
          iDestruct "rBE" as "[F $]". destruct isEx.
          + iDestruct "ox" as "[ox %]".
            iDestruct (ExWrite_agree with "oE ox") as %?. subst tx'.
            iFrame "oE ox". have Letx: (tx ≤ t')%positive by etrans.
            rewrite (decide_True _ _ Letx) bi.and_elim_l.
            by iMod ("VS" with "F") as "[$ $]".
          + iFrame "ox oE".
            case decide => ?; [rewrite bi.and_elim_l|rewrite bi.and_elim_r];
              by iMod ("VS" with "F") as "[$ $]".
        - iFrame "rBE".
          have Eqm2: (C ∖ cbends C) !! t' = Some m'
            by apply lookup_difference_Some.
          rewrite /resD /gps_res (big_sepM_delete _ _ _ _ Eqm2) Eqv'.
          iDestruct "rD" as "[F $]". destruct isEx.
          + iDestruct "ox" as "[ox %]".
            iDestruct (ExWrite_agree with "oE ox") as %?. subst tx'.
            iFrame "oE ox". have Letx: (tx ≤ t')%positive by etrans.
            rewrite (decide_True _ _ Letx).
            iDestruct "F" as "[F|F]"; [rewrite bi.and_elim_r|rewrite bi.and_elim_l];
              by iMod ("VS" with "F") as "[$ $]".
          + iFrame "ox oE". case decide => ?.
            * iDestruct "F" as "[F|F]";
                [rewrite bi.and_elim_r bi.and_elim_l|rewrite bi.and_elim_l];
                by iMod ("VS" with "F") as "[$ $]".
            * iDestruct "F" as "[F|[F|F]]";
                [rewrite bi.and_elim_r bi.and_elim_l|rewrite bi.and_elim_l
                |rewrite bi.and_elim_r bi.and_elim_r];
                by iMod ("VS" with "F") as "[$ $]". }

      iDestruct (gps_ownA_split with "[$oA $oE $oN]") as "oA".

      iApply ("Post" $! _ v' (μ t') t'
                (default ∅ m'.(mrel) ⊔ Va) (default ∅ m'.(mrel))
                with "[- $s']"). iSplit.
      { iPureIntro. split; [done|]. split; [by etrans|].
        split; [|done]. eapply seen_local_mono; [by apply Ext2|].
        eapply read_helper_seen_local; [|eauto]. done. }

      iFrame "ox P". iRight. iFrame "Rd R'". iSplit; [done|].
      iExists μ, C, tx', Va, _,_,_. iFrame.
      iPureIntro. do 3 (split; [done|]). split; [|split; [solve_lat|done]].
      do 2 (split; [done|]). move => ?? /JV [??]. split; [|solve_lat].
      eapply seen_local_mono; [|done]. solve_lat.

    - (* CAS succeeds *)
      have ?: tr ⊑ t'.
      { change (Some tr ⊑ Some t'). etrans; first apply Extr. by inversion RH. }
      have Exts: sr ⊑ μ t'.
      { rewrite -Eqsr. apply SORTED;
            [exists mr; by inversion Eqvr|by exists m'; rewrite Eqv'|done]. }

      iSpecialize ("VS" $! t' (μ t') with "[%//]"). rewrite bi.and_elim_l.
      subst b v'.

      (* get full interpretation *)
      have Eqm': cbends C !! t' = Some m'.
      { apply map_filter_lookup_Some. split; [done|].
        move => /= t1 m1 Eqm1 NEqt1 Eqf. rewrite /cell_adj /= in Eqf.
        by rewrite Eqf Eqt in Eqm1. }

      rewrite /resBE /gps_res (big_sepM_delete _ _ _ _ Eqm')
              Eqv' /=. iDestruct "rBE" as "(F & rBE)".

      have Ext4: Vs 𝓥.(cur) 𝓥.(frel) ow ⊑ default ∅ m.(mrel) ⊔ Va.
      { etrans; last apply lat_join_sqsubseteq_l.
        destruct OW; subst ow; simpl; (etrans; first apply Ext1).
        - have ExtR:= (write_helper_relaxed_mrel _ _ _ _ _ _ WH).
          etrans; last by eapply default_view_bot. solve_lat.
        - have ExtR:= (write_helper_release_mrel _ _ _ _ _ _ WH).
          etrans; last by eapply default_view_bot. solve_lat. }
      have Ext5: default ∅ m'.(mrel) ⊔ Va ⊑ default ∅ m.(mrel) ⊔ Va
        by rewrite LeRel.
      have Ext5': Vs 𝓥.(cur) 𝓥.(frel) ow ⊔ (default ∅ m'.(mrel) ⊔ Va)
                  ⊑ default ∅ m.(mrel) ⊔ Va by apply lat_join_lub.
      have Ext6: default ∅ m.(mrel) ⊔ Va ⊑ Vs (cur 𝓥') (acq 𝓥') or.
      { rewrite /Vs. destruct OR; subst or; simpl; apply lat_join_lub.
        - apply default_view_bot in HACQ. by rewrite HACQ.
        - etrans; [apply LeVaC|]. etrans; [apply cur_acq|apply Ext3].
        - apply default_view_bot in CUR. by rewrite CUR.
        - etrans; [apply LeVaC|apply Ext3]. }

      set t:= (t'+1)%positive.
      iApply step_fupd_intro; [done|].
      iIntros "!> s' hist' at nar aw" (WFC').
      iApply ("Post" $! _ vr (μ t') t
                        (default ∅ m.(mrel) ⊔ Va) (default ∅ m.(mrel))).

      iAssert (|={E}=> (F γF t' (μ t') #vr) (default ∅ (mrel m') ⊔ Va) ∗
                        oEX ∗ gps_ownA_E γ tx' ∗
                        P (Vs 𝓥.(cur) 𝓥.(frel) ow))%I
          with "[F VS oE ox P]" as ">(F & ox & oE & P)".
      { case decide => Lt; [by iFrame|]. destruct isEx.
        - iDestruct "ox" as "[ox %]".
          iDestruct (ExWrite_agree with "oE ox") as %?. subst tx'.
          exfalso. apply Lt. by etrans.
        - iFrame "ox oE".
          rewrite (monPred_mono _ _ _
            (lat_join_sqsubseteq_r (default ∅ m'.(mrel) ⊔ Va) (Vs 𝓥.(cur) 𝓥.(frel) ow))).
          by iMod ("VS" with "P F"). }

      have Let': (t' < t)%positive by rewrite /t; lia.

      iModIntro.
      iFrame "s'". iSplit; [done|]. iFrame "ox P".
      iLeft; iExists t', (default ∅ m'.(mrel)). iSplit.
      { iPureIntro. do 4 (split; [done|]). split; [by rewrite LeRel|].
        split; [done|]. split; last split; [..|done].
        - inversion WH. rewrite /write_Rw decide_True; [|by destruct OW; subst ow].
          etrans; first apply Ext1.
          destruct (𝓥x.(rel) !! l) eqn:HREL; last done. solve_lat.
        - etrans; last by apply lat_join_sqsubseteq_l.
          rewrite /Vs. destruct OW; subst ow; simpl.
          + have ?:= (write_helper_relaxed_mrel_frel _ _ _ _ _ _ WH).
            by destruct m.(mrel).
          + have ?:= write_helper_release_mrel_cur _ _ _ _ _ _ WH.
            by destruct m.(mrel). }

      have Eqv2 : memval_val_rel m.(mval) vw by rewrite Eqv; constructor.
      iFrame "R' F".

      iDestruct 1 as (ζ') "(#Prt & W & MAX)". iDestruct "MAX" as %MAX.
      iIntros (s Exts').
      iDestruct (Writer_exact with "oA W") as %?. subst ζ'.

      (* update Writer and get new Reader *)
      have HN: toState μ C !! t = None by rewrite /toState lookup_imap Eqt.
      iMod (Writer_update_insert_Reader _ _ _ _ (s, (vw, default ∅ m.(mrel)))
                                        HN with "oA W") as "(oA & W' & #R'')".

      iAssert ((PrtSeenView l t s vw γ) (default ∅ (mrel m) ⊔ Va))%I
        with "[Prt]" as "#Prt2".
      { iExists (default ∅ m.(mrel)).
        iDestruct "Prt" as (? Va' ?) "(na & ?)".
        iExists Va',_. iFrame "na R''". iPureIntro. split; [|solve_lat].
        eapply seen_local_mono; [apply lat_join_sqsubseteq_l|].
        eapply write_helper_seen_local_write; last exact WH.
        by destruct OW; subst ow. }

      iSplitL "W'".
      { iExists _. iFrame "Prt2 W'".
        iPureIntro. move => t1. case (decide (t1 = t)) => [->//|?].
        rewrite lookup_insert_ne; [|done]. move => /MAX Le.
        etrans; [apply Le|]. rewrite /t. apply Pos.lt_le_incl. lia. }
      iFrame "R''".

      iIntros "!> F Fr".
      iExists (<[t := s]> μ), (<[t:=m]> C), tx', Va, _, rs, (ws ∪ {[t]}).
      rewrite -toState_insert; [|done].
      rewrite gps_ownA_split EqC'. iFrame "hist' at aw nar oA oE oN".

      have Letx: (tx' ≤ t')%positive by apply MAX.
      have ?: (tx' < t)%positive by eapply Pos.le_lt_trans.

      iDestruct (GPSRaw_CAS_resources IW γF μ C tx' Va t' #vr m' t s vw m
                  Letx Let' with "rBE rD F Fr") as "[$ $]"; [done..|].

      iPureIntro. split; last split; last split; last split; last split;
        last split; last split; last split.
      + move => t0 [m0 [Eqm0 Eqv0]]. move : Eqm0.
        case (decide (t0 = t)) => [->|?];
          [rewrite lookup_insert fn_lookup_insert /=
          |rewrite lookup_insert_ne; last done]. (* TODO: duplicated proofs *)
        { move => _ [td [Ld1 [Ld2 DISJ]]] s0. etrans; last apply Exts'.
          apply FINAL; [by exists m'; rewrite Eqv'|].
          have Lt: (td < t)%positive.
          { apply Pos.le_lteq in Ld2 as [Lt2|];
              [done|subst td; by rewrite lookup_insert in DISJ]. }
          have Led: (td ≤ t')%positive.
          { apply Pos.lt_succ_r. clear -Lt. move : Lt. rewrite /t. lia. }
          exists td. repeat split; [done..|].
          rewrite lookup_insert_ne in DISJ; [done|].
          move => ?. subst t. lia. }
        { move => Eqt0 [td [? [? DISJ]]].
          rewrite/final_st fn_lookup_insert_ne; [|done].
          apply FINAL; [by eexists|]. exists td. repeat split; [done..|].
          move : DISJ. case (decide (t=td)) => [->|?]. by rewrite lookup_insert.
          by rewrite lookup_insert_ne. }
      + move => t1 t2 [m1 [Eq1 ?]] [m2 [Eq2 ?]]. move : Eq1 Eq2.
        case (decide (t1 = t)) => [->|NEq];
          [rewrite lookup_insert fn_lookup_insert
          |rewrite lookup_insert_ne; last done;
           rewrite fn_lookup_insert_ne; last done];
        (case (decide (t2 = t)) => [->|?];
          [rewrite lookup_insert fn_lookup_insert
          |rewrite lookup_insert_ne; last done;
           rewrite fn_lookup_insert_ne; last done]); [done| | |].
        * move => _ St2 Le2. apply FINAL; first by eexists.
          exists t. repeat split; [by apply Pos.lt_le_incl|done|by inversion ADD].
        * move => St1 _ Le1. etrans; last apply Exts'.
          apply SORTED; [by eexists|exists m'; by rewrite Eqv'|].
          case (decide (t1 ≤ t')%positive) => [//|/Pos.lt_nle Lt].
          exfalso. apply NEq. apply : anti_symm; [done|]. rewrite /t. clear -Lt.
          rewrite Pos.add_1_r. by apply Pos.le_succ_l.
        * move => St1 St2. apply SORTED; by eexists.
      + by rewrite -EqC'.
      + eapply inv_view_mono; last apply inv_view_insert_hist; [|done|]; last first.
        * split; [done|].
          split. { eapply atw_local_mono; [done| |done]. solve_lat. }
          move => ?? /JV [??].  split; [|solve_lat].
          eapply seen_local_mono; [|done]. solve_lat.
        * clear -WH JV Eqt' WRLX.
          apply lat_join_lub; [apply lat_join_lub|]; [done|..].
          { inversion WH. rewrite /= (decide_True _ _ WRLX). solve_lat. }
          { apply JV in Eqt' as [_ Eqt']. inversion_clear WH.
            rewrite writeRw_included Eqt' /= (decide_True _ _ WRLX). solve_lat. }
      + solve_lat.
      + done.
      + rewrite (toState_insert _ _ _ _ _ _ Eqv2) lookup_insert_ne; [done|].
        move => ?. subst tx'. by apply (irreflexive_eq (R:= (<)%positive) t t).
      + move => t0 m0. case (decide (t0 = t)) => [->|?].
        rewrite lookup_insert => [[<-]]. by rewrite Eqv.
        rewrite lookup_insert_ne; [|done]. by apply ND.
      + move => t0 [m0].
        case (decide (t0 = t)) => [->|?];
          [rewrite lookup_insert|rewrite lookup_insert_ne; [|done]].
        * move => _ Let3 [td [Le4 [/Pos.le_lteq [Le5|Le5] NIN]]];
            last by rewrite Le5 lookup_insert in NIN.
          rewrite lookup_insert_ne in NIN; last first.
          { move => ?; subst td; lia. }
          apply (RC t'); [by eexists|done|]. exists td. repeat split; auto.
          apply Pos.lt_succ_r. by rewrite -Pos.add_1_r.
        * move => Eqt0 Let0 [td [? [? NIN]]]. apply (RC t0); [by eexists|done|].
          exists td. repeat split; [done..|].
          case (decide (td = t)) => ?;
            [subst td; by rewrite lookup_insert in NIN
            |by rewrite lookup_insert_ne in NIN].
  Qed.

  Lemma GPSRaw_CAS_int_write
    IW (isEx dropR: bool) (tr tx: time) q sr vVr orf or ow (vr: Z) (vw: val)
    γ γF Vo Va rs 𝓥 E E'
    (P: vProp) (Q Q1 Q2: time → pr_stateT Prtcl → vProp)
    (R : time → pr_stateT Prtcl → lit → vProp)
    (Extr : seen_local l tr 𝓥.(cur)) (SUB: E' ⊆ E) :
    ↑histN ⊆ E →
    orf = Relaxed ∨ orf = AcqRel →
    or = Relaxed ∨ or = AcqRel → ow = Relaxed ∨ ow = AcqRel →
    let Vs V1 V2 o := if decide (AcqRel ⊑ o) then V1 else V2 in
    let Wv t s v γ : vProp :=
      (∃ ζ, PrtSeenView l t s v γ ∗ ⎡ Writer γ ζ ⎤ ∗
            ⌜∀ t', is_Some (ζ !! t') → t' ⊑ t⌝)%I in
    let Rs t s v : vProp :=
      (∃ (tx: time), PrtSeen l t s v γ ∗ ⎡ exwr γ tx q ⎤ ∗ ⌜(tx ≤ t)%positive⌝)%I in
    let oEX := (if isEx then exwr γ tx q ∗ ⌜(tx ≤ tr)%positive⌝ else True)%I in
    let VS t' s' :=
      ((<obj> (▷ F γF t' s' #vr ={E}=∗ ▷ Q1 t' s' ∗ ▷ Q2 t' s')) ∗
             (P -∗ ▷ Q2 t' s' ={E}=∗ ▷ Wv t' s' #vr γ ∗
                ∃ s'', ⌜s' ⊑ s''⌝ ∗
                  (∀ t , ⌜(t' < t)%positive⌝ -∗ ▷ Wv t s'' vw γ -∗
                      (if isEx && dropR then ▷ Rs t s'' vw else True) ={E}=∗
                      (<obj> (▷ Q1 t' s' ={E}=∗ ▷ F_read γF t' s' #vr)) ∗
                      |={E,E'}▷=> Q t s'' ∗ F γF t s'' vw)))%I in
    {{{ seen 𝓥 ∗ P (Vs 𝓥.(cur) 𝓥.(frel) ow) ∗
        ▷ (∀ t' s' v, ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
            (if isEx then (F γF t' s' v ∨ F_read γF t' s' v)
             else (F γF t' s' v ∨ F_read γF t' s' v ∨ IW l γF t' s' v)) -∗
            ⌜∃ vl, v = #vl ∧ lit_comparable vr vl⌝) ∅ ∗
        (∀ t' s', ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
            ((▷ if isEx then True else (P -∗ IW l γF t' s' #vr ={E}=∗ False)) ∧
             VS t' s') ∧
            (▷ ∀ (v: lit), ⌜lit_neq vr v⌝ -∗
              <obj> ((F_read γF t' s' #v ={E}=∗ F_read γF t' s' #v ∗ R t' s' v) ∧
                     (F γF t' s' #v ={E}=∗ F γF t' s' #v ∗ R t' s' v) ∧
                     if isEx then True else
                     (IW l γF t' s' #v ={E}=∗ IW l γF t' s' #v ∗ R t' s' v))))
             (Vs 𝓥.(cur) 𝓥.(frel) ow)
        ∗ ▷ (gps_inv IP IW l γ γF true Vo)
        ∗ Reader γ {[tr := (sr, vVr)]}
        ∗ naViewSeen l γ Va rs 𝓥.(cur)
        ∗ oEX }}}
      CAS #l #vr vw orf or ow at 𝓥 @ E
    {{{ b (v': lit) s' (t': time) V' Vt' 𝓥', RET #b at 𝓥';
        seen 𝓥'
        ∗ ⌜𝓥 ⊑ 𝓥' ∧ sr ⊑ s' ∧ seen_local l t' 𝓥'.(cur) ∧ Vt' ⊔ Va = V'⌝
        ∗ gps_inv IP IW l γ γF true (Vo ⊔ 𝓥'.(cur))
        ∗ ( (⌜b = true ∧ v' = LitInt vr ∧ (tr < t')%positive
              ∧ 𝓥.(cur) !!w l ⊏ Some t'
              ∧ 𝓥.(rel) !! l ⊑ Some V'
              ∧ Vs 𝓥'.(cur) 𝓥'.(frel) ow ⊑ V'
              ∧ V' ⊑ Vs 𝓥'.(cur) 𝓥'.(acq) or⌝
            ∗ (if dropR then True else oEX)
            ∗ Reader γ {[t' := (s', (vw, Vt'))]}
            ∗ Q t' s' (Vs 𝓥'.(cur) 𝓥'.(acq) or))
        ∨ ( ⌜b = false ∧ lit_neq vr v' ∧ (tr ≤ t')%positive 
              ∧ V' ⊑ Vs 𝓥'.(cur) 𝓥'.(acq) orf⌝
            ∗ oEX
            ∗ Reader γ {[t' := (s', (#v', Vt'))]}
            ∗ R t' s' v' (Vs 𝓥'.(cur) 𝓥'.(acq) orf)
            ∗ P (Vs 𝓥'.(cur) 𝓥'.(frel) ow) ))
        }}}.
  Proof.
    iIntros (? ORF OR OW Vs Wv Rs oEX VS Φ)
            "(s & P & Comp & VS & Inv & #R & naV & ox) Post".
    iDestruct "Inv" as (μ C tx' Va' rsa rs' ws)
                      "(>hist & >at & >aw & >nar & >oA & rBE & rD & >H)".
    iDestruct (naViewSeen_local_seen with "oA naV") as %(?&?&LeVaC&LNA&LNA2).
    subst Va' rs'. iClear "naV".
    iDestruct "H" as %(FINAL&SORTED&GOOD&[AT [AW JV]]&LeVa & NAL & IS & ND & RC).

    iDestruct (gps_ownA_Reader_singleton_sub with "oA R") as %EQL.
    have Eqsr := toState_lookup_state _ _ _ _ _ EQL.
    destruct (toState_lookup_message _ _ _ _ _ EQL) as [mr [Eqr [Eqvr EqVr]]].

    iDestruct (gps_ownA_split with "oA") as "[oA [oE #oN]]".

    iAssert (▷ ⌜∀ t m v, C !! t = Some m → 𝓥.(cur) !!w l ⊑ Some t
                → memval_val_rel m.(mval) v
                → ∃ vl : lit, v = #vl ∧ lit_comparable vr vl⌝)%I
    with "[rBE rD Comp ox oE]" as "#> COMP".
    { iNext. iIntros (t m v Eqt Le Eqv).
      have Eqs := toState_lookup_state_inv μ _ _ _ _ Eqv Eqt.
      have Letr: (tr ≤ t)%positive.
      { change (Some tr ⊑ Some t). by etrans; [apply Extr|apply Le]. }
      have Les: sr ⊑ μ t.
      { rewrite -Eqsr. apply SORTED; [exists mr; by inversion Eqvr
                                      |exists m; by inversion Eqv|done]. }
      rewrite (monPred_mono _ _ _ (lat_bottom_sqsubseteq (default ∅ m.(mrel) ⊔ Va))).
      iApply ("Comp" $! t (μ t) v with "[%] [-]"); [done|..].
      destruct ((cbends C) !! t) as [tm|] eqn:Eqtm.
      - rewrite /resBE /gps_res (big_sepM_lookup _ _ _ _ Eqtm).
        have ?: tm = m.
        { move : Eqtm => /map_filter_lookup_Some [Eqtm _]. rewrite Eqt in Eqtm.
          by inversion Eqtm. } subst tm.
        destruct isEx.
        + iDestruct "ox" as "[ox %]".
          iDestruct (ExWrite_agree with "oE ox") as %?. subst tx'.
          have Letx: (tx ≤ t)%positive by etrans.
          inversion Eqv; rewrite (decide_True _ _ Letx); by iFrame "rBE".
        + inversion Eqv; case decide => ?; by iFrame "rBE".
      - have Eqtm2: (C ∖ cbends C) !! t = Some m by apply lookup_difference_Some.
        rewrite /resD /gps_res (big_sepM_lookup _ _ _ _ Eqtm2). destruct isEx.
        + iDestruct "ox" as "[ox %]".
          iDestruct (ExWrite_agree with "oE ox") as %?. subst tx'.
          have Letx: (tx ≤ t)%positive by etrans.
          inversion Eqv; rewrite (decide_True _ _ Letx);
            by iDestruct "rD" as "[$|$]".
        + inversion Eqv; case decide => ?; iDestruct "rD" as "[$|$]". }
    iClear "Comp". iDestruct "COMP" as %COMP.

    have RRLX: Relaxed ⊑ or by destruct OR; subst or.
    have FRLX: Relaxed ⊑ orf by destruct ORF; subst orf.
    have WRLX: Relaxed ⊑ ow by destruct OW; subst ow.
    iApply wp_fupd.
    iApply (iwp_cas_int l vr vw orf or ow C _ _ _ _ _ _ E E'
              with "s hist at nar aw");
      [done|done|done|exists tr, mr; by inversion Eqvr|done..|].

    iIntros (b rd v' 𝓥' C' t' (WFC&m'&𝓥x&Eqt'&Eqv'&Ext1&Ext2&AT'&CCase)).

    have Eqv'2: memval_val_rel m'.(mval) #v' by rewrite Eqv'; constructor.
    have Eqs' := toState_lookup_state_inv μ _ _ _ _ Eqv'2 Eqt'.
    have Ext3: 𝓥 ⊑ 𝓥' by etrans.

    destruct CCase as [(? & ? & ? & ATW' & RH)
    |(? & RH & ? & m & EqC' & Eqt & Eqv
      & LeRel & Let1 & Let2 & ADD & HACQ & CUR & ATW' & WH)].
    - (* CAS fails *)
      have ?: tr ⊑ t'.
      { change (Some tr ⊑ Some t'). etrans; first apply Extr. by inversion RH. }
      have Exts: sr ⊑ μ t'. { rewrite -Eqsr. apply SORTED;
          [exists mr; by inversion Eqvr|by exists m'; rewrite Eqv'|done]. }

      iSpecialize ("VS" $! t' (μ t') with "[%]"); [done|].
      subst b C'. rewrite bi.and_elim_r.
      iSpecialize ("VS" $! v' with "[%]"); [done|].
      iApply step_fupd_intro; [done|]. iIntros "!> s' hist' at nar aw" (WFC').

      iMod (gps_ownA_M_fork_Reader_singleton _ _ _ _ Eqs' with "oA") as "[oA R']".

      have LeV': default ∅ m'.(mrel) ⊔ Va ⊑ Vs 𝓥'.(cur) 𝓥'.(acq) orf.
      { destruct ORF; subst orf; simpl; (etrans; last by apply Ext2);
          apply lat_join_lub.
        - by eapply read_helper_view_relaxed.
        - etrans; [apply LeVaC|]. etrans; [apply cur_acq|apply Ext1].
        - by eapply read_helper_view_acq.
        - etrans; [apply LeVaC|apply Ext1]. }

      iAssert (|={E}=> R t' (μ t') v' (Vs 𝓥'.(cur) 𝓥'.(acq) orf) ∗
                        resBE IP IW l γF μ C tx' Va ∗
                        resD IP IW l γF μ C tx' Va ∗
                        oEX ∗ gps_ownA_E γ tx')%I
        with "[VS rBE rD ox oE]" as ">(Rd & rBE & rD & ox & oE)".
      { iSpecialize ("VS" $! (default ∅ (mrel m') ⊔ Va)).
        destruct ((cbends C) !! t') as [m2|] eqn:Eq2.
        - iFrame "rD".
          have ?: m2 = m'. { move : Eq2 => /map_filter_lookup_Some [Eq2 _].
              rewrite Eqt' in Eq2. by inversion Eq2. } subst m2.
          rewrite bi.and_elim_r /resBE /gps_res (big_sepM_delete _ _ _ _ Eq2) Eqv'.
          iDestruct "rBE" as "[F $]". destruct isEx.
          + iDestruct "ox" as "[ox %]".
            iDestruct (ExWrite_agree with "oE ox") as %?. subst tx'.
            iFrame "oE ox". have Letx: (tx ≤ t')%positive by etrans.
            rewrite (decide_True _ _ Letx) bi.and_elim_l.
            by iMod ("VS" with "F") as "[$ $]".
          + iFrame "ox oE".
            case decide => ?; [rewrite bi.and_elim_l|rewrite bi.and_elim_r];
              by iMod ("VS" with "F") as "[$ $]".
        - iFrame "rBE".
          have Eqm2: (C ∖ cbends C) !! t' = Some m'
            by apply lookup_difference_Some.
          rewrite /resD /gps_res (big_sepM_delete _ _ _ _ Eqm2) Eqv'.
          iDestruct "rD" as "[F $]". destruct isEx.
          + iDestruct "ox" as "[ox %]".
            iDestruct (ExWrite_agree with "oE ox") as %?. subst tx'.
            iFrame "oE ox". have Letx: (tx ≤ t')%positive by etrans.
            rewrite (decide_True _ _ Letx).
            iDestruct "F" as "[F|F]"; [rewrite bi.and_elim_r|rewrite bi.and_elim_l];
              by iMod ("VS" with "F") as "[$ $]".
          + iFrame "ox oE". case decide => ?.
            * iDestruct "F" as "[F|F]";
                [rewrite bi.and_elim_r bi.and_elim_l|rewrite bi.and_elim_l];
                by iMod ("VS" with "F") as "[$ $]".
            * iDestruct "F" as "[F|[F|F]]";
                [rewrite bi.and_elim_r bi.and_elim_l|rewrite bi.and_elim_l
                |rewrite bi.and_elim_r bi.and_elim_r];
                by iMod ("VS" with "F") as "[$ $]". }

      iDestruct (gps_ownA_split with "[$oA $oE $oN]") as "oA".

      iApply ("Post" $! _ v' (μ t') t'
                (default ∅ m'.(mrel) ⊔ Va) (default ∅ m'.(mrel))
                with "[- $s']"). iSplit.
      { iPureIntro. split; [done|]. split; [by etrans|].
        split; [|done]. eapply seen_local_mono; [by apply Ext2|].
        eapply read_helper_seen_local; [|eauto]. done. }

      iSplitR "Rd R' P ox".
      { iExists μ, C, tx', Va, _,_,_. iFrame.
        iPureIntro. do 3 (split; [done|]). split; [|split; [solve_lat|done]].
        do 2 (split; [done|]). move => ?? /JV [??]. split; [|solve_lat].
        eapply seen_local_mono; [|done]. solve_lat. }

      iRight. iFrame "Rd R' ox". iSplit; [done|].
      move : OW => [->|->]; (rewrite /Vs /= monPred_mono; [by iFrame|]).
      by apply frel_mono. by apply cur_mono.

    - (* CAS succeeds *)
      have ?: tr ⊑ t'.
      { change (Some tr ⊑ Some t'). etrans; first apply Extr. by inversion RH. }
      have Exts: sr ⊑ μ t'.
      { rewrite -Eqsr. apply SORTED;
            [exists mr; by inversion Eqvr|by exists m'; rewrite Eqv'|done]. }

      iSpecialize ("VS" $! t' (μ t') with "[%//]").
      subst b v'. rewrite bi.and_elim_l.

      (* get full interpretation *)
      have Eqm': cbends C !! t' = Some m'.
      { apply map_filter_lookup_Some. split; [done|].
        move => /= t1 m1 Eqm1 NEqt1 Eqf. rewrite /cell_adj /= in Eqf.
        by rewrite Eqf Eqt in Eqm1. }

      rewrite /resBE /gps_res (big_sepM_delete _ _ _ _ Eqm')
              Eqv' /=. iDestruct "rBE" as "(F & rBE)".

      case (decide (tx' ≤ t')%positive) => [Letx|NLe]; last first.
      { rewrite bi.and_elim_l.
        iApply step_fupd_intro; [done|]. iIntros "!> s hist at nar aw" (WFC').
        destruct isEx.
        - iDestruct "ox" as "[ox %]".
          iDestruct (ExWrite_agree with "oE ox") as %?. subst tx'.
          exfalso. apply NLe. by etrans.
        - iSpecialize ("VS" with "P").
          rewrite {1}(monPred_mono _ _ _
                        (lat_join_sqsubseteq_l (Vs 𝓥.(cur) 𝓥.(frel) ow)
                                                (default ∅ m'.(mrel) ⊔ Va))).
          by iMod ("VS" with "F") as "?". }

      rewrite bi.and_elim_r. iDestruct "VS" as "[VS1 VS2]".

      set t:= (t'+1)%positive.
      have Let': (t' < t)%positive by rewrite /t; lia.
      have ?: (tx' < t)%positive by eapply Pos.le_lt_trans.
      have ?: (tr < t)%positive by eapply Pos.le_lt_trans; [|apply Let'].

      iMod ("VS1" $! (default ∅ m'.(mrel) ⊔ Va) with "[$F]") as "[Q1 Q2]".
      iSpecialize ("VS2" with "P").
      rewrite (monPred_mono _ _ _ (lat_join_sqsubseteq_l (Vs 𝓥.(cur) 𝓥.(frel) ow)
                                            (default ∅ m'.(mrel) ⊔ Va))).
      iMod ("VS2" with "[$Q2]") as "[W VS]".
      iDestruct "W" as (ζ') "(#Prt & >W & >MAX)". iDestruct "MAX" as %MAX.
      iAssert (▷ ⌜seen_local l t' (Vs (cur 𝓥) (frel 𝓥) ow
                        ⊔ (default ∅ m'.(mrel) ⊔ Va))⌝)%I as ">#SL".
      { iDestruct "Prt" as (Vw Va' rs') "(naV' & Rw & [$ _])". } iDestruct "SL" as %SL.
      have SLw: seen_local l t' (𝓥.(cur) ⊔ default ∅ m'.(mrel)).
      { eapply seen_local_mono; last exact SL.
        rewrite LeVaC. apply lat_join_lub; [|solve_lat].
        clear -OW. rewrite /Vs.
        destruct OW; subst ow; simpl; [rewrite frel_cur|]; solve_lat. } clear SL.

      iDestruct (Writer_exact with "oA W") as %?. subst ζ'.
      iDestruct "VS" as (s Exts') "VS".

      (* update Writer and get new Reader *)
      have HN: toState μ C !! t = None by rewrite /toState lookup_imap Eqt.
      iMod (Writer_update_insert_Reader _ _ _ _ (s, (vw, default ∅ m.(mrel)))
                                        HN with "oA W") as "(oA & W' & #R')".
      iSpecialize ("VS" $! t with "[//]").

      (* get new interp *)
      have Ext4: Vs 𝓥.(cur) 𝓥.(frel) ow ⊑ default ∅ m.(mrel) ⊔ Va.
      { etrans; last apply lat_join_sqsubseteq_l.
        destruct OW; subst ow; simpl; (etrans; first apply Ext1).
        - have ExtR:= (write_helper_relaxed_mrel _ _ _ _ _ _ WH).
          etrans; last by eapply default_view_bot. solve_lat.
        - have ExtR:= (write_helper_release_mrel _ _ _ _ _ _ WH).
          etrans; last by eapply default_view_bot. solve_lat. }
      have Ext5: default ∅ m'.(mrel) ⊔ Va ⊑ default ∅ m.(mrel) ⊔ Va
        by rewrite LeRel.
      have Ext5': Vs 𝓥.(cur) 𝓥.(frel) ow ⊔ (default ∅ m'.(mrel) ⊔ Va)
                  ⊑ default ∅ m.(mrel) ⊔ Va by apply lat_join_lub.
      rewrite Ext5'.

      iAssert (▷ (PrtSeenView l t s vw γ) (default ∅ (mrel m) ⊔ Va))%I
        with "[Prt]" as "#Prt2".
      { iExists (default ∅ m.(mrel)).
        iDestruct "Prt" as (? Va' ?) "(na & ?)".
        iExists Va',_. iFrame "na R'". iPureIntro. split; [|solve_lat].
        eapply seen_local_mono; [apply lat_join_sqsubseteq_l|].
        eapply write_helper_seen_local_write; last exact WH.
        by destruct OW; subst ow. }

      iAssert ((if dropR then True else oEX) ∗
               (if dropR then oEX else True))%I with "[ox]" as "[ox1 ox2]".
      { destruct dropR; by iFrame "ox". }
      iMod ("VS" with "[W'] [ox2]") as "[FR QF]".
      { iExists _. iFrame "Prt2 W'".
        iPureIntro. move => t1. case (decide (t1 = t)) => [->//|?].
        rewrite lookup_insert_ne; [|done]. move => /MAX Le.
        etrans; [apply Le|]. rewrite /t. apply Pos.lt_le_incl. lia. }
      { iDestruct (PrtSeenView_PrtSeen with "Prt2") as "Prt3".
        destruct dropR; [|by rewrite andb_false_r].
        destruct isEx; [|done]. iExists tx. iDestruct "ox2" as "[$ %]".
        iFrame "Prt3". iPureIntro. by eapply Pos.lt_le_incl, Pos.le_lt_trans. }

      (* get F_read for the read event *)
      iMod ("FR" $! (default ∅ m'.(mrel) ⊔ Va) with "[$Q1]") as "Fr".
      iMod "QF" as "QF". iIntros "!> !>". iMod "QF" as "[Q F]".
      have Ext6: default ∅ m.(mrel) ⊔ Va ⊑ Vs (cur 𝓥') (acq 𝓥') or.
      { rewrite /Vs. destruct OR; subst or; simpl; apply lat_join_lub.
        - apply default_view_bot in HACQ. by rewrite HACQ.
        - etrans; [apply LeVaC|]. etrans; [apply cur_acq|apply Ext3].
        - apply default_view_bot in CUR. by rewrite CUR.
        - etrans; [apply LeVaC|apply Ext3]. }

      iIntros "!> s' hist' at nar aw" (WFC') "!>".
      iApply ("Post" $! _ vr s t (default ∅ m.(mrel) ⊔ Va) (default ∅ m.(mrel))).
      iFrame "s'". iSplit; [|iSplitR "R' Q ox1"; last first].
      { iPureIntro. split; [solve_lat|]. split; [by etrans|done]. }
      { iLeft. iFrame "R' Q ox1". iPureIntro. do 4 (split; [done|]).
        split; last split; [..|done].
        - inversion WH. rewrite /write_Rw decide_True; [|by destruct OW; subst ow].
          etrans; first apply Ext1.
          destruct (𝓥x.(rel) !! l) eqn:HREL; last done. solve_lat.
        - etrans; last by apply lat_join_sqsubseteq_l.
          rewrite /Vs. destruct OW; subst ow; simpl.
          + have ?:= (write_helper_relaxed_mrel_frel _ _ _ _ _ _ WH).
            by destruct m.(mrel).
          + have ?:= write_helper_release_mrel_cur _ _ _ _ _ _ WH.
            by destruct m.(mrel). }

      have Eqv2 : memval_val_rel m.(mval) vw by rewrite Eqv; constructor.
      iExists (<[t := s]> μ), (<[t:=m]> C), tx', Va, _, rs, _.
      rewrite -toState_insert; [|done].
      rewrite gps_ownA_split EqC'. iFrame "hist' at aw nar oA oE oN".

      iDestruct (GPSRaw_CAS_resources IW γF μ C tx' Va t' #vr m' t s vw m
                  Letx Let' with "rBE rD F Fr") as "[$ $]"; [done..|].

      (* TODO: some overlaps with writes *)
      iPureIntro. split; last split; last split; last split; last split;
        last split; last split; last split.
      + move => t0 [m0 [Eqm0 Eqv0]]. move : Eqm0.
        case (decide (t0 = t)) => [->|?];
          [rewrite lookup_insert fn_lookup_insert /=
          |rewrite lookup_insert_ne; last done]. (* TODO: duplicated proofs *)
        { move => _ [td [Ld1 [Ld2 DISJ]]] s0. etrans; last apply Exts'.
          apply FINAL; [by exists m'; rewrite Eqv'|].
          have Lt: (td < t)%positive.
          { apply Pos.le_lteq in Ld2 as [Lt2|];
              [done|subst td; by rewrite lookup_insert in DISJ]. }
          have Led: (td ≤ t')%positive.
          { apply Pos.lt_succ_r. clear -Lt. move : Lt. rewrite /t. lia. }
          exists td. repeat split; [done..|].
          rewrite lookup_insert_ne in DISJ; [done|].
          move => ?. subst t. lia. }
        { move => Eqt0 [td [? [? DISJ]]].
          rewrite/final_st fn_lookup_insert_ne; [|done].
          apply FINAL; [by eexists|]. exists td. repeat split; [done..|].
          move : DISJ. case (decide (t=td)) => [->|?]. by rewrite lookup_insert.
          by rewrite lookup_insert_ne. }
      + move => t1 t2 [m1 [Eq1 ?]] [m2 [Eq2 ?]]. move : Eq1 Eq2.
        case (decide (t1 = t)) => [->|NEq];
          [rewrite lookup_insert fn_lookup_insert
          |rewrite lookup_insert_ne; last done;
           rewrite fn_lookup_insert_ne; last done];
        (case (decide (t2 = t)) => [->|?];
          [rewrite lookup_insert fn_lookup_insert
          |rewrite lookup_insert_ne; last done;
           rewrite fn_lookup_insert_ne; last done]); [done| | |].
        * move => _ St2 Le2. apply FINAL; first by eexists.
          exists t. repeat split; [by apply Pos.lt_le_incl|done|by inversion ADD].
        * move => St1 _ Le1. etrans; last apply Exts'.
          apply SORTED; [by eexists|exists m'; by rewrite Eqv'|].
          case (decide (t1 ≤ t')%positive) => [//|/Pos.lt_nle Lt].
          exfalso. apply NEq. apply : anti_symm; [done|]. rewrite /t. clear -Lt.
          rewrite Pos.add_1_r. by apply Pos.le_succ_l.
        * move => St1 St2. apply SORTED; by eexists.
      + by rewrite -EqC'.
      + eapply inv_view_mono; last apply inv_view_insert_hist; [|done|]; last first.
        * split; [done|].
          split. { eapply atw_local_mono; [done| |done]. solve_lat. }
          move => ?? /JV [??].  split; [|solve_lat].
          eapply seen_local_mono; [|done]. solve_lat.
        * clear -WH JV Eqt' WRLX.
          apply lat_join_lub; [apply lat_join_lub|]; [done|..].
          { inversion WH. rewrite /= (decide_True _ _ WRLX). solve_lat. }
          { apply JV in Eqt' as [_ Eqt']. inversion_clear WH.
            rewrite writeRw_included Eqt' /= (decide_True _ _ WRLX). solve_lat. }
      + solve_lat.
      + done.
      + rewrite (toState_insert _ _ _ _ _ _ Eqv2) lookup_insert_ne; [done|].
        move => ?. subst tx'. by apply (irreflexive_eq (R:= (<)%positive) t t).
      + move => t0 m0. case (decide (t0 = t)) => [->|?].
        rewrite lookup_insert => [[<-]]. by rewrite Eqv.
        rewrite lookup_insert_ne; [|done]. by apply ND.
      + move => t0 [m0].
        case (decide (t0 = t)) => [->|?];
          [rewrite lookup_insert|rewrite lookup_insert_ne; [|done]].
        * move => _ Let3 [td [Le4 [/Pos.le_lteq [Le5|Le5] NIN]]];
            last by rewrite Le5 lookup_insert in NIN.
          rewrite lookup_insert_ne in NIN; last first.
          { move => ?; subst td; lia. }
          apply (RC t'); [by eexists|done|]. exists td. repeat split; auto.
          apply Pos.lt_succ_r. by rewrite -Pos.add_1_r.
        * move => Eqt0 Let0 [td [? [? NIN]]]. apply (RC t0); [by eexists|done|].
          exists td. repeat split; [done..|].
          case (decide (td = t)) => ?;
            [subst td; by rewrite lookup_insert in NIN
            |by rewrite lookup_insert_ne in NIN].
  Qed.

  (* FIXME: duplicating GPSRaw_CAS_int *)
  (* TODO: Q might want to know about the timestamp/state it read *)
  Lemma GPSRaw_CAS_int_strong_write
    IW (tr tx: time) sr vVr orf or ow (vr: Z) (vw: val) γ γF q Vo Va rs 𝓥 E E'
    (P: vProp) (Q Q1 Q2: time → pr_stateT Prtcl → vProp)
    (R : time → pr_stateT Prtcl → lit → vProp) (dropR: bool)
    (Extr : seen_local l tr 𝓥.(cur))
    (LEx: (tx ≤ tr)%positive) (SUB: E' ⊆ E) :
    ↑histN ⊆ E →
    orf = Relaxed ∨ orf = AcqRel →
    or = Relaxed ∨ or = AcqRel → ow = Relaxed ∨ ow = AcqRel →
    let Vs V1 V2 o := if decide (AcqRel ⊑ o) then V1 else V2 in
    let Wv t s v γ : vProp :=
      (∃ ζ, PrtSeenView l t s v γ ∗ ⎡ Writer γ ζ ⎤ ∗
        ⌜∀ t', is_Some (ζ !! t') → t' ⊑ t⌝)%I in
    let Rs t s v : vProp :=
      (∃ (tx: time), PrtSeen l t s v γ ∗ ⎡ exwr γ tx q ⎤ ∗ ⌜(tx ≤ t)%positive⌝)%I in
    {{{ seen 𝓥 ∗ P (Vs 𝓥.(cur) 𝓥.(frel) ow)
        ∗ ▷ (∀ t' s' v, ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
              (F γF t' s' v ∨ F_read γF t' s' v) -∗
              ⌜∃ vl, v = #vl ∧ lit_comparable vr vl⌝) ∅
        ∗ (∀ t' s', ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
            ((<obj> (▷ F γF t' s' #vr ={E}=∗ ▷ Q1 t' s' ∗ ▷ Q2 t' s')) ∗
             (P -∗ ▷ Q2 t' s' ={E}=∗ ▷ Wv t' s' #vr γ ∗
                ∃ s'', ⌜s' ⊑ s''⌝ ∗
                  (∀ t , ⌜(t' < t)%positive⌝ -∗ ▷ Wv t s'' vw γ -∗
                      (if dropR then ▷ Rs t s'' vw else True) ={E}=∗
                      (<obj> (▷ Q1 t' s' ={E}=∗ ▷ F_read γF t' s' #vr)) ∗
                      |={E,E'}▷=> Q t s'' ∗ F γF t s'' vw)) )
               ∧ (▷ ∀ (v: lit), ⌜lit_neq vr v⌝ -∗
                      <obj> ((F_read γF t' s' #v ={E}=∗
                                F_read γF t' s' #v ∗ R t' s' v) ∧
                       (F γF t' s' #v ={E}=∗ F γF t' s' #v ∗ R t' s' v))))
             (Vs 𝓥.(cur) 𝓥.(frel) ow)
        ∗ ▷ (gps_inv IP IW l γ γF true Vo)
        ∗ Reader γ {[tr := (sr, vVr)]}
        ∗ naViewSeen l γ Va rs 𝓥.(cur)
        ∗ exwr γ tx q }}}
      CAS #l #vr vw orf or ow at 𝓥 @ E
    {{{ b (v': lit) s' (t': time) V' Vt' 𝓥', RET #b at 𝓥';
        seen 𝓥'
        ∗ ⌜𝓥 ⊑ 𝓥' ∧ sr ⊑ s' ∧ seen_local l t' 𝓥'.(cur) ∧ Vt' ⊔ Va = V'⌝
        ∗ gps_inv IP IW l γ γF true (Vo ⊔ 𝓥'.(cur))
        ∗ ( (⌜b = true ∧ v' = LitInt vr ∧ (tr < t')%positive
              ∧ 𝓥.(cur) !!w l ⊏ Some t'
              ∧ 𝓥.(rel) !! l ⊑ Some V'
              ∧ Vs 𝓥'.(cur) 𝓥'.(frel) ow ⊑ V'
              ∧ V' ⊑ Vs 𝓥'.(cur) 𝓥'.(acq) or⌝
            ∗ Reader γ {[t' := (s', (vw, Vt'))]}
            ∗ (if dropR then True else exwr γ tx q)
            ∗ Q t' s' (Vs 𝓥'.(cur) 𝓥'.(acq) or))
        ∨ ( ⌜b = false ∧ lit_neq vr v' ∧ (tr ≤ t')%positive
              ∧ V' ⊑ Vs 𝓥'.(cur) 𝓥'.(acq) orf⌝
            ∗ exwr γ tx q
            ∗ Reader γ {[t' := (s', (#v', Vt'))]}
            ∗ R t' s' v' (Vs 𝓥'.(cur) 𝓥'.(acq) orf)
            ∗ P (Vs 𝓥'.(cur) 𝓥'.(frel) ow) ))
        }}}.
  Proof.
    iIntros (? ORF OR OW Vs Wv Rs Φ)
            "(s & P & Comp & VS & Inv & R & naV & ox) Post".
    iApply (GPSRaw_CAS_int_write IW true dropR tr tx q _ _ orf or ow vr vw γ γF
              _ _ _ _ E E' P Q Q1 Q2 R Extr SUB
              with "[$s $P $Comp VS $Inv $R $naV $ox]"); [done..| |].
    { iFrame (LEx). iIntros (t' s' V' Ext' Exts').
      iSpecialize ("VS" $! t' s'). rewrite monPred_at_wand.
      iSpecialize ("VS" $! V' Ext' Exts'). iSplit.
      - iSplit; [done|]. rewrite bi.and_elim_l. by iDestruct "VS" as "[$ $]".
      - rewrite bi.and_elim_r. by setoid_rewrite bi.and_True. }

    iIntros "!>" (b v' s' t' V' Vt' 𝓥') "(s' & FACT & Inv & CASE)".
    iApply ("Post" $! b v' s' t' V' Vt' 𝓥' with "[$s' $FACT $Inv CASE]").
    iDestruct "CASE" as "[(FACT & ox & R & Q)|(FACT & ox & R & Q)]";
      [iLeft|iRight].
    - iFrame "FACT R Q". destruct dropR; [done|]. by iDestruct "ox" as "[$ ?]".
    - iFrame "FACT R Q". by iDestruct "ox" as "[$ ?]".
  Qed.

  Lemma GPSRaw_CAS_int_weak_write
    IW (tr: time) sr vVr orf or ow (vr: Z) (vw: val) γ γF Vo Va rs 𝓥 E E'
    (P: vProp) (Q Q1 Q2: time → pr_stateT Prtcl → vProp)
    (R : time → pr_stateT Prtcl → lit → vProp)
    (Extr : seen_local l tr 𝓥.(cur)) (SUB: E' ⊆ E) :
    ↑histN ⊆ E →
    orf = Relaxed ∨ orf = AcqRel →
    or = Relaxed ∨ or = AcqRel → ow = Relaxed ∨ ow = AcqRel →
    let Vs V1 V2 o := if decide (AcqRel ⊑ o) then V1 else V2 in
    let Wv t s v γ : vProp :=
      (∃ ζ, PrtSeenView l t s v γ ∗ ⎡ Writer γ ζ ⎤ ∗
            ⌜∀ t', is_Some (ζ !! t') → t' ⊑ t⌝)%I in
    {{{ seen 𝓥 ∗ P (Vs 𝓥.(cur) 𝓥.(frel) ow) ∗
        ▷ (∀ t' s' v, ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
              (F γF t' s' v ∨ F_read γF t' s' v ∨ IW l γF t' s' v) -∗
              ⌜∃ vl, v = #vl ∧ lit_comparable vr vl⌝) ∅ ∗
        (∀ t' s', ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
            ((▷ (P -∗ IW l γF t' s' #vr ={E}=∗ False)) ∧
             ((<obj> (▷ F γF t' s' #vr ={E}=∗ ▷ Q1 t' s' ∗ ▷ Q2 t' s')) ∗
             (P -∗ ▷ Q2 t' s' ={E}=∗ ▷ Wv t' s' #vr γ ∗
                ∃ s'', ⌜s' ⊑ s''⌝ ∗
                  (∀ t , ⌜(t' < t)%positive⌝ -∗ ▷ Wv t s'' vw γ ={E}=∗
                      (<obj> (▷ Q1 t' s' ={E}=∗ ▷ F_read γF t' s' #vr)) ∗
                      |={E,E'}▷=> Q t s'' ∗ F γF t s'' vw)) )) ∧
            (▷ ∀ (v: lit), ⌜lit_neq vr v⌝ -∗
              <obj> ((F_read γF t' s' #v ={E}=∗ F_read γF t' s' #v ∗ R t' s' v) ∧
                     (F γF t' s' #v ={E}=∗ F γF t' s' #v ∗ R t' s' v) ∧
                     (IW l γF t' s' #v ={E}=∗ IW l γF t' s' #v ∗ R t' s' v))))
             (Vs 𝓥.(cur) 𝓥.(frel) ow)
        ∗ ▷ (gps_inv IP IW l γ γF true Vo)
        ∗ Reader γ {[tr := (sr, vVr)]}
        ∗ naViewSeen l γ Va rs 𝓥.(cur) }}}
      CAS #l #vr vw orf or ow at 𝓥 @ E
    {{{ b (v': lit) s' (t': time) V' Vt' 𝓥', RET #b at 𝓥';
        seen 𝓥'
        ∗ ⌜𝓥 ⊑ 𝓥' ∧ sr ⊑ s' ∧ seen_local l t' 𝓥'.(cur) ∧ Vt' ⊔ Va = V'⌝
        ∗ gps_inv IP IW l γ γF true (Vo ⊔ 𝓥'.(cur))
        ∗ ( (⌜b = true ∧ v' = LitInt vr ∧ (tr < t')%positive
              ∧ 𝓥.(cur) !!w l ⊏ Some t'
              ∧ 𝓥.(rel) !! l ⊑ Some V'
              ∧ Vs 𝓥'.(cur) 𝓥'.(frel) ow ⊑ V'
              ∧ V' ⊑ Vs 𝓥'.(cur) 𝓥'.(acq) or⌝
            ∗ Reader γ {[t' := (s', (vw, Vt'))]}
            ∗ Q t' s' (Vs 𝓥'.(cur) 𝓥'.(acq) or))
        ∨ ( ⌜b = false ∧ lit_neq vr v' ∧ (tr ≤ t')%positive 
              ∧ V' ⊑ Vs 𝓥'.(cur) 𝓥'.(acq) orf⌝
            ∗ Reader γ {[t' := (s', (#v', Vt'))]}
            ∗ R t' s' v' (Vs 𝓥'.(cur) 𝓥'.(acq) orf)
            ∗ P (Vs 𝓥'.(cur) 𝓥'.(frel) ow) ))
        }}}.
  Proof.
    iIntros (? ORF OR OW Vs Wv Φ) "(s & P & Comp & VS & Inv & R & naV) Post".
    iApply (GPSRaw_CAS_int_write IW false false tr 1 1%Qp _ _ orf or ow vr vw
              γ γF _ _ _ _ E E' P Q Q1 Q2 R Extr SUB
              with "[$s $P $Comp VS $Inv $R $naV]"); [done..| |].
    { iIntros (t' s' V' Ext' Exts').
      iSpecialize ("VS" $! t' s'). rewrite monPred_at_wand.
      iSpecialize ("VS" $! V' Ext' Exts'). iSplit; last by rewrite bi.and_elim_r.
      rewrite bi.and_elim_l. iSplit; first by rewrite bi.and_elim_l.
      rewrite bi.and_elim_r. iDestruct "VS" as "[$ VS]".
      iIntros (? ->) "P". iSpecialize ("VS" with "P").
      iIntros (? ->) "Q". iMod ("VS" with "[$Q]") as "[$ VS]".
      iDestruct "VS" as (s Ext) "VS". iModIntro. iExists s. iFrame (Ext).
      iIntros (t'' ? ExtV Ext''). iSpecialize ("VS" $! t'' Ext'').
      rewrite ExtV. iIntros (? ->) "W". iIntros (? ExtV2) "_".
      iMod ("VS" with "[$W]") as "VS". rewrite ExtV2.
      by iDestruct "VS" as "[$ $]". }

    iIntros "!>" (b v' s' t' V' Vt' 𝓥') "(s' & FACT & Inv & CASE)".
    iApply ("Post" $! b v' s' t' V' Vt' 𝓥' with "[$s' $FACT $Inv CASE]").
    iDestruct "CASE" as "[(FACT & _ & R & Q)|(FACT & _ & R & Q)]";
      [iLeft|iRight]; by iFrame.
  Qed.

  (* CAS for normal protocols *)
  Lemma GPSRaw_CAS_int
    IW (tr tx: time) sr vVr orf or ow (vr: Z) (vw: val) γ γF q Vo Va rs 𝓥 E E'
    (P: vProp) (Q Q1 Q2: time → pr_stateT Prtcl → vProp)
    (R : time → pr_stateT Prtcl → lit → vProp)
    (Extr : seen_local l tr 𝓥.(cur))
    (LEx: (tx ≤ tr)%positive) (SUB: E' ⊆ E) :
    ↑histN ⊆ E →
    orf = Relaxed ∨ orf = AcqRel →
    or = Relaxed ∨ or = AcqRel → ow = Relaxed ∨ ow = AcqRel →
    let Vs V1 V2 o := if decide (AcqRel ⊑ o) then V1 else V2 in
    {{{ seen 𝓥 ∗ P (Vs 𝓥.(cur) 𝓥.(frel) ow) ∗
        ▷ (∀ t' s' v, ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
              (F γF t' s' v ∨ F_read γF t' s' v) -∗
              ⌜∃ vl, v = #vl ∧ lit_comparable vr vl⌝) ∅ ∗
        (∀ t' s', ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
            ((<obj> (▷ F γF t' s' #vr ={E}=∗ ▷ Q1 t' s' ∗ ▷ Q2 t' s')) ∗
             (P -∗ ▷ Q2 t' s' ={E}=∗ ▷ ⎡ ∃ ζ, Writer γ ζ ⎤ ∗
                ∃ s'', ⌜s' ⊑ s''⌝ ∗
                  (∀ t , ⌜(t' < t)%positive⌝ -∗
                    ▷ (⎡ (∃ ζ, Writer γ ζ) ⎤ ∗ PrtSeen l t s'' vw γ) ={E}=∗
                      (<obj> (▷ Q1 t' s' ={E}=∗ ▷ F_read γF t' s' #vr)) ∗
                      |={E,E'}▷=> Q t s'' ∗ F γF t s'' vw)) )
               ∧ (▷ ∀ (v: lit), ⌜lit_neq vr v⌝ -∗
                      <obj> ((F_read γF t' s' #v ={E}=∗
                                F_read γF t' s' #v ∗ R t' s' v) ∧
                       (F γF t' s' #v ={E}=∗ F γF t' s' #v ∗ R t' s' v))))
             (Vs 𝓥.(cur) 𝓥.(frel) ow) ∗
        ▷ (gps_inv IP IW l γ γF false Vo) ∗
        Reader γ {[tr := (sr, vVr)]} ∗
        naViewSeen l γ Va rs 𝓥.(cur) ∗
        exwr γ tx q }}}
      CAS #l #vr vw orf or ow at 𝓥 @ E
    {{{ b (v': lit) s' (t': time) V' Vt' 𝓥', RET #b at 𝓥';
        seen 𝓥'
        ∗ ⌜𝓥 ⊑ 𝓥' ∧ sr ⊑ s'
          ∧ seen_local l t' 𝓥'.(cur) ∧ Vt' ⊔ Va = V'⌝
        ∗ gps_inv IP IW l γ γF false (Vo ⊔ 𝓥'.(cur))
        ∗ exwr γ tx q
        ∗ ( (⌜b = true ∧ v' = LitInt vr ∧ (tr < t')%positive
              ∧ 𝓥.(cur) !!w l ⊏ Some t'
              ∧ 𝓥.(rel) !! l ⊑ Some V'
              ∧ Vs 𝓥'.(cur) 𝓥'.(frel) ow ⊑ V'
              ∧ V' ⊑ Vs 𝓥'.(cur) 𝓥'.(acq) or⌝
            ∗ Reader γ {[t' := (s', (vw, Vt'))]}
            ∗ Q t' s' (Vs 𝓥'.(cur) 𝓥'.(acq) or))
        ∨ ( ⌜b = false ∧ lit_neq vr v' ∧ (tr ≤ t')%positive
              ∧ V' ⊑ Vs 𝓥'.(cur) 𝓥'.(acq) orf⌝
            ∗ Reader γ {[t' := (s', (#v', Vt'))]}
            ∗ R t' s' v' (Vs 𝓥'.(cur) 𝓥'.(acq) orf)
            ∗ P (Vs 𝓥'.(cur) 𝓥'.(frel) ow) ))
        }}}.
  Proof.
    iIntros (? ORF OR OW Vs Φ)
            "(s & P & Comp & VS & Inv & #R & naV & ox) Post".
    iDestruct "Inv" as (μ C tx' Va' rsa rs' ws)
                        "(>hist & >at & >aw & >nar & >oA & rBE & rD & >H)".
    iDestruct (naViewSeen_local_seen with "oA naV") as %(?&?&LeVaC&LNA&LNA2).
    subst Va' rs'. iClear "naV".
    iDestruct "H" as %(FINAL&SORTED&GOOD&[AT [AW JV]]&LeVa & NAL & IS & ND & RC).

    iDestruct (gps_ownA_Reader_singleton_sub with "oA R") as %EQL.
    have Eqsr := toState_lookup_state _ _ _ _ _ EQL.
    destruct (toState_lookup_message _ _ _ _ _ EQL) as [mr [Eqr [Eqvr EqVr]]].

    iDestruct (gps_ownA_split with "oA") as "[oA [oE #oN]]".
    iDestruct (ExWrite_agree with "oE ox") as %?. subst tx'.

    iAssert (▷ ⌜∀ t m v, C !! t = Some m → 𝓥.(cur) !!w l ⊑ Some t
                → memval_val_rel m.(mval) v
                → ∃ vl : lit, v = #vl ∧ lit_comparable vr vl⌝)%I
    with "[rBE rD Comp]" as "#> COMP".
    { iNext. iIntros (t m v Eqt Le Eqv).
      have Eqs := toState_lookup_state_inv μ _ _ _ _ Eqv Eqt.
      have Letr: (tr ≤ t)%positive.
      { change (Some tr ⊑ Some t). by etrans; [apply Extr|apply Le]. }
      have Les: sr ⊑ μ t.
      { rewrite -Eqsr. apply SORTED; [exists mr; by inversion Eqvr
                                      |exists m; by inversion Eqv|done]. }
      rewrite (monPred_mono _ _ _ (lat_bottom_sqsubseteq (default ∅ m.(mrel) ⊔ Va))).
      iApply ("Comp" $! t (μ t) v with "[%] [-]"); [done|..].
      have Letx: (tx ≤ t)%positive by etrans.
      destruct ((cbends C) !! t) as [tm|] eqn:Eqtm.
      - rewrite /resBE /gps_res (big_sepM_lookup _ _ _ _ Eqtm).
        have ?: tm = m.
        { move : Eqtm => /map_filter_lookup_Some [Eqtm _]. rewrite Eqt in Eqtm.
          by inversion Eqtm. } subst tm.
        inversion Eqv; rewrite (decide_True _ _ Letx); by iFrame "rBE".
      - have Eqtm2: (C ∖ cbends C) !! t = Some m by apply lookup_difference_Some.
        rewrite /resD /gps_res (big_sepM_lookup _ _ _ _ Eqtm2).
        inversion Eqv; rewrite (decide_True _ _ Letx);
          by iDestruct "rD" as "[$|$]". } iClear "Comp".
    iDestruct "COMP" as %COMP.

    have RRLX: Relaxed ⊑ or by destruct OR; subst or.
    have FRLX: Relaxed ⊑ orf by destruct ORF; subst orf.
    have WRLX: Relaxed ⊑ ow by destruct OW; subst ow.
    iApply wp_fupd.
    iApply (iwp_cas_int l vr vw orf or ow C _ _ _ _ _ _ E E'
              with "s hist at nar aw");
      [done|done|done|exists tr, mr; by inversion Eqvr|done..|].


    iIntros (b rd v' 𝓥' C' t' (WFC&m'&𝓥x&Eqt'&Eqv'&Ext1& Ext2 & AT' & CCase)).

    have Eqv'2: memval_val_rel m'.(mval) #v' by rewrite Eqv'; constructor.
    have Eqs' := toState_lookup_state_inv μ _ _ _ _ Eqv'2 Eqt'.
    have Ext3: 𝓥 ⊑ 𝓥' by etrans.

    destruct CCase as [(? & ? & ? & ATW' & RH)
    |(? & RH & ? & m & EqC' & Eqt & Eqv
      & LeRel & Let1 & Let2 & ADD & HACQ & CUR & ATW' & WH)].
    - (* CAS fails *)
      have ?: tr ⊑ t'.
      { change (Some tr ⊑ Some t'). etrans; first apply Extr. by inversion RH. }
      have Exts: sr ⊑ μ t'. { rewrite -Eqsr. apply SORTED;
          [exists mr; by inversion Eqvr|by exists m'; rewrite Eqv'|done]. }
      have Letx: (tx ≤ t')%positive by etrans.

      iSpecialize ("VS" $! t' (μ t') with "[%]"); [done|].
      subst b C'. rewrite bi.and_elim_r.
      iSpecialize ("VS" $! v' with "[%]"); [done|].
      iApply step_fupd_intro; [done|]. iIntros "!> s' hist' at nar aw" (WFC').

      iMod (gps_ownA_M_fork_Reader_singleton _ _ _ _ Eqs' with "oA") as "[oA R']".
      iDestruct (gps_ownA_split with "[$oA $oE $oN]") as "oA".

      have LeV': default ∅ m'.(mrel) ⊔ Va ⊑ Vs 𝓥'.(cur) 𝓥'.(acq) orf.
      { destruct ORF; subst orf; simpl; (etrans; last by apply Ext2);
          apply lat_join_lub.
        - by eapply read_helper_view_relaxed.
        - etrans; [apply LeVaC|]. etrans; [apply cur_acq|apply Ext1].
        - by eapply read_helper_view_acq.
        - etrans; [apply LeVaC|apply Ext1]. }

      iAssert (|={E}=> R t' (μ t') v' (Vs 𝓥'.(cur) 𝓥'.(acq) orf) ∗
                        resBE IP IW l γF μ C tx Va ∗
                        resD IP IW l γF μ C tx Va)%I
        with "[VS rBE rD]" as ">(Rd & rBE & rD)".
      { iSpecialize ("VS" $! (default ∅ (mrel m') ⊔ Va)).
        destruct ((cbends C) !! t') as [m2|] eqn:Eq2.
        - iFrame "rD".
          have ?: m2 = m'. { move : Eq2 => /map_filter_lookup_Some [Eq2 _].
              rewrite Eqt' in Eq2. by inversion Eq2. } subst m2.
          rewrite bi.and_elim_r /resBE /gps_res (big_sepM_delete _ _ _ _ Eq2)
                  Eqv' (decide_True _ _ Letx).
          iDestruct "rBE" as "[F $]". by iMod ("VS" with "F") as "[$ $]".
        - iFrame "rBE".
          have Eqm2: (C ∖ cbends C) !! t' = Some m'
            by apply lookup_difference_Some.
          rewrite /resD /gps_res (big_sepM_delete _ _ _ _ Eqm2)
                  Eqv' (decide_True _ _ Letx). iDestruct "rD" as "[F $]".
          iDestruct "F" as "[F|F]"; [rewrite bi.and_elim_r|rewrite bi.and_elim_l];
            by iMod ("VS" with "F") as "[$ $]". }

      iApply ("Post" $! _ v' (μ t') t' (default ∅ m'.(mrel) ⊔ Va)
                        (default ∅ m'.(mrel)) with "[- $s' $ox]"). iSplit.
      { iPureIntro. split; [done|]. split; [by etrans|].
        split; [|done]. eapply seen_local_mono; [by apply Ext2|].
        eapply read_helper_seen_local; [|eauto]. done. }

      iSplitR "Rd R' P".
      { iExists μ, C, tx, Va, _, rs, _. iFrame.
        iPureIntro. do 3 (split; [done|]). split; [|split; [solve_lat|done]].
        do 2 (split; [done|]). move => ?? /JV [??]. split; [|solve_lat].
        eapply seen_local_mono; [|done]. solve_lat. }

      iRight. iFrame "Rd R'". iSplit; [done|].
      move : OW => [->|->]; (rewrite /Vs /= monPred_mono; [by iFrame|]).
      by apply frel_mono. by apply cur_mono.

    - (* CAS succeeds *)
      have ?: tr ⊑ t'.
      { change (Some tr ⊑ Some t'). etrans; first apply Extr. by inversion RH. }
      have Exts: sr ⊑ μ t'. { rewrite -Eqsr. apply SORTED;
        [exists mr; by inversion Eqvr|by exists m'; rewrite Eqv'|done]. }

      iSpecialize ("VS" $! t' (μ t') with "[%]"); [done|].
      subst b v'. rewrite bi.and_elim_l. iDestruct "VS" as "[VS1 VS2]".

      set t:= (t'+1)%positive.
      have Letx: (tx ≤ t')%positive by etrans.
      have Let': (t' < t)%positive by rewrite /t; lia.
      have ?: (tx < t)%positive by eapply Pos.le_lt_trans.
      have ?: (tr < t)%positive by eapply Pos.le_lt_trans; [|apply Let'].

      (* get full interpretation *)
      have Eqm': cbends C !! t' = Some m'.
      { apply map_filter_lookup_Some. split; [done|].
        move => /= t1 m1 Eqm1 NEqt1 Eqf. rewrite /cell_adj /= in Eqf.
        by rewrite Eqf Eqt in Eqm1. }

      rewrite /resBE /gps_res (big_sepM_delete _ _ _ _ Eqm')
              Eqv' (decide_True _ _ Letx) /=. iDestruct "rBE" as "(F & rBE)".

      iMod ("VS1" $! (default ∅ m'.(mrel) ⊔ Va) with "[$F]") as "[Q1 Q2]".
      iSpecialize ("VS2" with "P").

      rewrite (monPred_mono _ _ _ (lat_join_sqsubseteq_l (Vs 𝓥.(cur) 𝓥.(frel) ow)
                                               (default ∅ m'.(mrel) ⊔ Va))).
      iMod ("VS2" with "[$Q2]") as "[W VS]".
      iDestruct "W" as (ζ') ">W". iDestruct (Writer_exact with "oA W") as %?.
      subst ζ'. iDestruct "VS" as (s Exts') "VS".

      (* update Writer and get new Reader *)
      have HN: toState μ C !! t = None by rewrite /toState lookup_imap Eqt.
      iMod (Writer_update_insert_Reader _ _ _ _ (s, (vw, default ∅ m.(mrel)))
                                        HN with "oA W") as "(oA & W' & R')".
      iSpecialize ("VS" $! t with "[//]").

      have Ext4: Vs 𝓥.(cur) 𝓥.(frel) ow ⊑ default ∅ m.(mrel) ⊔ Va.
      { etrans; last apply lat_join_sqsubseteq_l.
        destruct OW; subst ow; simpl; (etrans; first apply Ext1).
        - have ExtR:= (write_helper_relaxed_mrel _ _ _ _ _ _ WH).
          etrans; last by eapply default_view_bot. solve_lat.
        - have ExtR:= (write_helper_release_mrel _ _ _ _ _ _ WH).
          etrans; last by eapply default_view_bot. solve_lat. }
      have Ext5: default ∅ m'.(mrel) ⊔ Va ⊑ default ∅ m.(mrel) ⊔ Va
        by rewrite LeRel.
      have Ext5': Vs 𝓥.(cur) 𝓥.(frel) ow ⊔ (default ∅ m'.(mrel) ⊔ Va)
                  ⊑ default ∅ m.(mrel) ⊔ Va by apply lat_join_lub.
      rewrite Ext5'.

      iAssert (▷ (PrtSeen l t s vw γ) (default ∅ (mrel m) ⊔ Va))%I
        with "[R']" as "#Prt2".
      { iExists (default ∅ m.(mrel)), Va, _. iNext. rewrite naViewSeen_eq.
        iFrame "R' oN". iPureIntro. split; [split; first solve_lat|].
        eapply na_local_mono; [done| |apply NAL]. solve_lat.
        eapply seen_local_mono; [apply lat_join_sqsubseteq_l|].
        eapply write_helper_seen_local_write; last exact WH.
        by destruct OW; subst ow. }

      (* get new interp *)
      iMod ("VS" with "[W']") as "[FR QF]".
      { iFrame "Prt2". by iExists _. }
      (* get F_read for the read event *)
      iMod ("FR" $! (default ∅ m'.(mrel) ⊔ Va) with "[$Q1]") as "Fr".
      iMod "QF" as "QF". iIntros "!> !>". iMod "QF" as "[Q F]".

      have Ext6: default ∅ m.(mrel) ⊔ Va ⊑ Vs (cur 𝓥') (acq 𝓥') or.
      { rewrite /Vs. destruct OR; subst or; simpl; apply lat_join_lub.
        - apply default_view_bot in HACQ. by rewrite HACQ.
        - etrans; [apply LeVaC|]. etrans; [apply cur_acq|apply Ext3].
        - apply default_view_bot in CUR. by rewrite CUR.
        - etrans; [apply LeVaC|apply Ext3]. }

      iIntros "!> s' hist' at nar aw" (WFC') "!>".
      iApply ("Post" $! _ vr s t (default ∅ m.(mrel) ⊔ Va) (default ∅ m.(mrel))).
      iFrame "s' ox". iSplit; [|iSplitR "R' Q"; last first].
      { iPureIntro. split; [solve_lat|]. split; [by etrans|done]. }
      { iLeft. iFrame "R' Q". iPureIntro. do 4 (split; [done|]).
        split; last split; [..|done].
        - inversion WH. rewrite /write_Rw decide_True; [|by destruct OW; subst ow].
          etrans; first apply Ext1.
          destruct (𝓥x.(rel) !! l) eqn:HREL; last done. solve_lat.
        - etrans; last by apply lat_join_sqsubseteq_l.
          rewrite /Vs. destruct OW; subst ow; simpl.
          + have ?:= (write_helper_relaxed_mrel_frel _ _ _ _ _ _ WH).
            by destruct m.(mrel).
          + have ?:= write_helper_release_mrel_cur _ _ _ _ _ _ WH.
            by destruct m.(mrel). }

      have Eqv2 : memval_val_rel m.(mval) vw by rewrite Eqv; constructor.
      iExists (<[t := s]> μ), (<[t:=m]> C), tx, Va, _, rs, _.
      rewrite -toState_insert; [|done].
      rewrite gps_ownA_split EqC'. iFrame "hist' at aw nar oA oE oN".

      iDestruct (GPSRaw_CAS_resources IW γF μ C tx Va t' #vr m' t s vw m
                  Letx Let' with "rBE rD F Fr") as "[$ $]"; [done..|].

      (* TODO: some overlaps with writes *)
      iPureIntro. split; last split; last split; last split; last split;
        last split; last split.
      + move => t0 [m0 [Eqm0 Eqv0]]. move : Eqm0.
        case (decide (t0 = t)) => [->|?];
          [rewrite lookup_insert fn_lookup_insert /=
          |rewrite lookup_insert_ne; last done]. (* TODO: duplicated proofs *)
        { move => _ [td [Ld1 [Ld2 DISJ]]] s0. etrans; last apply Exts'.
          apply FINAL; [by exists m'; rewrite Eqv'|].
          have Lt: (td < t)%positive.
          { apply Pos.le_lteq in Ld2 as [Lt2|];
              [done|subst td; by rewrite lookup_insert in DISJ]. }
          have Led: (td ≤ t')%positive.
          { apply Pos.lt_succ_r. clear -Lt. move : Lt. rewrite /t. lia. }
          exists td. repeat split; [done..|].
          rewrite lookup_insert_ne in DISJ; [done|].
          move => ?. subst t. lia. }
        { move => Eqt0 [td [? [? DISJ]]].
          rewrite/final_st fn_lookup_insert_ne; [|done].
          apply FINAL; [by eexists|]. exists td. repeat split; [done..|].
          move : DISJ. case (decide (t=td)) => [->|?]. by rewrite lookup_insert.
          by rewrite lookup_insert_ne. }
      + move => t1 t2 [m1 [Eq1 ?]] [m2 [Eq2 ?]]. move : Eq1 Eq2.
        case (decide (t1 = t)) => [->|NEq];
          [rewrite lookup_insert fn_lookup_insert
          |rewrite lookup_insert_ne; last done;
           rewrite fn_lookup_insert_ne; last done];
        (case (decide (t2 = t)) => [->|?];
          [rewrite lookup_insert fn_lookup_insert
          |rewrite lookup_insert_ne; last done;
           rewrite fn_lookup_insert_ne; last done]); [done| | |].
        * move => _ St2 Le2. apply FINAL; first by eexists.
          exists t. repeat split; [by apply Pos.lt_le_incl|done|by inversion ADD].
        * move => St1 _ Le1. etrans; last apply Exts'.
          apply SORTED; [by eexists|exists m'; by rewrite Eqv'|].
          case (decide (t1 ≤ t')%positive) => [//|/Pos.lt_nle Lt].
          exfalso. apply NEq. apply : anti_symm; [done|]. rewrite /t. clear -Lt.
          rewrite Pos.add_1_r. by apply Pos.le_succ_l.
        * move => St1 St2. apply SORTED; by eexists.
      + by rewrite -EqC'.
      + eapply inv_view_mono; last apply inv_view_insert_hist; [|done|]; last first.
        * split; [done|].
          split. { eapply atw_local_mono; [done| |done]. solve_lat. }
          move => ?? /JV [??].  split; [|solve_lat].
          eapply seen_local_mono; [|done]. solve_lat.
        * clear -WH JV Eqt' WRLX.
          apply lat_join_lub; [apply lat_join_lub|]; [done|..].
          { inversion WH. rewrite /= (decide_True _ _ WRLX). solve_lat. }
          { apply JV in Eqt' as [_ Eqt']. inversion_clear WH.
            rewrite writeRw_included Eqt' /= (decide_True _ _ WRLX). solve_lat. }
      + solve_lat.
      + done.
      + rewrite (toState_insert _ _ _ _ _ _ Eqv2) lookup_insert_ne; [done|].
        move => ?. subst tx. by apply (irreflexive_eq (R:= (<)%positive) t t).
      + split; [|done]. move => t0 m0. case (decide (t0 = t)) => [->|?].
        rewrite lookup_insert => [[<-]]. by rewrite Eqv.
        rewrite lookup_insert_ne; [|done]. by apply ND.
  Qed.

  Lemma GPSRaw_CAS_int_simple
    IW (tr tx: time) sr vVr orf or ow (vr: Z) (vw: val) γ γF q Vo Va rs 𝓥 E
    (P: vProp) (Q Q1 Q2: time → pr_stateT Prtcl → vProp)
    (R : time → pr_stateT Prtcl → lit → vProp)
    (Extr : seen_local l tr 𝓥.(cur)) (LEx: (tx ≤ tr)%positive) :
    ↑histN ⊆ E →
    orf = Relaxed ∨ orf = AcqRel →
    or = Relaxed ∨ or = AcqRel → ow = Relaxed ∨ ow = AcqRel →
    let Vs V1 V2 o := if decide (AcqRel ⊑ o) then V1 else V2 in
    {{{ seen 𝓥 ∗ P (Vs 𝓥.(cur) 𝓥.(frel) ow)  ∗
        ▷ (∀ t' s' v, ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
              (F γF t' s' v ∨ F_read γF t' s' v) -∗
              ⌜∃ vl, v = #vl ∧ lit_comparable vr vl⌝) ∅ ∗
        (∀ t' s', ⌜sr ⊑ s' ∧ tr ⊑ t'⌝ -∗
            ((<obj> (▷ F γF t' s' #vr ={E}=∗ ▷ Q1 t' s' ∗ ▷ Q2 t' s')) ∗
             (P -∗ ▷ Q2 t' s' ={E}=∗ ∃ s'', ⌜s' ⊑ s''⌝ ∗
                  (∀ t , ⌜(t' < t)%positive⌝ -∗ ▷ (PrtSeen l t s'' vw γ) ={E}=∗
                      (<obj> (▷ Q1 t' s' ={E}=∗ ▷ F_read γF t' s' #vr)) ∗
                      |={E}▷=> Q t s'' ∗ F γF t s'' vw)) )
               ∧ (▷ ∀ (v: lit), ⌜lit_neq vr v⌝ -∗
                      <obj> ((F_read γF t' s' #v ={E}=∗
                                F_read γF t' s' #v ∗ R t' s' v) ∧
                       (F γF t' s' #v ={E}=∗ F γF t' s' #v ∗ R t' s' v))))
             (Vs 𝓥.(cur) 𝓥.(frel) ow)
        ∗ ▷ (gps_inv IP IW l γ γF false Vo)
        ∗ Reader γ {[tr := (sr, vVr)]}
        ∗ naViewSeen l γ Va rs 𝓥.(cur)
        ∗ exwr γ tx q
        ∗ (∃ ζ, Writer γ ζ) }}}
      CAS #l #vr vw orf or ow at 𝓥 @ E
    {{{ b (v': lit) s' (t': time) V' Vt' 𝓥', RET #b at 𝓥';
        seen 𝓥'
        ∗ ⌜𝓥 ⊑ 𝓥' ∧ sr ⊑ s'
          ∧ seen_local l t' 𝓥'.(cur) ∧ Vt' ⊔ Va = V'⌝
        ∗ gps_inv IP IW l γ γF false (Vo ⊔ 𝓥'.(cur))
        ∗ exwr γ tx q
        ∗ (∃ ζ, Writer γ ζ)
        ∗ ( (⌜b = true ∧ v' = LitInt vr ∧ (tr < t')%positive
              ∧ 𝓥.(cur) !!w l ⊏ Some t'
              ∧ 𝓥.(rel) !! l ⊑ Some V'
              ∧ Vs 𝓥'.(cur) 𝓥'.(frel) ow ⊑ V'
              ∧ V' ⊑ Vs 𝓥'.(cur) 𝓥'.(acq) or⌝
            ∗ Reader γ {[t' := (s', (vw, Vt'))]}
            ∗ Q t' s' (Vs 𝓥'.(cur) 𝓥'.(acq) or))
        ∨ ( ⌜b = false ∧ lit_neq vr v' ∧ (tr ≤ t')%positive
              ∧ V' ⊑ Vs 𝓥'.(cur) 𝓥'.(acq) orf⌝
            ∗ Reader γ {[t' := (s', (#v', Vt'))]}
            ∗ R t' s' v' (Vs 𝓥'.(cur) 𝓥'.(acq) orf)
            ∗ P (Vs 𝓥'.(cur) 𝓥'.(frel) ow) ))
        }}}.
  Proof.
    iIntros (? ORF OR OW Vs Φ)
            "(s & P & Comp & VS & Inv & R & naV & ox & W) Post".

    set Q': time → pr_stateT Prtcl → vProp := (λ t s, Q t s  ∗ ⎡ ∃ ζ, Writer γ ζ ⎤)%I.
    set R': time → pr_stateT Prtcl → lit → vProp
      := (λ t s (v: lit), R t s v ∗ ⎡ ∃ ζ, Writer γ ζ ⎤)%I.

    iApply (GPSRaw_CAS_int IW tr tx sr vVr orf or ow vr vw γ γF q Vo Va rs 𝓥 E E
                           P Q' Q1 Q2 R' Extr LEx
            with "[$s $P $Comp VS W $Inv $R $naV $ox]"); [done..| |].
    { iIntros (t' s' ?? Le). iSpecialize ("VS" $! t' s' Le).
      iSplit; [rewrite bi.and_elim_l|rewrite bi.and_elim_r].
      - iDestruct "VS" as "[$ VS]". iIntros (??) "P". iIntros (? Ext2) "Q2".
        iFrame "W". rewrite Ext2. iMod ("VS" with "P [$Q2]") as (s'') "(Le' & VS)".
        iModIntro. iExists s''. iFrame "Le'".
        iIntros (t ? Ext3 Lt). iSpecialize ("VS" $! t Lt). rewrite Ext3.
        iIntros (? ->) "[W Prt]". iMod ("VS" with "[$Prt]") as "[$ VS]".
        iMod "VS". iIntros "!> !> !>". iFrame "W". by iMod "VS" as "[$ $]".
      - iNext. iIntros (v ?? NEq). iSpecialize ("VS" $! v NEq).
        rewrite 2!monPred_at_objectively. iIntros (V). iSpecialize ("VS" $! V).
        iSplit; [rewrite bi.and_elim_l|rewrite bi.and_elim_r];
          iIntros (??) "F"; iFrame "W"; by iApply ("VS" with "F"). }

    iNext. iIntros (b v' s' t' V' Vt' 𝓥') "(s & FACTS & Inv & ox & CASE)".

    iApply ("Post" $! b v' s' t' V' Vt' with "[- $s $FACTS $Inv $ox]").
    iDestruct "CASE" as "[(FACTS & R & Q & $)|(FACTS & R & (Rs & $) & P)]".
    - iLeft. by iFrame.
    - iRight. by iFrame.
  Qed.
End RawRules.
